// New style comment marker used to force error in 16-bit dBW/VdB
//---------------------------------------------------------------------
// Program......: Center32.PRG
// Programmer...: Jim Sare (jimsare@ameritech.net)
// Last Modified: 04/16/1998
// Notes........: Adaptation of original 16-bit version of
//              : Center.PRG for use with VdB7. This is to accommodate
//              : the migration of 16-bit apps to 32-bit.
//              : Contains routines to:
//              : Determine screen size in pixels.
//              : Determine workarea size (screen excluding taskbar) in pixels.
//              : Determine number of pixels/form unit (Turpin).
//              : Determine window client area size in pixels.
//              : Determine window size in pixels.
//              : Center a form in the screen or the VdB MDI client area.
//              : Position a form to 1 of 9 positions.
//              : Move forms to any location faster than the Move method.
//              : Determine VdB MDI client area size in form units.
//              : Place UI objects on the VdB desktop.
// Written for..: Visual dBASE (32-bit versions)
// Rev. History.: 03/17/1995 -- Original
//              : 05/15/1995 -- Added MoveForm, SetAppClientControl,
//              :               CenterAppClientControl procedures and
//              :               changed <> to #.
//              : 05/16/1995 -- Additional comments added.
//              : 06/21/1995 -- Added and updated comments.
//              : 07/31/1995 -- Fixed 'Calls' in SetAppClientControl coments.
//              : 08/03/1995 -- Changed parameters to locals.
//              : 09/29/1996 -- Correction to CenterForm for child windows.
//              : 12/14/1997 -- Initial adaptation for 32-bit.
//              : 04/11/1998 -- Added PositionForm, GetWorkareaX, and
//              :               GetWorkareaY procedures.
//              :               Changed CenterForm to use PositionForm.
//              :               Minor comment cleanup.
//              : 04/16/1998 -- Added offset parameters to PositionForm.
//  Usage:......: SET PROCEDURE TO :dUFLP:center32 ADDITIVE
//                ? GetScrenX() // or whatever you need to do from here
//                ** Note that this program can be used with the
//                ** dUFLP source alias, instructions on setting
//                ** this up are in the WHATS.NEW and README.TXT
//                ** files. If you use this source alias, be sure
//                ** your project (when building your executable)
//                ** includes the full path, as the project explorer
//                ** does not understand source aliases.
//---------------------------------------------------------------------

msgbox( "This is not 'running' program, but a collection of useful"+;
        "functions and procedures ... see header for details.",;
        "Can't run this!", 16)
return

FUNCTION GetScreenX // Returns screen width in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine screen width in pixels.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 05/15/1995 -- Changed <> to #.
//             : 06/21/1995 -- Updated comments.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: None
//             : WinAPI: GetSystemMetrics
// Called by...: Any
// Usage.......: nScreenWidth = GetScreenX()
// Returns.....: Numeric -- Screen width in pixels.
// Parameters..: None
//---------------------------------------------------------------------

	If Type("GetSystemMetrics") # "FP"
		extern CINT GetSystemMetrics(CINT) User32
	EndIf
	#define SM_CXSCREEN		0		// Screen width in pixels
	RETURN GetSystemMetrics(SM_CXSCREEN)
// EOP: GetScreenX

FUNCTION GetScreenY // Returns screen height in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine screen height in pixels.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 05/15/1995 -- Changed <> to #.
//             : 06/21/1995 -- Updated comments.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: None
//             : WinAPI: GetSystemMetrics
// Called by...: Any
// Usage.......: nScreenHeight = GetScreenY()
// Returns.....: Numeric -- Screen height in pixels.
// Parameters..: None
//---------------------------------------------------------------------

	If Type("GetSystemMetrics") # "FP"
		extern CINT GetSystemMetrics(CINT) User32
	endIf
	#define SM_CYSCREEN		1		// Screen height in pixels
	RETURN GetSystemMetrics(SM_CYSCREEN)
// EOP: GetScreenY

FUNCTION GetWorkareaX // Returns width of screen workarea in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 04/11/1998
// Notes.......: Determine width of screen excluding taskbar in pixels.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 04/11/1998 -- Original
// Calls.......: None
//             : WinAPI: SystemParametersInfo
// Called by...: Any
// Usage.......: nWidth = GetWorkareaX()
// Returns.....: Numeric -- Workarea width in pixels.
// Parameters..: None
//---------------------------------------------------------------------
	LOCAL xRECT

	If Type("SystemParametersInfo") # "FP"
		extern CLOGICAL SystemParametersInfo(CUINT, CUINT, CPTR, CUINT) User32;
						From "SystemParametersInfoA"
	EndIf
	#ifndef SPI_GETWORKAREA
		#define SPI_GETWORKAREA	48
	#endif
	xRECT = Space(8)
	SystemParametersInfo(SPI_GETWORKAREA, 0, xRECT, 0)
	RETURN Int(xRECT.GetByte(8) +;
				BitLShift(xRECT.GetByte(9), 8) +;
				BitLShift(xRECT.GetByte(10), 16) +;
				BitLShift(xRECT.GetByte(11), 24))
// EOP: GetWorkareaX

FUNCTION GetWorkareaY // Returns height of screen workarea in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 04/11/1998
// Notes.......: Determine height of screen excluding taskbar in pixels.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 04/11/1998 -- Original
// Calls.......: None
//             : WinAPI: SystemParametersInfo
// Called by...: Any
// Usage.......: nHeight = GetWorkareaY()
// Returns.....: Numeric -- Workarea height in pixels.
// Parameters..: None
//---------------------------------------------------------------------
	LOCAL xRECT

	If Type("SystemParametersInfo") # "FP"
		extern CLOGICAL SystemParametersInfo(CUINT, CUINT, CPTR, CUINT) User32;
						From "SystemParametersInfoA"
	EndIf
	#ifndef SPI_GETWORKAREA
		#define SPI_GETWORKAREA	48
	#endif
	xRECT = Space(8)
	SystemParametersInfo(SPI_GETWORKAREA, 0, xRECT, 0)
	RETURN Int(xRECT.GetByte(12) +;
				BitLShift(xRECT.GetByte(13), 8) +;
				BitLShift(xRECT.GetByte(14), 16) +;
				BitLShift(xRECT.GetByte(15), 24))
// EOP: GetWorkareaX

FUNCTION UnitPixelsX(oForm) // Returns pixels/form unit (Turpin) on X axis
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine pixels/horizontal form unit (Turpin).
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: ClientPixWidth
// Called by...: Any
// Usage.......: nPixelsXUnit = UnitPixelsX(oFORM)
// Returns.....: Numeric -- Number of pixels/horizontal form unit.
// Parameters..: oForm = Object reference to the form.
//---------------------------------------------------------------------

	RETURN ClientPixWidth(oForm.hWND) / oForm.Width
// EOP: UnitPixelsX

FUNCTION UnitPixelsY(oForm) // Returns pixels/form unit (Turpin) on Y axis
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine pixels/vertical form unit (Turpin).
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: ClientPixHeight
// Called by...: Any
// Usage.......: nPixelsYUnit = UnitPixelsY(oFORM)
// Returns.....: Numeric -- Number of pixels/vertical form unit.
// Parameters..: oForm = Object reference to the form.
//---------------------------------------------------------------------

	RETURN ClientPixHeight(oForm.hWND) / oForm.Height
// EOP: UnitPixelsY

FUNCTION ClientPixWidth(nHandle) // Returns client area width in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine the width in pixels of the client area of a window.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 05/15/1995 -- Changed <> to #.
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: None.
//             : WinAPI: GetClientRect
// Called by...: Any
// Usage.......: nClientWidthPix = ClientPixWidth(f.hWND)
// Returns.....: Numeric
// Parameters..: nHandle = Valid Windows handle (hWND) to the window.
//---------------------------------------------------------------------
	LOCAL xRECT

	If Type("GetClientRect") # "FP"
		extern CLOGICAL GetClientRect(CHANDLE, CPTR) User32
	EndIf
	xRECT = Space(8)
	GetClientRect(nHandle, xRECT)
	RETURN Int(xRECT.GetByte(8) +;
				BitLShift(xRECT.GetByte(9), 8) +;
				BitLShift(xRECT.GetByte(10), 16) +;
				BitLShift(xRECT.GetByte(11), 24))
// EOP: ClientPixWidth

FUNCTION ClientPixHeight(nHandle) // Returns client area height in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine the height in pixels of the client area of a window.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 05/15/1995 -- Changed <> to #.
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: None.
//             : WinAPI: GetClientRect
// Called by...: Any
// Usage.......: nClientHeightPix = ClientPixHeight(f.hWND)
// Returns.....: Numeric
// Parameters..: nHandle = Valid Windows handle (hWND) to the window.
//---------------------------------------------------------------------
	LOCAL xRECT

	If Type("GetClientRect") # "FP"
		extern CLOGICAL GetClientRect(CHANDLE, CPTR) User32
	EndIf
	xRECT = Space(8)
	GetClientRect(nHandle, xRECT)
	RETURN Int(xRECT.GetByte(12) +;
				BitLShift(xRECT.GetByte(13), 8) +;
				BitLShift(xRECT.GetByte(14), 16) +;
				BitLShift(xRECT.GetByte(15), 24))
// EOP: ClientPixHeight

FUNCTION WindowPixWidth(nHandle) // Returns full window width in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine the width in pixels of a window.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 05/15/1995 -- Changed <> to #.
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: None.
//             : WinAPI: GetWindowRect
// Called by...: Any
// Usage.......: nWidthPix = WindowPixWidth(f.hWND)
// Returns.....: Numeric
// Parameters..: h = Valid Windows handle (hWND) to the window.
//---------------------------------------------------------------------
	LOCAL xRECT, nPixWidth

	If Type("GetWindowRect") # "FP"
		extern CLOGICAL GetWindowRect(CHANDLE, CPTR) User32
	EndIf
	xRECT = Space(8)
	GetWindowRect(nHandle, xRECT)
	nPixWidth = Int(xRECT.GetByte(8) +;
							BitLShift(xRECT.GetByte(9), 8) +;
							BitLShift(xRECT.GetByte(10), 16) +;
							BitLShift(xRECT.GetByte(11), 24)) -;
						Int(xRECT.GetByte(0) +;
								BitLShift(xRECT.GetByte(1), 8) +;
								BitLShift(xRECT.GetByte(2), 16) +;
								BitLShift(xRECT.GetByte(3), 24)) + 1
	RETURN IIf(nPixWidth < 0, nPixWidth + 65536, nPixWidth)
// EOP: WindowPixWidth

FUNCTION WindowPixHeight(nHandle) // Returns full window height in pixels
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Determine the height in pixels of a window.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 05/15/1995 -- Changed <> to #.
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: None.
//             : WinAPI: GetWindowRect
// Called by...: Any
// Usage.......: nHeightPix = WindowPixHeight(f.hWND)
// Returns.....: Numeric
// Parameters..: nHandle = Valid Windows handle (hWND) to the window.
//---------------------------------------------------------------------
	LOCAL xRECT, nPixWidth

	If Type("GetWindowRect") # "FP"
		extern CLOGICAL GetWindowRect(CHANDLE, CPTR) User32
	EndIf
	xRECT = Space(8)
	GetWindowRect(nHandle, xRECT)
	nPixWidth = Int(xRECT.GetByte(12) +;
							BitLShift(xRECT.GetByte(13), 8) +;
							BitLShift(xRECT.GetByte(14), 16) +;
							BitLShift(xRECT.GetByte(15), 24)) -;
						Int(xRECT.GetByte(4) +;
								BitLShift(xRECT.GetByte(5), 8) +;
								BitLShift(xRECT.GetByte(6), 16) +;
								BitLShift(xRECT.GetByte(7), 24)) + 1
	RETURN IIf(nPixWidth < 0, nPixWidth + 65536, nPixWidth)
// EOP: WindowPixHeight

PROCEDURE CenterForm(oForm) // Centers MDI form to VdB client area or non-MDI form to screen area
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 04/11/1998
// Notes.......: This function is provided for compatibility with the
//             : 16-bit version of Center.PRG for migrating apps from
//             : 16-bit VdB to 32-bit VdB. You should use the form's
//             : AutoCenter property instead.
//             : Center a MDI form to the VdB client area or non-MDI form
//             : to the screen area. This is best accomplished by setting
//             : the f.Top = -300 in the form's header and then call
//             : CenterForm(f) in the form's OnOpen to center the form.
//             : Provides quickest method to center a form for 2 reasons:
//             : 1. Uses the WinAPI call SetWindowPos which only draws the
//             :  form once on the screen.
//             : 2. Centers the form directly based on the form's size
//             :  relative to the VdB client area size (MDI), or workarea
//             :  size (non-MDI).
//             : This procedure can be used with modeless and modal forms.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 03/17/1995 -- Original
//             : 05/15/1995 -- Changed <> to #.
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 09/29/1996 -- Change to center child window in parent client.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
//             : 04/11/1998 -- Change to use PositionForm
// Calls.......: PositionForm
// Called by...: Any
// Usage.......: In the WFM header:
//             : f = New MyForm()  // Substitute the ClassName of your form.
//             : f.Top = -300      // Override class Top to open off screen.
//             : f.OnOpen = {; CenterForm(this)}  // Center form in OnOpen.
//             : * f.MDI = False   // if opened with ReadModal().
//             : f.Open()          // Or f.ReadModal().
//             : RETURN            // Prevent opening 2 copies of form.
//             : * End Header...
// Returns.....: None
// Parameters..: oForm = Object reference to the form object.
//---------------------------------------------------------------------

	PositionForm(oForm, 4)
// EOP: CenterForm

PROCEDURE PositionForm(oForm, nWhere, nXOff, nYOff)  // To 1 of 9 positions
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 04/11/1998
// Notes.......: This procedure will position an MDI form within.the VdB
//             : FrameWindow (MDIClient window) boundary or a non-MDI form
//             : within the usable screen boundary.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 04/11/1998 -- Original
//             : 04/16/1998 -- Added nXOff and nYOff offset parameters.
// Calls.......: UnitPixelsY, UnitPixelsY, MDIClientUnitWidth,
//             : MDIClientUnitHeight, GetWorkareaX, GetWorkareaY,
//             : WindowPixWidth, WindowPixHeight, MoveForm
// Called by...: Any
// Usage.......: In the WFM header:
//             : f = New MyForm()  // Substitute the ClassName of your form.
//             : f.Top = -300      // Override class Top to open off screen.
//             : f.OnOpen = {; PositionForm(this, 5)}  // Position center-right
//             : * f.MDI = False   // if opened with ReadModal().
//             : f.Open()          // Or f.ReadModal().
//             : RETURN            // Prevent opening 2 copies of form.
//             : * End Header...
// Returns.....: None
// Parameters..: oForm = Object reference to a form.
//             : nWhere = 0  // Top-Left
//             : nWhere = 1  // Top-Center
//             : nWhere = 2  // Top-Right
//             : nWhere = 3  // Center-Left
//             : nWhere = 4  // Center
//             : nWhere = 5  // Center-Right
//             : nWhere = 6  // Bottom-Left
//             : nWhere = 7  // Bottom-Center
//             : nWhere = 8  // Bottom-Right
//             : nXOff = Horizontal offset in form units from nWhere.
//             : nYOff = Vertical offset in form units from nWhere.
//             : nXOff and nYOff can be negative.
//             : To position a form 1 form unit above the bottom boundary
//             : and 1 form unit left of the right boundary:
//             : PositionForm(f, 8, -1, -1)
//---------------------------------------------------------------------
	LOCAL nUPX, nUPY, nParentWidth, nParentHeight, nTop, nLeft

	nUPX = UnitPixelsX(oForm)
	nUPY = UnitPixelsY(oForm)
	If oForm.MDI
		nParentWidth = MDIClientUnitWidth(oForm)
		nParentHeight = MDIClientUnitHeight(oForm)
	Else
		nParentWidth = (GetWorkareaX() + 2) / nUPX
		nParentHeight = (GetWorkareaY() + 2) / nUPY
	EndIf
	If (nWhere < 3) Or (nWhere > 8)
		nTop = 0
	ElseIf nWhere < 6
		nTop = (nParentHeight - (WindowPixHeight(oForm.hWND) / nUPY)) / 2
	Else
		nTop = nParentHeight - (WindowPixHeight(oForm.hWND) / nUPY)
	EndIf
	If Mod(nWhere, 3) = 2
		nLeft = nParentWidth - (WindowPixWidth(oForm.hWND) / nUPX)
	ElseIf Mod(nWhere, 3) = 1
		nLeft = (nParentWidth - (WindowPixWidth(oForm.hWND) / nUPX)) / 2
	Else
		nLeft = 0
	EndIf
	MoveForm(oForm,;
				nLeft + IIf(Empty(nXOff), 0, nXOff),;
				nTop + IIf(Empty(nYOff), 0, nYOff))
// EOP: PositionForm

PROCEDURE MoveForm(oForm, nLeft, nTop) // Moves a form to a new position.
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: MoveForm is faster than VdB FORM.Move() method. It also
//             : prevents the form from being drawn twice.
//             : This procedure can be used with any VdB form (modeless,
//             : modal, MDI or SDI).
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 05/15/1995 -- Original
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: UnitPixelsX, UnitPixelsY
//             : WinAPI: SetWindowPos, UpdateWindow
// Called by...: Any
// Usage.......: To provide a faster form open, in the WFM header:
//             : f = New MyForm    // Substitute the ClassName of your form.
//             : f.Top = -300      // Override class Top to open off screen.
//             : f.OnOpen = {; MoveForm(this, 4, 2)}  // Move form in OnOpen.
//             : * f.MDI = False     // if opened with ReadModal().
//             : f.Open()          // Or f.ReadModal().
//             : RETURN            // Prevent opening 2 copies of form.
//             : * End Header...
//             : Or, to move the form from anywhere in your program:
//             : MoveForm(FORM, 4, 2)
// Returns.....: None
// Parameters..: oForm = Object reference to the form object to be moved.
//             : nLeft = New Left position for the form in form units (Turpins).
//             : nTop = New Top position for the form in form units (Turpins).
//---------------------------------------------------------------------
	LOCAL nNewLeft, nNewTop

	nNewLeft = Int(nLeft * UnitPixelsX(oForm))
	nNewTop = Int(nTop * UnitPixelsY(oForm))
	If Type("SetWindowPos") # "FP"
		extern CLOGICAL SetWindowPos(CHANDLE, CHANDLE, CINT, CINT, CINT, CINT, CINT) User32
	EndIf
	If Type("UpdateWindow") # "FP"
		extern CLOGICAL UpdateWindow(CHANDLE) User32
	EndIf
	#ifndef	SWP_NOSIZE
		#define	SWP_NOSIZE	1	// Don't resize the form
	#endif
	SetWindowPos(oForm.hWND, 0, nNewLeft, nNewTop, 0, 0, SWP_NOSIZE)
	UpdateWindow(oForm.hWND)
// EOP: MoveForm

FUNCTION MDIClientUnitWidth(oForm) // Returns MDIClient width in form units
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: This procedure calculates and returns the width of the
//             : VdB MDIClient window in form units (Turpins).
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 05/16/1995 -- Original
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: ClientPixWidth, UnitPixelsX
//             : WinAPI: GetParent
// Called by...: Any
// Usage.......: nWidth = MDIClientUnitWidth(FORM)
// Returns.....: Numeric - The width of the MDIClient window in form units.
// Parameters..: oForm = Object reference to a form object.
//---------------------------------------------------------------------

	If Type("GetParent") # "FP"
		extern CHANDLE GetParent(CHANDLE) User32
	EndIf
	// Calculate the size of the MDIClient window in form units
	RETURN ClientPixWidth(GetParent(oForm.hWND)) / UnitPixelsX(oForm)
// EOP: MDIClientUnitWidth

FUNCTION MDIClientUnitHeight(oForm) // Returns MDIClient height in form units
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: This procedure calculates and returns the height of the
//             : VdB MDIClient window in form units (Turpins).
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 05/16/1995 -- Original
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: ClientPixHeight, UnitPixelsY
//             : WinAPI: GetParent
// Called by...: Any
// Usage.......: nHeight = MDIClientUnitHeight(FORM)
// Returns.....: Numeric - The height of the MDIClient window in form units.
// Parameters..: oForm = Object reference to a form.
//---------------------------------------------------------------------

	If Type("GetParent") # "FP"
		extern CHANDLE GetParent(CHANDLE) User32
	EndIf
	// Calculate the size of the MDIClient window in form units
	RETURN ClientPixHeight(GetParent(oForm.hWND)) / UnitPixelsY(oForm)
// EOP: MDIClientUnitHeight

PROCEDURE CenterAppClientControl(oCtlRef) // Center a VdB UI object in VdB client window
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Will center a VdB UI object such as an Image or Text object
//             : in the VdB MDIClient window.
//             : 1. The MDIClient window does not support colored pens for
//             :    drawing text. It only supports black on white. So, no
//             :    matter what you set Text.ColorNormal to, it will display
//             :    in black and white when the Text object is moved to the
//             :    VdB MDIClient window.
//             : 2. You cannot move or resize the controls once they are
//             :    assigned to the MDIClient window. If you attempt to move
//             :    or resize the controls, they will disappear.
//             : 3. You can code event handlers and methods for controls
//             :    placed in the MDIClient window although this has not been
//             :    thoroughly tested at this time.
//             : 4. The form on which the controls were created must remain
//             :    open. If you close the form on which the controls
//             :    originated after moving the controls to the MDIClient
//             :    window, the controls will disappear from the MDIClient
//             :    window.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 05/15/1995 -- Original
//             : 05/16/1995 -- Additional comments added.
//             : 06/21/1995 -- Updated comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: WindowPixWidth, WindowPixHeight
//             : WinAPI: GetParent, SetParent, SetWindowPos, UpdateWindow
// Called by...: Any
// Usage.......: SetAppClientControl(FORM.Image1)
//             : SetAppClientControl(FORM.Text1)
// Returns.....: None
// Parameters..: oCtlRef = Object reference of control to be placed in
//             : MDIClient window.
//---------------------------------------------------------------------
	LOCAL oForm, hMDIClient, nNewLeft, nNewTop

	If Type("GetParent") # "FP"
		extern CHANDLE GetParent(CHANDLE) User32
	EndIf
	oForm = New Form()
	oForm.Visible = False
	oForm.Open()
	hMDIClient = GetParent(oForm.hWND)
	oForm.Close()
	// Compute the position to center the image horizontally
	nNewLeft = ( (WindowPixWidth(hMDIClient) - WindowPixWidth(oCtlRef.hWND)) / 2 )
	// Compute the position to center the image vertically
	nNewTop = ( (WindowPixHeight(hMDIClient) - WindowPixHeight(oCtlRef.hWND)) / 2 )
	If nNewLeft < 0 Or nNewLeft > WindowPixWidth(hMDIClient)
		nNewLeft = 0 // Image is wider than MDIClient window
	EndIf
	If nNewTop < 0 Or nNewTop > WindowPixHeight(hMDIClient)
		nNewTop = 0 // Image is taller than MDIClient window
	EndIf
	If Type("SetWindowPos") # "FP"
		extern CLOGICAL SetWindowPos(CHANDLE, CHANDLE, CINT, CINT, CINT, CINT, CINT) User32
	EndIf
	If Type("SetParent") # "FP"
		extern CHANDLE SetParent(CHANDLE, CHANDLE) User32
	EndIf
	If Type("UpdateWindow") # "FP"
		extern CLOGICAL UpdateWindow(CHANDLE) User32
	EndIf
	#ifndef	SWP_NOSIZE
		#define	SWP_NOSIZE	1	// Don't resize the image
	#endif
	SetWindowPos(oCtlRef.hWND, 0, nNewLeft, nNewTop, 0, 0, SWP_NOSIZE)
	SetParent(oCtlRef.hWND, hMDIClient) // Make the MDIClient window the parent of object
	UpdateWindow(oCtlRef.hWND)          // Paint quickly
// EOP: CenterAppClientControl

PROCEDURE SetAppClientControl(oCtlRef) // Place a VdB UI object in VdB client window
//---------------------------------------------------------------------
// Programmer..: Jim Sare (jimsare@ameritech.net)
// Date........: 12/14/1997
// Notes.......: Will place a VdB UI object such as an Image or Text object
//             : in the VdB MDIClient window.
//             : The positioning of the controls in the MDIClient window
//             : is relative to the upper-left corner of the MDIClient
//             : window. So, when using SetParent to move the control(s)
//             : from the form to the MDIClient window, the controls(s)
//             : will be positioned in the MDIClient window in the same
//             : position relative to the top-left corner of the MDIClient
//             : window as they were in on the original form at the time
//             : of the SetParent call.
//             : See CenterAppClientControl for additional notes on using
//             : this procedure.
// Written for.: Visual dBASE (32-bit versions)
// Rev. History: 05/15/1995 -- Original
//             : 05/16/1995 -- Additional comments added.
//             : 06/21/1995 -- Updated comments.
//             : 07/31/1995 -- Fixed 'Calls' in comments.
//             : 08/03/1995 -- Changed parameters to locals.
//             : 12/14/1997 -- Initial adaptation for 32-bit.
// Calls.......: None
//             : WinAPI: GetParent, SetParent, UpdateWindow
// Called by...: Any
// Usage.......: SetAppClientControl(FORM.Image1)
//             : SetAppClientControl(FORM.Text1)
// Returns.....: None
// Parameters..: oCtlRef = Object reference of control to be placed in
//             : MDIClient window.
//---------------------------------------------------------------------
	LOCAL oForm, hMDIClient

	If Type("GetParent") # "FP"
		extern CHANDLE GetParent(CHANDLE) User32
	EndIf
	oForm = New Form()
	oForm.Visible = False
	oForm.Open()
	hMDIClient = GetParent(oForm.hWND)
	oForm.Close()
	If Type("SetParent") # "FP"
		extern CHANDLE SetParent(CHANDLE, CHANDLE) User32
	EndIf
	SetParent(oCtlRef.hWND, hMDIClient) // Make the MDIClient window the parent of oCtlRef
	If Type("UpdateWindow") # "FP"
		extern CLOGICAL UpdateWindow(CHANDLE) User32
	EndIf
	UpdateWindow(oCtlRef.hWND)          // Paint quickly
// EOP: SetAppClientControl
// EOF: Center32.PRG

