/*
   GetDirs.PRG
   This is a routine to procure a listing of all directories 
   on a specified drive, starting at the root. This creates
   an array, and returns a pointer to the array, which could be used
   to load a listbox or other ...

   Posted by Keith Chuvala on the Borland Visual dBASE newsgroups
   Minor modifications by Ken Mayer for readability ...

   Added use of Set Exact so folk using language drivers
   such as BLLT1DA0 and similar won't have problems ...

   Usage:    set procedure to getdirs additive
             aDirs = GetDirs( "C:" )
             // number of elements:
             ? aDirs.size
*/
   parameter cDrive
   local cDir, cPath, lDone, nDcnt, cAttributes, cFilename, nEntries
   private cOldExact
   private a,d  // So we can type() elements....

   // save setting for Exact
   cOldExact = set("exact")
   // change:
   set exact on

   if type("cDrive") # "C"
      cDrive = ""
   endif
   
   cPath = ""       // Holds the "current" directory
   lDone = false      // Set to true when nothing left to do
   a = new array(0) // Array to hold each directory's entries
   d = new array(0) // Directory list we're building
   nDcnt = 0        // Pointer into d array for traversing tree
  
   do while not lDone
      cDir = cdrive + iif(isblank(cPath),"\","") + cPath + ;
                      iif(isblank(cPath),"","\") + "*.*"
      a.dir(cDir,"D")              // Check docs on ADIR() for details on structure!
      nEntries = int(a.size / 5)   

      if nEntries > 0              // Not an empty directory
         for i = 1 to nEntries
            cAttributes = a[i,5]   // Check docs on ADIR() for details
            cFilename   = a[i,1]
            if (right(cAttributes,1) == "D") and ;
               (left(cFilename,1) # ".")      // Ignore . and .. entries
               d.add(cPath + "\" + cFilename) // Add this one to the list
            endif
         next
      endif
      nDcnt = nDcnt + 1            // Get the next entry in D...
      if nDcnt <= d.size
         cPath = d[nDcnt]          // ...and process it next
         a.resize(1,5)             // Reset the a array, otherwise the size might
      else                         //  be wrong for the next directory we read
         lDone = true               // No more entries, so go bye-bye.
      endif
   enddo

   // reset exact:
   set exact &cOldExact.

return d                           // Return a reference to the list array

