/*
RowEdit.cfm
For VdB7
Author: Peter Rorlick
Last modified: June 29, 1998   (Removed RecnoQuery requirement)

RowEditForm custom form class
It has OK and Cancel buttons, and handles all logic concerning
edit/append, save/abandon of a row or record.
Handles a rowset or workarea.

By Peter Rorlick

IMPORTANT: You MUST copy the following lines to the header of your form:
 private modeArg   // 2nd parameter (first is bModal).
 modeArg = argvector(2)
 private IndexKeyValue     // 3rd parameter
 IndexKeyValue = argvector(3)
 private oGrid
 oGrid = argvector(4)  // 4th param (optional) to ensure refresh of the grid.
 ****************************************************************************

cMode must be "New" or "Edit" (it can also contain an ampersand,
which will be stripped out when cMode is copied to Form.cMode).

Examples of how you can call your form:
 Do MyForm.wfm with false, 'Edit', form.rowset.fields['ClientNo'].value, form.zGrid1
 Do MyForm.wfm with true, This.Text, form.rowset.fields['ClientNo'].value   // "This" would be a pushbutton.
 Do MyForm.wfm with true, 'New', form.query1.rowset.fields['ClientNo'].value

If you want to over-write the Init method of this form, your Init method
should start with the line:
  RowEditForm::Init()  // call super's init
or the equivalent:
  super::init()

Something like this might be a good idea:
  Proc Init
    RowEditForm::Init()  // call super's init
    form.text = form.cMode + ' ' + trim(form.rowset.fields['LastName'].value)
  return

After calling this form, you probably will want to
RefreshBrowse(oBrowseObject) or oGridObject.Refresh()
Note that passing the 4th parameter oGrid takes care of this automatically.

"Replicate" mode is automatically supported in this version only
when operating on a rowset.  If you're using workareas, you
have to code for Replicate functionality in your form's OnOpen event.

*/
*****************************************************************

class RowEditForm of ZFORM custom from "zControl.cc"
   set procedure to xtrabtns.cc additive
   set procedure to zControl.cc additive
   with (this)
      onMouseMove = {;class::FlashButton_BlinkOff()}
      canClose = class::FORM_CANCLOSE
      height = 7
      left = 5.71
      top = 0.32
      width = 61.2857
   endwith


   this.OKPOPBUTTON1 = new OkPopBUTTON(this)
   with (this.OKPOPBUTTON1)
      onClick = class::OKPOPBUTTON1_ONCLICK
      left = 47.7143
      top = 1.8636
      pageno = 0
      value = false
      enabled = true    // was .F.
   endwith

/*
   this.OKANDLOGOUTBUTTON = new OkPopBUTTON(this)
   with (this.OKANDLOGOUTBUTTON)
      onClick = class::OKANDLOGOUTBUTTON_ONCLICK
      left = 47.7143
      top = 0.2727
      text = "Ok + logout"
      upBitmap = ""
      pageno = 0
      value = false
   endwith
*/

   this.CANCELPOPBUTTON1 = new CancelPopBUTTON(this)
   with (this.CANCELPOPBUTTON1)
      onClick = class::CANCELPOPBUTTON1_ONCLICK
      onOpen = {; if Type('Form.lAbort')=='L' and Form.lAbort ; Form.Close() ; Endif}
      left = 47.7143
      top = 3.4545
      pageno = 0
      value = false
   endwith


   Proc Init
*   if type('modeArg')#'C'   // If programmer runs the form from the FD by hitting F2...
*     * For design-time testing:
*     modeArg = 'Edit'
*     nRecordNum = 1
*   endif
   Form.CloseBy = 0     // For Form_CanClose - initialized here to an arbitrary non-char value.
   Form.cMode = modearg
   set proc to funclib.prg addi
   Form.cMode = Proper(StrTran(Form.cMode,chr(38),''))  // Strip out any ampersands.
   if not Form.cMode == 'New' and Type('Form.Rowset')=='O'
     * move the row pointer to the appropriate row:
     // Form.Rowset.GotoRecord(nRecordNum)
     * RecNoQuery's rowset.recno() method causes instability on some tables!
     * So, instead of passing the record# to this form, we'll expect to
     * pass an index key expression that identifies the record to be edited.
     Form.Rowset.FindKey(IndexKeyValue)   // Programmer must set the rowset.indexName in the RowEditForm.
   endif
   Form.bookmark = IIF( Type('Form.Rowset')=='O', Form.Rowset.Bookmark(), Recno())
   if type('oGrid') == 'O'
     form.GridBookmark = oGrid.datalink.bookmark()
   endif
   if Form.cMode == 'Edit' or Form.cMode == 'Replicate'
     If ( Type('Form.Rowset')=='O' and Form.Rowset.EndOfSet ) or ( Type('Form.Rowset')#'O' and eof() )
       TMsgBox('At end of file. Please highlight a record first.')
       Form.lAbort = true  // Form will close in the OnOpen of the cancelButton.
       return
     Endif
   endif
   Form.Text = Form.cMode + ' record'
   if Form.cMode == 'Replicate' and Type('Form.Rowset')=='O'
     * Create an array to save the field values of the current row:
     local aValues, numFields, i, saveBeforeGetValue
     numFields = Form.Rowset.Fields.Size
     aValues = new Array()
     for i = 1 to numFields
       * saveBeforeGetValue = Form.Rowset.Fields[i].BeforeGetValue  // save it
       * Form.Rowset.Fields[i].BeforeGetValue := null   // Disable field-morphing (if any) temporarily.
       * aValues.Add(Form.Rowset.Fields[i].Value)
       * Form.Rowset.Fields[i].BeforeGetValue := saveBeforeGetValue   // Restore it
       aValues.Add( RealValue(Form.Rowset.Fields[i]) )
     next
   endif
   if Form.cMode == 'New' or Form.cMode == 'Replicate'
     If Type('Form.Rowset')=='O'
       Form.Rowset.BeginAppend()   // OODML
     Else
       Form.BeginAppend()  // Assume we're dealing with a workarea.
     Endif
   else
     If Type('Form.Rowset')=='O'
       ** The following line sometimes crashes with "Expecting object".
       ** But it's superfluous anyway, because the rowset automatically
       ** goes into Edit mode when the user types into a datalinked field.
       * Form.Rowset.BeginEdit()   // OODML
     Endif
   Endif
   if Form.cMode == 'Replicate' and Type('Form.Rowset')=='O'
     * Set the field values of the new row, according to what was saved in the array:
     local saveCanChange
     for i = 1 to numFields
       if Form.Rowset.Fields[i].type # 'AUTOINC'  // Autoincrement fields are read-only.
         saveCanChange = Form.Rowset.Fields[i].canChange
         Form.Rowset.Fields[i].canChange := null   // Disable field-morphing temporarily.
         Form.Rowset.Fields[i].Value = aValues[i]
         Form.Rowset.Fields[i].canChange = saveCanChange  // Restore field-morphing, if any.
       endif
     next
     form.rowset.refreshControls()
     Form.Repaint()  // Form.Rowset.RefreshControls() would suffice also.
   Endif
   If Type('Form.Rowset')=='O' and Form.cMode == 'Edit'
     * Enable the OK button only as soon as the user edits something:
     Form.Rowset.OnEdit = {;_app.activeform.okPopbutton1.Enabled = true }
   Endif
   return

/*
   function OKANDLOGOUTBUTTON_onClick    // only need this for web application.
      class::OkPopButton1_OnClick()
      close forms
      return
*/

   function OKPOPBUTTON1_onClick
     if Form.CheckForErrors()
       return
     endif
     try
       if Type('Form.Rowset')=='O'
         Form.Rowset.Save()   // OODML
         * FLUSH and UpdateModStamp() are performed automatically - see ZQuery class.
       else
         Form.SaveRecord()    // Assume we're dealing with a workarea.
         UpdateModStamp()
       endif
     catch ( DBException e )
       if 'KEY VIOLATION' $ upper(e.Message)   // Can happen with DISTINCT indexes.
         tmsgbox('Another record exists having the same key value.','Error',48)
         return false  // false flags the error to form_canClose()
       else  // some other error occurred...
         tmsgbox(e.Message,'Error while saving',48)
       endif
     endtry
     if Type('Form.Rowset')=='O'
       Form.Rowset.RefreshControls()  // Should this go BEORE OnSaveRow() ?
       form.OnSaveRow()         
     endif
     if type('oGrid') == 'O'
       if Form.cMode == 'New' or Form.cMode == 'Replicate'   //  if new record was just added.....
         * Make sure that oGrid's rowset is pointing to that new record:
         private r
         r = oGrid.datalink  // the rowset
         * Extract the table name from the query's .SQL string:
         local t, cTableName, nAt, c
         nAt = at(' from ',lower(r.parent.sql))
         c = subs(r.parent.sql,nAt+6)
         c = remChars( '"'+"'", c )  // remove quotes
         nAt = at('.',c)
         if nAt > 0   // if it contains a dot...
           cTableName = left(c,nAt-1)   // take everything that precedes the dot
         else
           nAt = at(' ',c)  // look for a SPACE
           if nAt == 0
             cTableName = c
           else
             cTableName = left(c,nAt-1)  // take everything that precedes the SPACE.
           endif
         endif
         t = new tableDef()
         t.tableName = cTableName
         t.load()
         r.GotoRecord(t.RecordCount)  // RecNoQuery class method
/*
         r.SaveIndexName = r.indexName   // save it
         r.indexName = ''   // This puts the rows in RANDOM order (not natural order!!!)
         r.parent.requery()  // This puts the rows in NATURAL order.
         ** But Requery() forces re-execution of query.execute() and kills the 
         ** master/detail links.  So we'll avoid Requery, and use
         ** r.GotoRecord(t.RecordCount) instead.
         r.last()     // sets pointer to the record that was just added.
         r.IndexName = r.SaveIndexName   // restore it
*/
         if type('r.SaveRangeValue') # 'U'   // Programmers must create this custom property when it's needed.
           local b
           b = r.bookmark()
           * Restore the current Range, which got cleared when we switched the IndexKey:
           r.setRange(r.SaveRangeValue)
           r.goto(b)
         endif
         * oGrid.parent.setFocus()   // unnecessary
         * oGrid.SetFocus()          // unnecessary
         r.refreshControls()
       endif
       oGrid.Refresh()
       if Form.cMode == 'Edit'
         * Ensure that the grid will be properly refreshed even if the user has just 
         * Edited a record and changed a key field:
         // oGrid.datalink.Goto(Form.GridBookmark)    // Does NOT work!!
       endif
     endif
     Form.CloseBy = 'OKButton'   // For Form_CanClose
     lChangesSaved = true  // This line is for PickerForm::EditorDialog()
     Form.Close()
   return true


   function CANCELPOPBUTTON1_onClick
     If Type('Form.Rowset')=='O'
       Form.Rowset.Abandon()    // OODML
       Form.Rowset.Goto(Form.bookmark)
     Else
       Form.AbandonRecord()     // Assume we're dealing with a workarea.
       Goto(Form.bookmark)
     Endif
     Form.CloseBy = 'CancelButton'   // For Form_CanClose
     Form.Close()
   Return


   Function Form_CanClose
   local lChangesMade
   lChangesMade = Form.IsRecordChanged()
   IF Type('Form.CloseBy')#'C' AND Form.CheckForErrors()
     Return false   // An error was found, so do not permit the form to be closed.
   Endif
   * VDB7 BUG: Form.IsRecordChanged() is not reliable!!!! Sometimes it incorrectly returns FALSE.
   If Type('Form.CloseBy')#'C' AND IIF( Type('Form.Rowset')=='O', Form.Rowset.Modified, lChangesMade )
     * Changes were made and an attempt is being made to close the
     * form, but the user did not click on the OK or Cancel buttons.
     ?? CHR(7)
     IF tMsgBox('Save changes?', form.text+': Save?', 52) = 6   // If yes
       if class::OKPopButton1_OnClick() = false
         * A Key violation or other DBE error occurred.
         return false
       endif
     Else
       class::CancelPopButton1_OnClick()
     Endif
   Endif
   Form.CloseBy = NULL
   Return true


   Function CheckForErrors
     * Check for errors and reutrn true if any are found.
     * You can override this in forms that are subclassed from this one.
     * Example:
     *   If empty(Form.NameField.Value)
     *     tmsgbox('Error: You must supply a name','Error',48)
     *     return true
     *   Endif
     *   ....etc....
     *   return false
     *
   return false

   Proc OnSaveRow
     * You should override this method in your own forms. This is the
     * place to do anything special that needs to be done AFTER the row
     * has been saved but BEFORE the form is closed.
   return

endclass
