   /*
   --------------------------------------------------------
   Filename.....: SaveAppFrameWin.prg
   ClassName....: None.
   Programmer...: Rick Miller.
   Date.........: June 15, 2000.
   Purpose......: Functions to position _app.framewin.
   Notes........: See below.
   Written for..: VdBASE 7.x and newer.
   Rev. History.: February 10, 2005 -  renamed all external functions.
   Dependencies.: None.
   Calls........: None.
   Called by....: Any.
   Usage........: set procedure to SaveAppFrameWin.prg additive
   Example......: See below.
   --------------------------------------------------------
   Example:
   --------------------------------------------------------
   At the start of an Application.
      bVar  =  SetFramePosion()

   In the canClose of the Application.
      bVar  =  SaveFramePosition()
   --------------------------------------------------------
   Functions:
   --------------------------------------------------------
   SaveFramePosition()
      The current _app.framewin position and
         _app.framewin.windowstate are written
         to an ini file.
      The ini file written to can then be used
         when starting an App to Position the
         _app.framewin where it was when last closed.
      This can be used with a Network App to set
         up the framewin for each user by writing to
         each user's own ini file.
   SetFramePosition()
      When the App starts (is at the Bottom of this file)
         the _app.framewin is positioned based upon
         the information in the ini file.
   --------------------------------------------------------
   Features:
   --------------------------------------------------------
      The framewin position values are only written
         when the framewin is in Normal mode.
      The Windowstate is always written.
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
   This is a cobbled together function from various routines
      found in the dUFLP. Hope it is of use to others.
   It was put together for the purpose of running a Network App
      and having the _app.framewin for each User re-appear where it
      was last closed by each user.
         For that:
         substitute "_app.inifile" with  "_app.userini"
   Do what You wish with the code.
   Had been working on this when Ken Mayer brought it up in
      the newsgroups.
   Seems much of the code came from stuff written by: 
      Jim Sare
      Keith Chuvala 
      Bowen Moursund
      Kenneth Chan
      Romain Strieff
   */
//---------------------------------------------------------------//
//---------------------------------------------------------------//
function GetFrameSize(nHwnd)
   Local nBot, nRight, xRect
   xRect =  space(8)
   if type("argVector(1)") # "N"
      nHwnd := int(_app.framewin.hwnd)
   endif
   if not type("RMM_GetWindowRect") == "FP"
      extern CLOGICAL RMM_GetWindowRect(CHANDLE, CPTR) ;
         user32 from "GetWindowRect"
   endif
   RMM_GetWindowRect(nHwnd, xRect)
   _app.framewin.left   =  GetRectInteger(xRect, 0)
   _app.framewin.top    =  GetRectInteger(xRect, 4)
   nRight   =  GetRectInteger(xRect, 8)
   _app.framewin.right  =  int(nRight)
   nBot     =  GetRectInteger(xRect, 12)
   _app.framewin.bot    =  int(nBot)
   _app.framewin.width  =  nRight - _app.framewin.left
   _app.framewin.height =  nBot - _app.framewin.top
return(xRect)
//---------------------------------------------------------------//
//---------------------------------------------------------------//
function GetRectInteger(xRect, nByte)
   nRet  =  ;
      xRect.GetByte(nByte) +;
      BitLShift(xRect.GetByte(nByte + 1), 8) +;
      BitLShift(xRect.GetByte(nByte + 2), 16) +;
      BitLShift(xRect.GetByte(nByte + 3), 24)
   if BitSet(nRet, 31)
      nRet  := - BitNot(nRet)
   endif
return int(nRet)
//---------------------------------------------------------------//
//---------------------------------------------------------------//
function GetIniValue(cSection, cEntry)
   /*
   Return a Value from the  _app.inifile
   Need to Pass the Section Name, Entry Name
   Returns the  Entry Name Value in Char
   */
   if not type("RMM_GetIniString") == "FP"
      extern CULONG RMM_GetIniString( ;
         CSTRING, CSTRING, CSTRING, CSTRING, CULONG, CSTRING) ;
         kernel32 from "GetPrivateProfileStringA"
   endif
   Local cBuff, cFile, nLen
   cBuff =  space(256)
   if type("_app.inifile") # "C"
      cFile =  "Test.ini"
   else
      cFile =  "" + _app.inifile
   endif
   nLen  =  RMM_GetIniString( ;
      cSection, cEntry, "", cBuff, 256, cFile )
return left(cBuff, nLen)
//---------------------------------------------------------------//
//---------------------------------------------------------------//
function SaveFramePosition
   if _app.framewin.windowstate == 0
      GetFrameSize()
      SaveIniValue("Desktop", "FrameLeft",    ltrim(str(_app.framewin.left)))
      SaveIniValue("Desktop", "FrameTop",     ltrim(str(_app.framewin.top)))
      SaveIniValue("Desktop", "FrameRight",   ltrim(str(_app.framewin.right)))
      SaveIniValue("Desktop", "FrameBottom",  ltrim(str(_app.framewin.bot)))
   endif
   SaveIniValue("Desktop", "FrameMaximized",  ltrim(str(_app.framewin.windowstate)))
return true
//---------------------------------------------------------------//
//---------------------------------------------------------------//
function SaveIniValue( cSection, cEntry, cValue)
   /*
   Write a Value to the  _app.inifile
   Need to Pass the Section Name, Entry Name, Entry Name Value
   */
   if not type("RMM_WriteIniString") == "FP"
      extern CLOGICAL RMM_WriteIniString( ;
         CSTRING, CSTRING, CSTRING, CSTRING) ;
         kernel32 from "WritePrivateProfileStringA"
   endif
   if type("_app.inifile") # "C"
      cFile =  "Test.ini"  // the name of exe + ".ini"
   else
      cFile =  "" + _app.inifile
   endif
return RMM_WriteIniString(cSection, cEntry, cValue, cFile)
//---------------------------------------------------------------//
//---------------------------------------------------------------//
function SetFramePosition
   Local bRet, nMax, nLeft, nTop, nRight, nBot, nHeight, nWidth
   bRet     =  false
   nMax     =  val(GetIniValue("Desktop", "FrameMaximized"))
   nLeft    =  val(GetIniValue("Desktop", "FrameLeft"))
   nTop     =  val(GetIniValue("Desktop", "FrameTop"))
   nRight   =  val(GetIniValue("Desktop", "FrameRight"))
   nBot     =  val(GetIniValue("Desktop", "FrameBottom"))
   nLeft    := iif(nLeft  >= 0,            nLeft, 0)
   nTop     := iif(nTop   >= 0,            nTop,  0)
   nRight   := iif(nRight > nLeft + 100,   nRight, 600)
   nBot     := iif(nBot   > nTop + 100,    nBot,   400)
   nWidth   =  nRight  - nLeft
   nHeight  =  nBot    - nTop
   _app.framewin.left   =  nLeft
   _app.framewin.right  =  nRight
   _app.framewin.top    =  nTop
   _app.framewin.bot    =  nBot
   _app.framewin.width  =  nWidth
   _app.framewin.height =  nHeight
   if nMax == 0  // Normal
      _app.framewin.windowState := 0
      if not type("RMM_MoveWindow") == "FP"
         extern CLOGICAL RMM_MoveWindow( ;
            CHANDLE, CINT, CINT, CINT, CINT, CLOGICAL) ;
            user32 from "MoveWindow"
      endif
      RMM_MoveWindow( ;
         _app.framewin.hwnd, nLeft, nTop, nWidth, nHeight, true)
      bRet  := true
   endif
return iif(bRet, true, false)
//---------------------------------------------------------------//
//          end of file.
//---------------------------------------------------------------//
