* ZControl.CC
* Custom controls of all kinds
* Author: Peter Rorlick
* Other contributing programmers: Saveli Kossenko

***********************************************************************
*CLASS zPushButton(FormObj) OF PopBUTTON(FormObj) Custom
*EndClass
**********************************************************************
CLASS zEntryField(FormObj) OF ENTRYFIELD(FormObj) Custom
  with (this)
    onGotFocus = {; this.keyboard( "{Home}" )}
    BorderStyle = 7
    ColorHighlight = 'N/RG+'  // yellow background when in focus.
    ColorNormal = 'N/W+'
    Height = 1     // was 0.8
    SelectAll = false   // So that selected strings are immediately visible even if they have lots of trailing spaces.
    // this.FontSize = 10  // was 8
    OnRightDblClick = CLASS::InsertDate
  endwith

  this.BaseClassName = 'ENTRYFIELD'

  ** Popup menu has been disabled because it messes up the menu prompts of the _app.framewin's menu.
  * this.OnOpen  = {; zEditor::DefineEditorPopup() }
  * this.OnRightMouseDown = zEditor::EditControl_OnRightMouseDown
  this.OnRightMouseDown = {;Do EdTools.wfm with (this), (form) }


  Proc InsertDate
    If Type('This.selectAll')#'U'  // Editors don't have a SelectAll property.
      local cSaveSelectAll
      cSaveSelectAll = this.SelectAll
      this.SelectAll = false  // We don't want to over-write the current contents with the selected date.
    Endif
    do InsertDt.wfm with true
    If Type('This.selectAll')#'U'   // If this is an EntryField
      This.SelectAll = cSaveSelectAll
      *This.Keyboard("{End}" + ' ' + _app.StringToInsert)  // Add to the end of what's already in the entryfield.
      Keyboard "{End}" + ' ' + _app.StringToInsert  // Add to the end of what's already in the entryfield.
    Else // we're in an Editor
      * This.Keyboard("{Ctrl+End}" + "{End}" + _app.StringToInsert)   // Insert at the cursor position.
      * This.Keyboard(_app.StringToInsert)   // Insert at the cursor position.
      Keyboard _app.StringToInsert   // Insert at the cursor position.
    Endif
  return

  FUNCTION Key(nChar, nPosition)
    * Trap Alt-D to invoke date-inserter tool.
    If nChar == 4  // Alt-D
      zEntryField::InsertDate()
      return false
    EndIf
  RETURN nChar

ENDCLASS
**********************************************************************
// This next one is good for names of people or places...
// Thanks to Edward Sanders for the idea.
//
CLASS ProperEntryField( oCont ) of zEntryfield( oCont ) CUSTOM

  this.speedTip := 'Press Alt + = to turn auto-casing on/off'
  this.lAutoCasing = true

  FUNCTION key( nChar, nPosition )
    if nextkey() # 0
      // Wait until there are no keystrokes pending, and...
      RETURN ( nChar # 255 )     // ...ignore keystroke if special key
    endif
    if nChar == 255
      if this.lAutoCasing
        this.value := ProperName(this.value)
      endif
      return false
    else
      if nChar = 61   // Alt-=
        this.lAutoCasing = not this.lAutoCasing
        if this.lAutoCasing
          this.ColorHighlight = 'N/RG+'  // Back to normal: yellow background when in focus.
        else
          this.ColorHighlight = 'N/CYAN'  // Cyan background indicates that AutoCasing is OFF.
        endif
      endif

      if (nChar > 31 and nChar < 255) or ;
          nChar = 7 or nChar = 8 or nChar = 127  // trap printable chars or Del or Backspace.
        keyboard chr(255)
      endif
    endif
  RETURN nChar

  func onLostFocus
    this.lAutoCasing = true
    this.ColorHighlight = 'N/RG+'  // Back to normal: yellow background when in focus.
  return

ENDCLASS
**********************************************************************
* Zoomable editor:
CLASS zEditor(FormObj) OF EDITOR(FormObj) Custom
  this.BaseClassName = 'EDITOR'
  this.Key = zEntryField::Key  // Trap Alt-D for date insertion

* this.OnRightMouseDown = CLASS::EditControl_OnRightMouseDown
* this.OnOpen      = {; CLASS::DefineZoomButton() ; CLASS::DefineEditorPopup() }

  *this.OnOpen      = {; CLASS::DefineZoomButton() }   // popup disabled because of conflict with _app.frameWin's menu
  this.BorderStyle = 7
  this.ScrollBar   = 1
*  this.OnGotFocus  = CLASS::YellowBackground
*  this.OnLostFocus = CLASS::WhiteBackground
  this.OnRightDblClick = zEntryField::InsertDate
  this.onLeftDblClick := class::Print
  this.ColorNormal='N/W+'
*  this.PopUpEnable := false   // Prevent glitch with modal forms?  No such glitch under 7.5?
  this.cuaTab := false
//  this.FontSize := 8
//  this.MarginVertical := 1

  this.lZoomModal = true   // Can be overridden. This flag is passed as (bModal)
                      // when ZoomEdit.wfm is called. If you want to enable
                      // HTML formatting, set this to FALSE in your form.

   Procedure Init
     * Instead of the OnOpen event, which is slow cause it occurs AFTER the form is open.
     CLASS::DefineZoomButton()
   return


  *-------------------------------------
  PROC YellowBackground   // onGotFocus
//  this.ColorNormal = 'N/RG+'
//  this.borderstyle=6

  *--------------------------------------
  PROC WhiteBackground   // onLostFocus
//  this.ColorNormal = 'N/W+'
//  this.borderstyle=7

  func onGotFocus
    _app.oCurrentEditor = this
    * on key label Ctrl+Z _app.oCurrentEditor.undo()  // doesn't work!

    on key label Ctrl+Tab zEditor::MoveFocusFromEditor()
    // This allows the user to press Ctrl-Tab to move to the next control.
    //  (We could do same for Ctrl-Shift-Tab maybe)
    // But this is not "standard" Windows behaviour, so users won't know 
    // about this unless we show them or unless they read our help files.

    on key label Tab keyb spac(3)  // Workaround to bug where editor.cuaTab=false misbehaves.

  return

  func onLostFocus
    on key label Tab  // Recover from workaround to cuaTab=false bug.
    on key label Ctrl+Tab
    * on key label Ctrl+Z
    _app.oCurrentEditor = null
  return

  func MoveFocusFromEditor      // Invoked by Ctrl-Tab while in a zEditor
    * Find the next item in the tabbing order of the form:
    private o
    o = _app.oCurrentEditor.before
    do while not type('o.setFocus') $ 'CB/FP' or ( type('o.Visible')=='L' and o.Visible==false )
      o = o.before
    enddo
    o.SetFocus()   // ...and give it focus.
  return


  function Print
    local f
    f = new form()
    f.top := -300
    f.text := ' '
    f.e = new editor(f)
    f.e.anchor := 6  // Container
    f.e.value := this.value
    f.onOpen = { ; form.print() ; form.close() }
    f.open()
  return


  *--------------------------------------
   PROC DefineEditorPopup

*   If true
*     return   // Disabled because this proc screws up _app.FrameWin's menu prompts.
*   Endif

   IF TYPE('this.parent.edit_pop') = 'O'
     RETU   && An editor popup is already defined for this form.
   ENDIF
   this.parent.edit_pop =new popup(this.parent) // of this.parent
   *
   *  Each popup menu item uses the Active property
   *  set by the EditControl's OnRightMouseDown event
   *  to manipulate the control that was clicked.
   *
   this.parent.edit_pop.cut = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.cut)
     Text = "Cu&t"
     shortCut = "Ctrl+X"
     OnClick = {;this.parent.active.cut()}
   endWith

   this.parent.edit_pop.copy = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.copy)
     Text = "&Copy"
     shortCut = "Ctrl+C"
     OnClick = {;this.parent.active.copy()}
   endWith

   this.parent.edit_pop.paste = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.paste)
     Text = "&Paste"
     shortCut = "Ctrl+V"
     OnClick = {;this.parent.active.paste()}
   endWith

   this.parent.edit_pop.DeleteAll = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.DeleteAll)
     Text = "&Delete all"
     OnClick = {;this.parent.active.value=''}
   endWith

   this.parent.edit_pop.undo = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.undo)
     Text = "&Undo"
     OnClick = {;this.parent.active.Undo()}
   endWith

   this.parent.edit_pop.selectAll = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.selectAll)
     Text = "&Select all"
     OnClick = {;tMsgBox('To select all text, press Ctrl-Home then Ctrl-Shift-End')}
     *OnClick {;this.parent.active.Keyboard( "{Ctrl+Home}" + REPL("{Shift+DownArrow}",99) )}
   endWith
/*
   this.parent.edit_pop.InsTodayN = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.InsTodayN)
     Text = "Insert today's date (&numeric)"
     OnClick = {;d=DtoS(DATE()); this.parent.active.Keyboard( SUBS(d,3,2)+'/'+SUBS(d,5,2)+'/'+SUBS(d,7,2) ) }
   endWith

   this.parent.edit_pop.InsTodayE = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.InsTodayE)
     Text = "Insert today's date (&English)"
     OnClick = {;this.parent.active.Keyboard( cDow(DATE())+', '+cMonth(Date())+' '+LTRIM(STR(DAY(DATE())))+', '+STR(YEAR(DATE()),4) ) }
   endWith

   this.parent.edit_pop.InsDateN = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.InsDateN)
     Text = "Insert a date (numeric)"
     OnClick = {; _app.SelectedDate=.F.; DO Calendar.WFM; IF .NOT.EMPTY(_app.SelectedDate); d=DtoS(_app.SelectedDate); this.parent.active.Keyboard( SUBS(d,3,2)+'/'+SUBS(d,5,2)+'/'+SUBS(d,7,2) ); ENDIF }
   endWith

   this.parent.edit_pop.InsDateE = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.InsDateE)
     Text = "Insert a date (English)"
     OnClick = {; _app.SelectedDate=.F.; DO Calendar.WFM; IF .NOT.EMPTY(_app.SelectedDate); d=_app.SelectedDate; this.parent.active.Keyboard( cDow(d)+', '+cMonth(d)+' '+LTRIM(STR(DAY(d)))+', '+STR(YEAR(d),4) ); ENDIF }
   endWith
*/
   this.parent.edit_pop.InsDate = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.InsDate)
     Text = "Insert a date"
     Shortcut = "Ctrl+D"
     OnClick = {; zEntryField::InsertDate()}
   endWith

   this.parent.edit_pop.Print = new MENU(this.parent.edit_pop)
   with (this.parent.edit_pop.Print)
     Text = "Pr&int"
     OnClick = {;PrintThisForm(This.Parent.Parent)}
   endWith


   *------------------------------------------------------
   Procedure EditControl_OnRightMouseDown(flags, col, row)
*   if true
*     return   // Disabled because this proc screws up _app.FrameWin's menu prompts.
*   endif
   *
   *  Set the popup position, taking into account the
   *  position of the control itself on the form
   *
   this.parent.edit_pop.top  = row + this.top
   this.parent.edit_pop.left = col + this.left
   *
   *  Give the popup menu a reference to the
   *  active control.
   *
   this.parent.edit_pop.active = this
   this.parent.edit_pop.open()


   Proc DefineZoomButton

      PRIV c

      oParentOfEditor = this.Parent
      c ='oParentOfEditor.Z' + TRIM(This.Name) + ' = new PUSHBUTTON(oParentOfEditor)'
      &c

      c = 'oParentOfEditor.Z' + TRIM(This.Name)
      this.oZoomButton = &c //For new disable/enable methods

      With (&c)
        Top        = This.Top+0.15
        Left       = This.Left+This.Width-(2.5+2.9)// This.Left+This.Width-(2.3+2.7)
        Height     = 0.9//0.7
        Width      = 2.9//2.3
        Text       = ""
        Group      = .T.
        * OnClick    = class::ZoomEdit
        OnClick    = {; this.oOwner.ZoomEdit() }
        onRightMouseDown = {; this.oOwner.ExportToTxtFile( this.oOwner.value ) }
        PageNo     = This.PageNo
        Enabled    = This.Enabled
        Visible    = This.Visible
        UpBitmap   = "RESOURCE #2001"
        TabStop    = .F.
        SpeedBar   = .T.
        SpeedTip   = "Left-click to zoom in.  Right-click to save as TXT file."
      EndWith

      priv d
      d = &c    // d is now the zoom button
      d.oOwner = this   // reference to the editor that "owns" the zoom button.
      d.OnFormResize = {; this.Move(this.oOwner.Left+This.oOwner.Width-(2.5+2.9), this.oOwner.Top+0.15) }
      * ; this.Visible=(this.oOwner.Height>this.height)  // Make button disappear if there is not sufficient room to display it.

      * c='oParentOfEditor.' + TRIM(This.Name)
      * d.oOwner=&c
   return

   *-------------------------------
   Proc ZoomEdit             // The OnClick event of the Zoom button.
     oEditor = This.oOwner
     _app.oEditor = oEditor
     do ZoomEdit.WFM with this.oOwner.lZoomModal
   return


//   function key( nChar, nPosition, bShift, bControl )
//      /*
//         This code is by Gary White -- it gets around
//         an odd bug in the editor.
//
//         If the rowset's autoEdit property is set to "false",
//         and:
//            You edit the memo and then save or abandon,
//            you are allowed to continue editing the memo.
//
//            If the memo has NOT been edited, then everything
//            works as expected. This call to the KEY event
//            appears to solve this problem.
//      */
//      if form.rowset.state == 1
//         return 0
//      endif

   func NumberOfLines
      local nSaveLineNo, n
      nSaveLineNo = this.lineNo
      this.lineNo := 9999999          // This trick no longer works in 7.5
      n = this.lineNo
      this.lineNo := nSaveLineNo
   return n

   func ExportToTxtFile( cText )
     if argCount()==0
       cText = this.value  // Assume THIS is an Editor control.
     endif
     local nH, cFileName
     if left(cText,1) == '<'  // If it looks like it begins with an HTML tag...
       cFileName = PutFile(translate('Save contents of editor to an HTML file'),null,'.HTM')
     else
       cFileName = PutFile(translate('Save contents of editor to text file'),null,'.txt')
     endif
     if not empty( cFileName )
       if file(cFileName) and tmsgbox(cFilename+' already exists. Overwrite?','Overwrite?',52)#6
         return  // User said no.
       endif
       nH = fcreate(cFileName)
       fputs(nH,cText)
       fclose(nH)
     endif
  return

ENDCLASS
**********************************************************************
CLASS zSpinBox(FormObj) OF SpinBox(FormObj) Custom
  with( this )
    borderStyle = 7
    ColorNormal='N/W+'
    ColorHighlight = 'N/RG+'
    Height = 1     // was 0.8
    FontSize = 10  // was 8
    function = 'J'  // Right-align
  endWith
  this.BaseClassName = 'SPINBOX'
ENDCLASS
**********************************************************************
CLASS zTabBox(FormObj) OF TabBox(FormObj) Custom
  this.BaseClassName = 'TABBOX'

  function tabCount
    private m_macro
    local a
    m_macro = substr( this.datasource, at(" ",this.datasource)+1 )
    a = &m_macro.
  return a.size 

ENDCLASS
**********************************************************************
CLASS zComboBox(FormObj) OF ComboBox(FormObj) Custom
  with (this)
    ColorHighlight = 'N/RG+'
    ColorNormal='N/W+'
    OnGotFocus  = CLASS::YellowBackground
    OnLostFocus = CLASS::WhiteBackground
    height = 1.05
    style = 2  // dropdown list
  endwith
  this.BaseClassName = 'COMBOBOX'

  PROC YellowBackground
    * this.ColorNormal = 'N/0x80ffff'    // On some (Win95 ?) machines, this color causes an ugly partly-gray background.
    this.ColorNormal = 'N/RG+'
    // Note that autodrop will be pretty ugly.

  PROC WhiteBackground
   this.ColorNormal = 'N/W+'

ENDCLASS
**********************************************************************
CLASS BooleanComboBox(FormObj) OF ZComboBox(FormObj) Custom

/*
For comboBoxes that are datalinked to boolean DBF fields.

For user-friendliness we'll offer the user 3 choices:
  Unspecified
  Yes
  No
and do field-morphing to translate these strings to/from true/false/null.

On forms where we provide the user with a default value, we could remove the
"Unspecified" option.

In fact, it would be easy to create a BooleanCombobox class that performs the
field-morphing automatically in its OnOpen, if the datalink is not empty.
*/

  this.width++  // Make it wider, to accommodate "Unspecified"

  Proc Init
    this.datasource := 'Array {"' + translate("Unspecified") +'","'+ translate("Yes") +'","'+ translate("No") +'"}'
    if not empty(this.dataLink)
      this.datalink.beforeGetValue = {|| translate(iif(this.value==null,'Unspecified',iif(this.value,'Yes','No'))) }
      this.dataLink.CanChange = {|newValue|; if localType(newValue)=='C' ; this.value:=iif(newvalue==Translate('Unspecified'),null,(newValue==Translate('Yes'))) ; endif ; return false }
    endif
  return

ENDCLASS

*********************************************************************
// by Xiaoming Zhong, April 26, 2000
// The following three controls are created for boe.wfm

CLASS DegreeSpinBox(FormObj) OF ZSpinBox(FormObj) Custom

  this.width = 5.6667
  this.rangeMax = 99
  this.rangeMin = 0
  this.rangeRequired = true 
  Proc Init
    if not empty(this.dataLink)
      this.dataLink.parent.parent.refreshControls()
    endif
  return

ENDCLASS

CLASS VarusComboBox(FormObj) of ZComboBox(FormObj) Custom

  this.width = 8.3
  Proc Init
    this.datasource := 'Array {"Valgus","Varus"}'
  return
ENDCLASS

CLASS IncreaseComboBox(FormObj) of ZComboBox(FormObj) Custom

  this.width = 13

  Proc Init
    this.datasource :='Array {"Increased", "Decreased", "WNL"}'
  return

ENDCLASS

**********************************************************************
CLASS zSeeker( fArg ) of Seeker( fArg ) CUSTOM

// By Peter Rorlick, Feb 1998
//  Does a SEEK on Backspace or Del.
//  Adds SeekPrefix functionality.
//  Yellow background when in focus.
//  Seeks in the work area described by the .Alias property (optional).
//  Beeps when seek fails.
//  Goes back to original row (not bottom) when seek fails.

  protect lNear, lExact, lFound, lRefresh, nMaxPos, seekRowset
  protect dataLink  // This control uses the active index to seek; DO NOT datalink!

  this.BaseClassName = 'ENTRYFIELD'
  this.BorderStyle = 7
  this.ColorHighlight = 'N/RG+'  // yellow background when in focus.
  this.ColorNormal='N/W+'
  this.Height = 1
  * The above 4 lines were copied from zEntryField class.

  This.SeekPrefix  = ""   // Modify this when a SET KEY TO ... is active.

  this.Alias = NULL       // This can be changed in a form's OnOpen.
                          // Does not apply when dealing with a RowSet.

  this.NumericSeek = false  // Set this to TRUE if the index key values are numeric.

  function onKey( nChar, nPosition, bShift, bControl )
    if nextkey() == 0 and nChar # 13   // Wait until there are no keystrokes pending
      if not empty( this.value )  //  and ( this.lFound or nPosition <= this.nMaxPos )
        this.lFound = this.seek()
        if not this.lFound
          beep()
          keyb "{End}" + "{Backspace}"   // "Reject" the last char and force another seek.
        endif
      else
        // The seeker is empty, so go top:
        if form.rowset == null       // if xbase:
          go top
        else
          this.seekRowset.first()
        endif
      endif
      if this.tabAfter > 0 and nPosition > this.tabAfter
        // Stuff tab if position is greater than tabAfter
        this.keyboard( chr(9) )
      endif
      * this.nMaxPos = max( nPosition, min( this.nMaxPos, len( trim( this.Value ))) )
    endif
  return

  function key( nChar, nPosition )
    do case
     CASE nChar==7 or nChar==127  // trap Del (that key fails to trigger the onKey event)
       // So we'll make sure it fires - we want a "seek" when the user presses Delete.
       this.keyboard( chr(255) )
    *CASE nChar = 229  // Ctrl-Backspace was pressed
       *  return false
     CASE nChar = 230  // up arrow was pressed
       if form.rowset == null   // if xbase:
         this.xSkip(-1)
       else   // oodml:
         this.oSkip(-1)
       endif
       return false
     CASE nChar = 231  // down arrow was pressed
       if form.rowset == null   // if xbase:
         this.xSkip(1)
       else   // oodml:
         this.oSkip(1)
       endif
       return false
    endcase
  return true

  function NormalizedValue
    if this.NumericSeek
      // Index keys are numeric...
      RETURN val( this.value )
    else
      // Index keys are upper-case strings...
      RETURN This.SeekPrefix + upper( trim( this.Value ))
    endif
  return

  function xSeek
    local nSaveRecn, lRet, nSaveWorkArea
    nSaveWorkArea = WorkArea()
    if this.Alias # NULL
      select (this.Alias)
    endif
    nSaveRecn = RECN()
    lRet = seek( this.normalizedValue() )
    if eof()               // Always move record pointer to real record
      * goto bottom  // This was Seeker's original behavior.
      if nSaveRecn <= RECC()
        GO nSaveRecn
      else
        goto bottom
      endif
    endif
    select (nSaveWorkArea)
  RETURN lRet

  function oSeek
    local lRet, bBookMark
    bBookMark = this.seekRowset.BookMark()
    lRet = this.seekRowset.findKey( this.normalizedValue() )   // Original seeker had FindKeyNearest()
    if not lRet
      * this.seekRowset.next( -1 )   // This was Seeker's original behavior.
      this.SeekRowset.goto(bBookMark)   // back to the row we were at before the seek.
      if this.seekRowset.endOfSet
        this.seekRowset.first()
      endif
    endif
  RETURN lRet

  function xSkip( n )
    local nSaveRecn, nSaveWorkArea
    nSaveWorkArea = WorkArea()
    if this.Alias # NULL
      select (this.Alias)
    endif
    nSaveRecn = RECN()
    skip n
    if eof() or bof()             // Always try to move record pointer to real record
      beep()
      if nSaveRecn <= RECC()
        GO nSaveRecn
      else
        go top
      endif
    endif
    select (nSaveWorkArea)
  return

  function oSkip( n )    // n is 1 or -1
    local bBookMark
    bBookMark = this.seekRowset.BookMark()
    if not this.seekRowset.next( n )
      beep()
      this.SeekRowset.goto(bBookMark)   // back to the row we were at before the attempted skip
      if this.seekRowset.endOfSet
        this.seekRowset.first()
      endif
    endif
  return

  function onGotFocus
    seeker::onGotFocus()
    * on key label Ctrl+Backspace keyb chr(229)
    on key label UpArrow        keyb chr(230)
    on key label DnArrow        keyb chr(231)
  return

  function onLostFocus
    seeker::onLostFocus()
    * on key label Ctrl+Backspace
    on key label UpArrow
    on key label DnArrow
  return

ENDCLASS
**************************************************************************
class zBrowse(parentObj) of BROWSE(parentObj) custom
  with (this)
    onGotFocus = class::ONGOTFOCUS
    onLostFocus = class::ONLOSTFOCUS
    onLeftMouseDown = class::ONLEFTMOUSEDOWN
    colorNormal = "WindowText/0x80ffff"
    colorHighLight = "WindowText/0x80ffff"
    height = 7.5
    width = 37.7143
    append = false
    modify = false
    delete = false
    OnLeftDblClick  = class::Print
    OnRightDblClick = class::Print
  endwith
  this.BaseClassName = 'BROWSE'
  GetGridFont()
  this.fontName = _app.GridFontName
  this.fontSize = _app.GridFontSize

  Function onLeftMouseDown
    this.SetFocus()
  Return

  function onGotFocus
    this.BorderStyle = 6   // DropShadow (to indicate visually that this control has focus).
  return

  function onLostFocus
    this.BorderStyle = 0   // Default (client style)
  return

  function Print
    if msgBox('Print ALL the rows and columns of this grid?','Print?',32+4)==6
      local f, nRecn
      f = new form()
      f.top := -300
      f.text := ' '
      f.b = new browse(f)
      f.b.anchor := 6  // Container
      f.b.fontName := this.fontName
      f.b.fontSize := this.fontSize
      f.b.alias := this.alias
      f.b.fields := this.fields
      freeze((form))
      nRecn = recn(this.alias)  // Save record pointer
      go top in (this.alias)   // Must go top, else it will print only from current to EOF.
      f.onOpen = { ; form.print() ; form.close() }
      f.open()
      go nRecn in (this.alias)
      unfreeze()
      form.setFocus()
    endif
  return

endclass
****************************************************************************
class FastLabel(parentObj) of Entryfield(parentObj) custom

* This is a faster version of a text control - actually an entryfield without
* a border that cannot receive focus.  It's like a dumbed-down Text control,
* and it instantiates faster than a Text object.

   with (this)
     border = false
     TabStop = false
     When = {||false}
     * onGotFocus = {; this.keyboard("{Tab}")}  // Needed?
     colorNormal = 'WindowText/btnFace'
     metric = 0        // Chars
     mousePointer = 1  // Arrow
     borderStyle = 3   // None
   endwith
   this.text = ''  // Create custom property here, so that constructor section of WFM can update it.
   this.BaseClassName = 'ENTRYFIELD'
   this.Value = 'FastLabel'

   proc init
     if this.value == 'FastLabel'  // If the .Value was not changed by the constructor section of the WFM...
       if not empty(this.text)
         // This condition should be true only for Text objects that have
         // been converted to FastLabels.
         this.value := this.text
       endif
     endif
   return

* Problem: This RUNS fine, but the Form Designer will kill any Text properties
* that were set by the programmer back when zText was based on a Text control.
* In other words, it is problematic only when working with forms that were
* originally designed with text controls that were not FastLabel controls.

* Note: Classify.prg should be modified so that converts .Text properties of
*       zText objects to .Value properties.

* For fastest instantiation of a form that has a LOT of text objects, design
* it with texts, and once the design is finalized, take a bitmap of the running
* form, then redesign the form, remove all the text objects, add a new image
* (send to back) that shows the bitmap.  The result is that all the text controls
* have been replaced by a single image object, and it will open much faster.
* Keep a copy of the WFM that contains the text objects, in case you ever want
* to redesign the form.

endclass
****************************************************************************
class zTextLabel(parentObj) of TextLabel(parentObj) custom
   with( this )
     OnRightMouseDown = {; do configTx.wfm with true, this.form, this}
     text = 'zTextLabel'
     height = 0.8
     width = 20
   endWith
   this.BaseClassName = 'TEXTLABEL'
   this.init = zText::Init
endclass
****************************************************************************
class zBaseText(parentObj) of Text(parentObj) custom

* Use this class instead of zText only when you don't want the user to be
* able to change the property values at runtime.  For example, when the .Text
* property must be able to change dynamically and programmatically during
* program execution.

   with (this)
      left = 5
      top = 2
*     OnOpen{; this.text = translate(this.text) }
   endwith
   this.BaseClassName = 'TEXT'

endclass
****************************************************************************
class zText(parentObj) of zBaseText(parentObj) custom

*  this.Visible = false  // slow-down and GDI hit only occurs when Visible, says Bowen.
   with( this )
     OnRightMouseDown = {; do configTx.wfm with true, this.form, this}
     text = 'ZText'
     wrap = false
     height = 0.8
     width = 20
   endWith
   this.BaseClassName = 'TEXT'

   Proc Init // SetUserProperties
   if fileInCurrentDirectory('ConfigTx.dbf')
     * Look up Form.ClassName+'.'+this.name, and set properties if found.
     if type('_app.qConfigTx')=='U'
       * Create and activate a query on ConfigTx.dbf (once only):
       _app.qConfigTx = new Query()
       _app.qConfigTx.SQL = 'select * from ConfigTx'
       _app.qConfigTx.Active = true
       _app.qConfigTx.Rowset.IndexName = 'ObjName'
     endif
     *
     ** To get the full object name for objects that reside on
     * containers/notebooks, we parse the ownership tree
     * (courtesy of George Burt):
     local mObject
     private mTemp1
     Mobject  = this.name
     Mtemp1 = this.parent
     do while type("Mtemp1.parent") # 'U'
       Mobject = Mtemp1.name+"."+Mobject
       Mtemp1 = Mtemp1.parent
     enddo
     * Mobject = "form."+Mobject
     * Mobject = lower(Mobject)
     *
     if _app.qConfigTx.Rowset.FindKey(Form.ClassName+'.'+mObject)
       this.Text        = Trim(_app.qConfigTx.Rowset.Fields['pText'].Value)
       this.Top         = _app.qConfigTx.Rowset.Fields['pTop'].Value
       this.Left        = _app.qConfigTx.Rowset.Fields['pLeft'].Value
       this.Width       = _app.qConfigTx.Rowset.Fields['pWidth'].Value
       this.Height      = _app.qConfigTx.Rowset.Fields['pHeight'].Value
       this.ColorNormal = Trim(_app.qConfigTx.Rowset.Fields['pColorNorm'].Value)
       this.Wrap        = _app.qConfigTx.Rowset.Fields['pWrap'].Value
       this.Function    = Trim(_app.qConfigTx.Rowset.Fields['pFunction'].Value)
       this.Picture     = Trim(_app.qConfigTx.Rowset.Fields['pPicture'].Value)
       this.FontName    = Trim(_app.qConfigTx.Rowset.Fields['pFontName'].Value)
       this.FontSize    = _app.qConfigTx.Rowset.Fields['pFontSize'].Value
       this.FontBold    = _app.qConfigTx.Rowset.Fields['pFontBold'].Value
       this.FontItalic  = _app.qConfigTx.Rowset.Fields['pFontItal'].Value
       this.Border      = _app.qConfigTx.Rowset.Fields['pBorder'].Value
       if _app.qConfigTx.Rowset.Fields['pBorder'].Value
         this.BorderStyle = 7   // Client style
       endif
     endif
   endif    //if fileInCurrentDirectory('ConfigTx.dbf')
   *

*   if this.visible = .f.   // default is False
*     * Create a FastLabel object on-the-fly, whose .Value=this.text
*     * and top/left/height/width = same as (this) :
*     private c, oLabel
*     c = 'this.parent.FL'+this.name+' = new FastLabel(this.parent)'
*     &c
*     c = 'this.parent.FL'+this.name
*     oLabel = &c
*     oLabel.Value  = this.text
*     oLabel.top    = this.top
*     oLabel.left   = this.left
*     oLabel.height = this.height
*     oLabel.width  = this.width
*     * Copy other properties as well?
*   endif

   * With this scheme,  zText objects will (by default) generate FastTexts,
   * improving speed of form instantiation.  If the programmer wants
   * to have an actual zText object on the form at runtime (allowing user
   * to right-click to set user-defined properties, or to take advantage
   * of HTML abilities of the Text object), he just has to set the object's
   *.visible property to TRUE in the form designer.

   return

/*
Design for user-configurable text objects
-----------------------------------------

ConfigTx.dbf
 ObjName    C 50  (e.g. "INJURYFORM.TEXT1" )   (Index on this field)
 pText      C
 pTop       N
 pLeft      N
 pWidth     N
 pHeight    N
 pColorNorm C
 pWrap      L
 pFunction  C
 pPicture   C
 pBorder    L  (If true then set BorderStyle=7)
 pFontName  C
 pFontSize  N
 pFontBold  L
 pFontItal  L
*/
endclass
****************************************************************************
class zCheckBox(parentObj) of CHECKBOX(parentObj) custom
   with (this)
      left = 5
      top = 2
   endwith
   this.BaseClassName = 'CHECKBOX'
endclass
****************************************************************************
class zListbox(parentObj) of LISTBOX(parentObj) custom
  with (this)
    height = 4
    left = 5
    top = 2
    width = 28
    *ColorHighlight = 'N/RG+'
    ColorNormal='N/W+'
    OnGotFocus  = CLASS::YellowBackground
    OnLostFocus = CLASS::WhiteBackground
  endwith
  this.BaseClassName = 'LISTBOX'

  PROC YellowBackground
    this.ColorNormal = 'N/0x80ffff'    // was: 'N/RG+'
  return

  PROC WhiteBackground
   this.ColorNormal = 'N/W+'
  return

  function SelectAll  // (nHwnd,lOn)
    //       if type("SendMessage") # "FP"
    //          extern CLONG SendMessage ( CHANDLE,CINT,CWORD,CLONG )  USER
    //       endif
    //       #define WM_USER htoi("400")
    //       #define LB_SETSEL (WM_USER+6)
    //       if lOn
    //         SendMessage(nHwnd,LB_SETSEL,1,-1)
    //       else
    //          SendMessage(nHwnd,LB_SETSEL,0,-1)
    //       endif
    *
    // Our kludgy method may fail if the list contains very many items...
    // It would be nice if there was a way to freeze the screen while
    // the pending chars are processed from the keyboard.  Maybe with a timer?
    * freeze()
    * on key label Alt-7 unfreeze()  // This trick doesn't work
    * keyboard repl("{Ctrl+PgUp}",99) + repl('{Shift+DnArrow}',this.count())
    * keyboard "{Ctrl+End}" + repl('{Shift+UpArrow}',this.count())   // + "{Alt+7}"   // Ctrl-End does nothing if there's less than 1 page full of items.
    keyboard repl("{Ctrl+PgDn}",99) + repl('{Shift+UpArrow}',this.count())   // + "{Alt+7}"   // Ctrl-End does nothing if there's less than 1 page full of items.
    this.setFocus()
    // We assume Multiple = true
  return

  function UnselectAll
   // keyboard "{Ctrl+Home}"
    Keyboard "{Home}" + repl("{UpArrow}", this.count())
    this.setFocus()
  return

endclass
****************************************************************************
class zRadioButton(parentObj) of RADIOBUTTON(parentObj) custom
   with (this)
      left = 5
      top = 2
   endwith
   this.BaseClassName = 'RADIOBUTTON'
endclass
****************************************************************************
class zLine(parentObj) of LINE(parentObj) custom
   this.BaseClassName = 'LINE'
endclass
****************************************************************************
class zImage(parentObj) of IMAGE(parentObj) custom
   this.BaseClassName = 'IMAGE'
endclass
****************************************************************************
class zGridColumn(parentObj) of GridColumn(parentObj) custom
   this.BaseClassName = 'GRIDCOLUMN'
endclass
****************************************************************************
class zGrid(parentObj) of GRID(parentObj) custom

   with (this)
      height = 7
      left = 4
      top = 2
      width = 41
      *onGotFocus = class::ONGOTFOCUS
      *onLostFocus = class::ONLOSTFOCUS
      *onLeftMouseDown = class::ONLEFTMOUSEDown
      *colorNormal = "WindowText/0x80ffff"
      *colorHighLight = "WindowText/0x80ffff"
      bgcolor = '0x80ffff'
      allowEditing = false
      allowAddRows = false
      *hasColumnLines = false
      hasRowLines = false
      rowSelect = true
   endwith
   this.BaseClassName = 'GRID'
   GetGridFont()
   this.cellHeight = _app.GridcellHeight

   Function onOpen
     // Try to ensure that the row has a "Selected" color when the grid opens:
     if not empty(this.datalink) and not this.datalink.endofset
       this.datalink.goto( this.datalink.bookmark() )
     endif
   return

   function init

     this.SetFont()
     
     // Hide the 1-pixel hole along the top of the grid:
     * if this.HasColumnHeadings
     *   this.Rectangle = New Rectangle(form)
     *   with(this.Rectangle)
     *     Top = this.Top
     *     Left = this.Left + 0.1
     *     Width = this.Width - 0.1
     *     Height = 0.2
     *     PageNo = this.PageNo
     *   endwith
     * endif

     if this.hasIndicator
       this.PrintButton = new sButton(this.parent)
       this.printButton.grid = this
       with (this.PrintButton)
         top = this.top + .13
         left = this.left + .29
         width = 3.0
         height = 0.82
         upBitMap = 'RESOURCE TS_PRINT'
         onClick = {; this.grid.print() }
         speedtip = 'Print all the rows and columns of the grid'
       endwith
     endif

   return

   function SetFont
     local i
     for i = 1 to this.columns.size
       this.columns[i].headingControl.fontName := _app.GridFontName
       this.columns[i].headingControl.fontSize := _app.GridFontSize
       this.columns[i].headingControl.fontBold := false
       this.columns[i].editorControl.fontName = _app.GridFontName
       this.columns[i].editorControl.fontSize = _app.GridFontSize
     next
   return

   Function onGotFocus
       if _app.ActiveForm == this.Form
         * If the onLostFocus event was triggered by the opening of
         * another form, then we do NOT want to change the border,
         * becuase doing so would give focus back to the original form,
         * which we don't want.
         if this.AllowEditing = false  // Work-around to bug in VdB that causes flashing.
           this.BorderStyle = 6   // DropShadow (to indicate visually that this control has focus).
           *this.BorderStyle = 5   // Double (to indicate visually that this control has focus).
         endif
       EndIf
   Return

   Function onLostFocus
       if _app.ActiveForm == this.Form
         * If the onLostFocus event was triggered by the opening of
         * another form, then we do NOT want to change the border,
         * becuase doing so would give focus back to the original form,
         * which we don't want.
         if this.AllowEditing = false
          this.BorderStyle = 0   // Default (client style)
         endif
       endif
   Return

   Function onLeftMouseDown
     if this.AllowEditing = false
       this.SetFocus()
     endif
   Return

   Function SetFields(aFieldList)
      * Each elemtent of the 2d array a FieldList must be in the format:
      *   {cExpression [,nWidth [,cCaption [,nEditorType]]]}
      * where cExpression can be a field name or a beforeGetValue expression for
      * a calculated field.
      * Example:
      *    aFieldList = {{'Trim(Surname)+FirstName',18,'Athlete'}, {'Birth'}, {'Sex'}}
      * This example results in 3 columns.
      *
      this.PrepareRowset(aFieldList)
      this.UpdateGridColumns(aFieldList)
      this.SetFont()
   Return

   Function PrepareRowset(aFieldList)
     // Adds calculated fields, as needed, to the rowset, in preparation for UpdateGridColumns().
     private cName, oColumn, yy, aFieldDecl, cCommand
     aFieldDecl = aFieldList
     For yy = 1 to aLen(aFieldDecl)
       If AT('(',aFieldDecl[yy][1])>0 Or AT('+',aFieldDecl[yy][1])>0
         * If the expression contains a parenthesis or a plus sign, we know it must be a calculated field.
         * Here we create that calculated field:
         oField = new Field()
         oField.FieldName := "CalcField" + yy
         This.Datalink.Fields.Add(oField)
         cCalcExpr = aFieldDecl[yy][1]
         c = "{|| " + cCalcExpr + " }"
         oField.BeforeGetValue = &c
       EndIf
     Next
     *
     * Make sure that a Dummy calc'd field exists:
     If TYPE("this.datalink.fields['DUMMY']")<>'O'
       oField = new Field()
       oField.FieldName := "Dummy"
       oField.BeforeGetValue := {||null}
       This.Datalink.Fields.Add(oField)
     EndIf
   return

   Function UpdateGridColumns(aFieldList)  // Adds and removes grid columns.
      this.datalink.NotifyControls := false

      private cName, oColumn, yy, aFieldDecl, cCommand

      aFieldDecl = aFieldList

      n = This.Columns.size
      For yy = This.Columns.size+1 to aFieldDecl.size // Adding columns if there is not enough
        This.Columns["COLUMN"+yy] = New GRIDCOLUMN(this)
      Next

      For yy = aFieldDecl.size+1 to This.Columns.size // Make extra columns "invisible"...
         cName='This.Columns["COLUMN'+yy+'"]'
         if TYPE(cName) == 'O'
           oColumn = &cName
           oColumn.Datalink := this.datalink.fields['Dummy']     // crashes here, column4
           oColumn.Width := 0  // This makes the column virtually invisible.
           oColumn.HeadingControl.Value := ''
         EndIf
      Next

      For yy = 1 to aLen(aFieldDecl)

        oColumn = This.Columns['COLUMN'+yy]

        // The next line ensures that we don't directly change the values
        // in a column from numeric to string, because the grid chokes on that.
        oColumn.DataLink := this.DataLink.Fields["Dummy"]
        oColumn.editorType := 0  // default (self-adjusts according to the field it gets datalinked to). Required for VdB 7.5.

        If AT('(',aFieldDecl[yy][1])>0 Or AT('+',aFieldDecl[yy][1])>0
          // Parens or plus sign means it's a beforeGetValue expression for a calculated field, called Field1, Field2, etc..
          oColumn.DataLink := this.DataLink.Fields['CalcField'+yy]
        Else
          // Else we assume that the expression is a field name:
          local cArraySubscript //By Serge on 1999.12.20
          cArraySubscript = aFieldDecl[yy][1]
          * oColumn.DataLink := this.DataLink.Fields[ aFieldDecl[yy][1] ]
          oColumn.DataLink := this.DataLink.Fields[ cArraySubscript ]
        EndIf

        If TYPE("aFieldDecl[yy][2]")=='N'
          oColumn.width := aFieldDecl[yy][2]
        EndIf

        If TYPE("aFieldDecl[yy][3]")=='C'
          oColumn.HeadingControl.Value := aFieldDecl[yy][3]
        EndIf

        * oColumn.HeadingControl.fontBold := false   // zgrid::SetFont() method does this.

        oColumn.editorControl.picture := repl('X',50)   // To ensure that fields that are morphed from LONG to STRING are fully displayed.

        If TYPE("aFieldDecl[yy][4]")=='N'
          oColumnEditorType := aFieldDecl[yy][4]
        EndIf

      Next

      this.datalink.NotifyControls := true
   return

   Proc Refresh( oForm )
     *
     if argCount() > 0
       freeze(oForm)
     endif
     *
     this.BorderStyle = 7  // This refreshes all the visible cells in the grid window, without scrolling!
     if type('this.datalink')=='O' and not this.datalink.EndOfSet
       local bm, oActive
       bm = this.datalink.bookmark()
       this.datalink.goto(bm)
       oActive = Form.ActiveControl
       this.setFocus()
       oActive.SetFocus()
     Endif
     * This is better than the native Refresh() method which has
     * the annoying behaviour of scrolling the display so that the
     * current record is at the top of the grid window.
     *
     if argCount() > 0
       Unfreeze()
     endif
     *
     * _app.g = this
     * do commandr.wfm with true  // for debug
     *
   return

   Function OnRightDblClick
     this.viewMemos()       // user can double-right-click to view memos.
   return

*   Function OnLeftDblClick
*     // Programmers might want to override this event, for an "open" or "edit" behaviour.
*     this.print()
*   return
// Disabled, as of September 12, 2000 (see new button, in Init method)


   function Print
     if msgBox('Print ALL the rows and columns of this grid?','Print?',32+4)==6
       try
         local f, r, bm, i
         r = this.datalink
         bm = r.bookMark()
         f = new form()
         f.top := -300
         f.text := ' '
         f.g = new grid(f)
         f.g.anchor := 6  // Container
         f.g.datalink := r
         f.g.cellHeight := this.cellHeight
         * f.g.columns := this.columns   // Doesn't work.
         for i = 1 to this.columns.size
           f.g.columns["Column"+i] = new GRIDCOLUMN(f.g)
           f.g.columns["Column"+i].dataLink   := this.columns["Column"+i].dataLink
           f.g.columns["Column"+i].editorType := this.columns["Column"+i].editorType
           f.g.columns["Column"+i].width      := this.columns["Column"+i].width
           f.g.columns["Column"+i].headingControl.value := this.columns["Column"+i].headingControl.value
           f.g.columns["Column"+i].headingControl.fontBold := this.columns["Column"+i].headingControl.fontBold
           f.g.columns["Column"+i].EditorControl.picture := this.columns["Column"+i].EditorControl.picture
         next
         f.onOpen = { ; form.print() ; form.close() }
         r.notifyControls := false
         r.first()
         f.open()
         r.goto(bm)
         r.notifyControls := true
       catch( exception e )
         r.goto(bm)
         r.notifyControls := true
         msgbox('Sorry, this grid cannot be printed.')
       endtry
     endif
   return


   function ViewMemos
     // Check the columns' datalinks (or, if columns is empty, check all the fields in the rowset).
     // If none are memo fields, nothing happens.
     // If exactly one is a memo, a dialog will pop up with a read-only editor.
     // If more than one column is a memo, I'll ask the user to choose (via a listbox)
     // which memo he wants to view.
     if this.datalink == null or this.datalink.endOfSet
       return  // no rowset, or the grid is empty.
     endif
     local i, aMemoFields
     aMemoFields = new Array()
     if this.columns.size = 0
       * No columns defined.  Therefore ALL the fields in the rowset are shown in the grid.
       * Look through these fields (this.datalink.fields) to find out which ones are memo fields:
       for i = 1 to this.datalink.Fields.size
         if this.datalink.Fields[i].Type == 'MEMO'  // if this is a memo field...
           aMemoFields.add( this.datalink.fields[i] )    // the field object
         endif
       next
     else
       for i = 1 to this.columns.size
         if this.columns[i].datalink.Type == 'MEMO'  // if this is a memo field...
           aMemoFields.add( this.columns[i].datalink )  // the field object
         endif
       next
     endif
     if aMemoFields.size == 0
       return   // No memo fields to view.
     else
       do ViewMemo.wfm with aMemoFields
     endif
   return

   func SetFocusToColumn( nTargetColumn )
     // Assumes RowSelect property is set to False.
     //
     // Here's the "new" way:
     // (Under VdB 7.5, setting a new value to the grid's frozenColumn property
     // does NOT change the CurrentColumn property.  This is different from 7.01)
     * priv c
     * c = 'this.frozenColumn = this.columns["Column' + nTargetColumn + '"].datalink.fieldName'  // the field name of the column
     * &c
     * this.setfocus()
     * this.frozenColumn = ""
     * this.SetFocus()
     *
     // Here's the old way:
     local n
     n = this.CurrentColumn - nTargetColumn
     // n is now the # of tabs or shift-tabs needed to
     // get the cursor into the desired grid column.
     if n < 0
       keyboard replicate( "{RightArrow}", abs(n) )  // Under 7.01 we used {Tab}
     elseif n > 0
       keyboard replicate( "{LeftArrow}", n )          // Under 7.01 we used {Shift+Tab}
     elseif n == 0
       keyboard "{RightArrow}{LeftArrow}"        // Need this under 7.5
     endif
     this.SetFocus()
     
   return

endclass
****************************************************************************
class UDGrid(parentObj) of zGRID(parentObj) custom // Extends zGrid

/*
Grid with user-definable layouts and sort orders.
This lets the user create/select layouts for a grid.

 Structure for a table of layout definitions:
   Layout_No    Autoinc
   Description  C 50
   IndexName    C 25
   LDefault     L
   Columns      C 100   // An array of selected ColumnDef_No's and/or field names.
   						      Example:  {'Name',2,'Sex,'Hospital_No',5}

 This record is added/edited in Layout.wfm, via Picker.wfm (e.g., via PkLayout.wfm).
 The listbox items in Layout.wfm will be the field names from the grid's rowset, PLUS a
 set of cCaption's extracted from ColumnDef fields.
 When the cCaption is not specified, we use the cFieldName element as the caption.
 When the Layout.wfm is OK'd, we convert the selected items to an array of ColumnDef_No's
 and store the stringified array into Columns field.

 Structure for a table of grid column definitions:  // Example: A record for every possible Patient grid column
   ColumnDef_No Autoinc
   ColumnDef    C 100    // {cExpression [,nWidth [,cCaption [,nEditorType]]]}
                            where cExpression can be a field name or a beforeGetValue expression for a calculated field.
                            Examples:
                             {'Name',18,'Athlete'}
                             {'Name',18}
                             {'DateOfBirth',12,'Birth date'}
                             {'edate11(this.parent["dateOfBirth"].value)',12,'Born'}

 How to use this class:
  Drop a UDGrid onto the form designer surface, and datalink it to a rowset.
  In the form's Init(), set the grid's property:
       form.udGrid1.NameOfTableContainingLayoutDefinitions = "whatever"
  and optionally this property as well:
       form.udGrid1.NameOfTableContainingSpecialGridColumnDefinitions = "Whatever"
  (You must specify the filename prefixes, instead of "Whatever".)
*/
   this.CurrentLayoutName = ''

   Function onRightMouseDown   //  On right-click, let user pick a different layout.
     _app.oTargetGrid = this
     _app.OtargetGridForm = Form
     * Important: For a calculated field whose BeforeGetValue needs a value from a
     * different rowset, you might have to refer to _app.OtargetGridForm.  Example:
     *   _app.OtargetGridForm.dataModRef1.Ref.Patients1.Rowset.Fields['LastName'].Value
     do PkLayout.wfm
   Return

   function ApplyLayoutName( cLayoutName )
     // Could be run like this :  _app.oTargetGrid.ApplyLayoutName(Form.LayoutPicker.Value)
     if this.qLayouts.rowset.findKey( padr(upper(cLayoutName),50) )
       this.ApplyLayout( this.qLayouts.rowset )
     else
        msgbox('Layout: '+ cLayoutName + ' not found.','Error',16)
     endif
   return

   Function ApplyLayout( rLayouts )  // Param is a rowset, we assume row pointer has been set.
     TRY
       private aColumns, i
       aColumns = StringToArray(rLayouts.fields['Columns'].value)   // Columns field contains an array of selected ColumnDef_No's, in string format.
       local aColumnDefs
       aColumnDefs = New Array()
       for i = 1 to aColumns.size
         if type('aColumns[i]') == 'C'
           * This one's a string, meaning it's simply a field name.
           aColumnDefs.add( { aColumns[i] } )
         else
           * Else it is a numeric reference to an Autoinc value in the qColumnDefs rowset...
           if this.qColumnDefs.rowset.FindKey( aColumns[i] )
             aColumnDefs.add( StringToArray(this.qColumnDefs.rowset.fields['ColumnDef'].value) )
           else
             msgbox('ColumnDef_No '+ aColumns[i] + ' not found.','Error',16)
           EndIf
         EndIf
       Next
       this.DataLink.indexName := trim(rLayouts.fields['IndexName'].value)
       this.DataLink.first()
       this.SetFields(aColumnDefs)
       // Parameters of SetFields method : {{cFieldName [,nWidth [,cCaption [,nEditorType]]]},...}
       this.CurrentLayoutName := trim(rLayouts.fields['Description'].value)
       if this.hScrollBar == 2   // if it's "auto"
         // Get the scrollbar to appear or disappear, as appropriate for the new layout:
         this.hScrollBar := 0
         this.hScrollBar := 2
       endif
       * this.refresh()
     CATCH ( Exception e )
       tMsgBox('Error in the grid layout definition.','Error in ApplyLayout()',16)
     EndTry
   return

   Function Init
     set proc to funclib.prg addi
     super::Init()
     *
     if type('this.NameOfTableContainingLayoutDefinitions') # 'C'
       return
     endif
     TRY
       _app.oTargetGrid = this
       _app.OtargetGridForm = Form
       *
       local cTable
       if type(  'this.NameOfTableContainingSpecialGridColumnDefinitions' ) == 'C'  // If this custom property is defined...
         cTable = this.NameOfTableContainingSpecialGridColumnDefinitions
         if not fileInCurrentDirectory( cTable + '.dbf' )
           CreateColumnsTable( cTable )  // Create empty table if it does not yet exist.
         endif
         this.qColumnDefs = AppQuery(cTable,'ColumnDef_No')  // Needed in ApplyLayout() method.
       endif
       *
       cTable = this.NameOfTableContainingLayoutDefinitions
       if not fileInCurrentdirectory( cTable + '.dbf' )
         CreateLayoutTable( cTable )  // Create empty table if it does not yet exist.
       endif
       this.qLayouts = AppQuery(cTable,'Description')
       *
       local r, lDefaultFound, lUserLayoutFound
       r = this.qLayouts.rowset
       * Look for the user's layout:
       lUserLayoutFound = false
                               
       if type('_app.Layout_No')=='N'
          if r.applyLocate("Layout_No = " + _app.Layout_No)
             lUserLayoutFound = true
             lDefaultFound = true
          else
            MsgBox("The user's layout for this grid is not found; default layout will be used.","Warning",48)
          endif
       endif
       
       if lUserLayoutFound == false
          
         if r.applyLocate("LDEFAULT = TRUE")
           lDefaultFound = true
         else
           lDefaultFound = false
           r.first()
         endif
       endif

       if not r.endOfSet       // If there is at least one layout defined...
         if not lDefaultFound
           MsgBox('There is no default layout for this grid; The first layout will be used.','Warning',48)
         endif
         this.applyLayout( r )
       EndIf
     CATCH ( Exception e )
       MsgBox('Error in the grid layout definition.','Error',16)
     EndTry
   return

EndClass
***************************************************************
func CreateLayoutTable
param cTable
  local nSaveWorkArea
  nSaveWorkArea = WorkArea()
  select select()
  create table &cTable. ( Layout_No    Autoinc,;
                          Description  Char(50),;
                          IndexName    Char(25),;
                          lDefault      boolean,;
                          Columns      Char(100) ;
                        )
  use (cTable) excl
  index on Layout_No tag Layout_No
  index on upper(description) tag description distinct
  use
  select (nSaveWorkArea)
return
***************************************************************
func CreateColumnsTable
param cTable
  local nSaveWorkArea
  nSaveWorkArea = WorkArea()
  select select()
  create table &cTable. ( ColumnDef_No Autoinc,;
                          ColumnDef    Char(100) ;
                        )
  use (cTable) excl
  index on ColumnDef_No tag ColumnDef_No
  use
  select (nSaveWorkArea)
return
****************************************************************************
class zContainer(parentObj) of CONTAINER(parentObj) custom
   with (this)
      width = 20
      height = 10
   endwith
   this.BaseClassName = 'CONTAINER'

endclass
****************************************************************************
class zProgress(parentObj) of PROGRESS(parentObj) custom
   with (this)
      printable = true
      height = 0.5
      width = 34
      borderStyle = 0        // Default
   endwith
   this.BaseClassName = 'PROGRESS'

endclass
****************************************************************************
class zNotebook(parentObj) of NOTEBOOK(parentObj) custom
   with (this)
      // OnLeftMouseDown =  class::OnLeftMouseDown
      focus  = 1
      height = 6
      width  = 33
   endwith
   this.BaseClassName = 'NOTEBOOK'

/* Function OnLeftMouseDown
    * Workaround to VdB bug that causes improper refresh of the controls on
    * a notebook that contains zGrids on each of its pages:
    local oldpage, newpage
    private oObject
    oldPage = This.CurSel
    This.SetFocus()
    This.SetFocus()

    oObject=This.First
    Do While (TYPE("oObject.SetFocus()")=="U")
     oObject=oObject.before
    EndDo
    oObject.SetFocus()

    newPage = This.CurSel
    This.CurSel = oldPage
    This.CurSel = newPage
    This.SetFocus()
   Return
*/

  function tabCount
    private m_macro
    local a
    m_macro = substr( this.datasource, at(" ",this.datasource)+1 )
    a = &m_macro.
  return a.size 

endclass
****************************************************************************
class zSlider(parentObj) of SLIDER(parentObj) custom
   with (this)
      onGotFocus = {;this.borderStyle=6}
      onLostFocus = {;this.borderStyle=3}
      height = 1.5
      width = 25
      borderStyle = 3        // None
      mousePointer = 13        // Hand
      value = 1
      rangeMax = 10
      rangeMin = 1
      tics = 0        // Auto
      ticsPos = 1        // Bottom Right
      vertical = false
      enableSelection = true
      startSelection = 0
      endSelection = 0
   endwith
   this.BaseClassName = 'SLIDER'
endclass
****************************************************************************
class zRectangle(parentObj) of RECTANGLE(parentObj) custom
   with (this)
      width = 25
      height = 5
   endwith
   this.BaseClassName = 'RECTANGLE'
endclass
****************************************************************************
class zTreeView(parentObj) of TREEVIEW(parentObj) custom
  with (this)
    height = 8
    width = 27
    ColorNormal='N/W+'
    OnGotFocus  = CLASS::YellowBackground
    OnLostFocus = CLASS::WhiteBackground
  endwith
  this.BaseClassName = 'TREEVIEW'

  PROC YellowBackground
    this.ColorNormal = 'N/RG+'

  PROC WhiteBackground
   this.ColorNormal = 'N/W+'

// Problem:
//  My treeview shall not be edited, so I set:
//     this.allowEditLabels = false
//    this.allowEditTree = false
//  although editing is not possible I can still delete branches of the tree by typing DEL.
// Solution:
//    this.onGotFocus = {;on key label del ??}
//    this.onLostFocus = {;on key label del}

endclass
****************************************************************************
class zReportViewer(parentObj) of REPORTVIEWER(parentObj) custom
   with (this)
      height = 10
      width = 36
   endwith
   this.BaseClassName = 'REPORTVIEWER'

endclass
****************************************************************************
class zReport of REPORT custom

   function Render
     set lock off
     super::render()
     set lock on
     // SET LOCK OFF to avoid file locks for reports, if ON it locks the file to
     // get accurate results in the summing and counting, if 20 people are busily
     // adding, deleting, changing records _during_ the run of the report, these
     // would be worthless. With LOCK OFF, that file lock is no longer automatic.
     // (Comment by Romain S.)
   return

endclass
****************************************************************************
class zStoredProc of STOREDPROC custom
   this.BaseClassName = 'STOREDPROC'
endclass
****************************************************************************
class zSession of SESSION custom
   this.BaseClassName = 'SESSION'
endclass
****************************************************************************
*class zQuery of RecnoQuery custom from "RecnoQry.cc"
class zQuery of Query custom
   this.BaseClassName = 'QUERY'

*  with (this)
*    OnOpen  = CLASS::QueryOnOpen
*    OnClose = CLASS::QueryOnClose
*  endwith
*
* function QueryOnOpen
*   If (TYPE("Form.DataModRef1.Ref.aQueries")<>'A')
*     Form.DataModRef1.Ref.aQueries=New Array()
*   EndIf
*   Form.DataModRef1.Ref.aQueries.Add(this.name)
*   *  ExMsgBox(this.name)        // No such property? What about Bowen's idea?
*   super::onOpen()
* return
*
* function QueryOnClose
*   * we need to delete this query from the array of queries...
*   Form.DataModRef1.Ref.aQueries.Delete(Form.DataModRef1.Ref.aQueries.Scan(this.name))
* return

 Procedure execute
   this.Rowset.AddCalcField = {|cFieldName,bBeforeGetValue| zQuery::AddCalcField(cFieldName,bBeforeGetValue) }
   // Note that we prefer to establish the rowset.AddCalcField
   // method BEFORE the super::Execute. Why? Because
   // the super::execute() causes the query's onOpen event
   // to fire!  So this way, we allow the programmer to
   // to utilize AddCalcField() in the query's onOpen,
   super::execute()   // Execute the query, to produce the rowset.
   if type('this.rowset.NativeSave') == 'U'  // If this was not invoked by a Requery()...
     *
     * if type('_app.aQueries') # 'O'
     *   _app.aQueries = new AssocArray()
     * endif
     * _app.aQueries[ ltrim(str(_app.aQueries.size+1)) ] = this
     *
     * Over-ride the rowset's Save() and Delete() and SetRange() methods:
     this.rowset.NativeSave        = this.rowset.Save
     this.rowset.NativeDelete      = this.rowset.Delete
     this.rowset.Save     = {; this.SaveAndFlush() }
     this.rowset.Delete   = {; this.DeleteAndFlush() }
     this.rowset.SaveAndFlush   = {; zQuery::SaveAndFlush() }
     this.rowset.DeleteAndFlush = {; zQuery::DeleteAndFlush() }
     this.rowset.AskUserToConfirmDelete = true   // Programmer can over-ride this (i.e., set it to FALSE) to suppres the "Are you sure you want to delete?" prompt.

     this.Rowset.GotoRecord = {|nRec| zQuery::GotoRecord(nRec) }
     this.Rowset.Count = {|| zQuery::Count() }
     this.Rowset.PreventEndOfSetErrors = {|x| zQuery::PreventEndOfSetErrors(x) }

     * Prevent accidental rowset navigation when user hits PgUp/PgDn:
	  this.Rowset.OnAppend  = {; zQuery::DisableNavigation() }
*???By Serge on 98.12.24:          this.Rowset.OnAppend  = {; zQuery::QueryRowsetOnAppend()}
     this.Rowset.OnEdit    = {; zQuery::DisableNavigation() }
     this.Rowset.OnSave    = {; zQuery::EnableNavigation() }
     this.Rowset.OnAbandon = {; zQuery::EnableNavigation() }

     /* This section removed because it causes the table to be NOT CLOSED when the
      * form closes.  Note that a native Query does not require an explicit
      * query.active=false (unlike this class, which does require it) in order
      * for the table to be "released" - which is necessary for ZAP/PACK/REINDEX/etc.
     this.rowset.NativeSetRange = this.rowset.SetRange
     this.rowset.MySetRange     = zQuery::MySetRange   // was:  = {|p1,p2| zQuery::MySetRange(p1,p2) }
     this.rowset.SetRange = this.rowset.MySetRange   // was:  = {|p1,p2| this.MySetRange(p1,p2) }
     ** The line above was commented out because it cause an infinite loop
     ** when MySetRange() was invoked.  (I don't know why, since SaveAndFlush()
     ** works perfectly using the same technique.) So, our programs should
     ** explicitly call MySetRange rather than SetRange.
     this.rowset.SwitchIndexButPreserveRange = {|cIndexName|; zQuery::SwitchIndexButPreserveRange(cIndexName) }
      * (end of removed section)
     */

   endif
   *
   *-Added by Serge on 98.12.24
*	procedure QueryRowsetOnAppend
*     super::DisableNavigation()
*   return
   *
 return

 Function SaveAndFlush   // This will be employed as a method of the rowset that results from this query.
   * "This" is a rowset.
   UpdateModStamp(this,true)   // 2nd param means that UpdateModStamp() will not issue a Save(), so this avoids infinite loop.
   this.NativeSave()
   this.flush()
   this.CanNavigate = {|| true}  //   this.EnableNavigation()  // Just to be sure, because the OnSave event does not fire if there were no pending changes.
 return

 Function DeleteAndFlush   // Will be employed as a method of the rowset.
   *
   if type('_app.UserHasReadOnlyPrivilege') == 'L' and _app.UserHasReadOnlyPrivilege
     tmsgbox('Cannot delete - your user privilege is read-only.','Permission denied',48)
     return
   endif
   *
   * "This" is a rowset.
   if type('this.AskUserToConfirmDelete')=='U' or this.AskUserToConfirmDelete
     local cPromptTitle
     if type('this.PromptTitleForDelete') $ 'CB/FP'  // Check for optional custom method.
       cPromptTitle = this.PromptTitleForDelete()   // It must return a string.
     else
       cPromptTitle = translate("Confirmation")  // This is the default, if there's no codeblock for the titile.
     endif
     if MLYesNo(translate("Are you sure that you want to delete this record?"),cPromptTitle,.t.) <> 1
       return
     endif
   endif
   UpdateModStamp(this,true)  // For Data Synch. Must do this BEFORE the Delete()
   // Try..catch around this next line?
   this.NativeDelete()   // Row pointer automatically moves to next row, or to end-of-set.
   this.flush()
   if this.EndOfSet
     this.Last()      // More user-friendly than leaving it at EndOfSet.
   else
     * Row pointer has automatically moved to the next row. VdB does not
     * consider this to be a navigation. Therefore we will explicitly fire
     * the rowset's OnNavigate event, if there is such an event:
     if not empty(this.onNavigate)
       this.onNavigate()
     endif
   endif
 return

 Procedure MySetRange(pLow,pHigh)   // Will be employed as a method of the rowset.
   this.Low = pLow
   this.High = iif( pHigh=false, pLow, pHigh )  // Accounts for possibility of only 1 param being passed.
   this.NativeSetRange(this.Low,this.High)
 return

 Procedure SwitchIndexButPreserveRange(cIndexName)   // Will be employed as a method of the rowset.
   this.IndexName = cIndexName
   if type('this.Low') # 'U'    // If a range is currently in place...
     this.NativeSetRange( this.Low, this.High )
   endif
 return

 function GoToRecord    // "this" is a rowset
     Parameters nRec
     local lRet
     lRet=.f.
     if pcount()=1 and type("nRec")="N"
       If Type("GotoRec") # "FP"
         extern cint GotoRec  (cHandle,cInt);
         IDAPI32 from "DbiSetToRecordNo"
       endif
       if (GoToRec(this.handle,nRec) == 0)
         lRet=.t.
         local bm
         bm = this.bookmark()
         this.goto(bm)    // need this to make the navigation "stick".
       endif
     endif
 return lRet


 function DisableNavigation
   * "This" is a rowset.  Prevent navigation, e.g. when an unsaved row buffer is pending:
   this.CanNavigate = {; tmsgBox('Cannot navigate when in APPEND or EDIT mode',ExtractTableNameFromSQL(this.parent.SQL)+' mode is '+this.state, 48) ; return false}
 return


 function EnableNavigation
   * "This" is a rowset.  Re-enable navigation:
   this.CanNavigate = {|| true}
 return


 function GetTableName
   local cRet
   cRet = ""
   if type( "GetTableNameFromQueryHandle" ) # "FP"
     extern cInt GetTableNameFromQueryHandle (cHandle, cString);
            IDAPI32 from "DbiGetCursorProps"
   endif
   nH = this.rowset.handle
   c=space(2048)
   if (GetTableNameFromQueryHandle( nH, c) == 0)
     cRet = c.left(255).lefttrim()
   endif
 return cRet


 function Count   // "this" is a rowset
   * The native rowset.count() method is unreliable.
   * So we'll do a brute-force count:
   local b, nCount, lSaveNotify
   lSaveNotify = this.notifyControls
   this.notifyControls := false
   b = this.bookmark()
   nCount = 0
   this.first()
   do while not this.endOfSet
     nCount++
     this.next()
   enddo
   this.goto(b)
   this.notifyControls := lSaveNotify
 return nCount

 function PreventEndOfSetErrors
   param fieldValueToReturnWhenAtEndOfSet
   //
   // "this" is a rowset.
   // Can be useful to prevent "End of set" errors in child tables (in
   // a MasterRowset relationship) which occur when a child-table's field's value
   // is referenced when the child table is at EndOfSet.
   //
   // The param can be string (or other value), or a codeblock or function pointer that
   // returns a value.  A CB/FP could be used here to advise the user which
   // parent row caused an endOfSet in the child table, or perhaps to log the
   // error somewhere. Example:
   //   b = {|oField| ; msgbox('Error: Child table at end of set at '+oField.parent.parentRowset.fields['Descrip'].value) ; return '??' }
   //   r.PreventEndOfSetErrors( b )
   //
   local cType, f, i
   cType = type('fieldValueToReturnWhenAtEndOfSet')
   if argCount() == 0  // If param was omitted...
     this.fieldValueToReturnWhenAtEndOfSet = null  // This is the default.
   else
     this.fieldValueToReturnWhenAtEndOfSet = fieldValueToReturnWhenAtEndOfSet
   endif
   f = this.fields
   for i = 1 to f.size
     if empty( f[i].beforeGetValue )
       if cType=='CB' or cType=='FP'
         f[i].beforeGetValue := {|| iif( this.parent.parent.endOfSet, this.parent.parent.fieldValueToReturnWhenAtEndOfSet((this)), this.value ) }
       else
         f[i].beforeGetValue := {|| iif( this.parent.parent.endOfSet, this.parent.parent.fieldValueToReturnWhenAtEndOfSet, this.value ) }
       endif
     endif
   next
 return

 func AddCalcField(cFieldName,bBeforeGetValue)
   // THIS is a rowset.
   local f
   f = new field()
   f.fieldName := cFieldName
   f.beforeGetValue := bBeforeGetValue
   this.fields.add(f)
 return

endclass
****************************************************************************
class zDatabase of DATABASE custom
   this.BaseClassName = 'DATABASE'
endclass
****************************************************************************
class zDataModRef of DATAMODREF custom
  * This one might be useless.
   this.BaseClassName = 'DATAMODREF'
endclass
****************************************************************************
*
* zForm (VdB7 base form) class
* Author: Peter Rorlick
* Thanks to Romain and Bowen for their ideas.
/*
  Things happen in the following order, each time you call form.open() or form.readModal():
   1. Form.init() method executes.
   2. Init() methods of objects on the form execute, one by one. (But only the first time the form is opened.)
   3. Form.AfterControlInits() method executes.
   4. The form becomes visible.
   5. Form.OnOpen() method executes.
   6. OnOPen() methods of objects on the form execute, one by one.

  Bear in mind that the CONSTRUCTOR section of your form fires ONLY when you do
    <something> = new WhateverForm()
*/
*
class zForm of FORM custom

  with (this)
     top = 3
     height = 15
     width = 50
     Left = 10
     text = 'zForm'
     Printable = false // as default is not prinatable from the RecPrint
     OnSize = Class::Form_OnSize
     AutoCenter = true
     onLeftDblClick = class::PrintControls
     onRightDblClick = class::PrintControls
  endwith
  this.BaseClassName = 'FORM'
  this.FirstTime = true
  this.Language = 'English'
  * this.AllowMultipleConcurrentInstances = true   // Programmers can set this to FALSE if they like, in their form's Init method.
  _app.ActiveForm = this

  Proc Open
    form.lIsInFocus = false  // Don't need to put this in the ReadModal proc, because modal forms are ALWAYS in focus.(?)
    if class::beforeOpen()
    	return super::open()
    else
    	return false
    endif

  Proc ReadModal
    if class::BeforeOpen()
      return super::ReadModal()
    else
    	return false
    endif

  Func BeforeOpen
    *
    // The next bit (suggested by George Burt) is a work-around to VdB7.01's bug
    // where (this.name) is unreliable. He says it "reminds" VdB what the real names are.
    // We probably don't need this anymore, however, since we now avoid using the Name
    // parameter when we define custom classes.
    * for i = 1 to form.elements.size
    *   Mtemp = form.elements[i].name
    * next

    //Modified on 2000.08.21, previous version stored in 
    // \obsolete\ZCONTROLOf20000816At1428.CC
    private lResult
    lResult = form.Init()
    if type('lResult')=='L' and not lResult
    	return false
    endif
    * If Class::PreventMultipleConcurrentInstances()
    *   return false   // Do not allow the form to open
    * Endif
    if form.FirstTime
      CLASS::RunControlInits((Form))   // Only runs the first time the form is opened.
    endif
    form.AfterControlInits()
    *
    if type('_app.lScripts') == 'U'    // We'll look for the existence of the file on disk only once, since that's a bit slow to do.
      _app.lScripts = FileInCurrentDirectory( 'Scripts.dbf' )
    endif
    if _app.lScripts  // if scripts.dbf exists...
      * Run the associated script file for this form, if one is defined:
      if type('_app.qScripts') == 'U'    // Run the query only once.
        _app.qScripts = new zQuery()
        _app.qScripts.sql = 'Select * from Scripts'
        _app.qScripts.active = .t.
        _app.qScripts.rowset.IndexName = 'ObjectName'
      endif
      if _app.qScripts.rowset.findkey(padr(this.className,30))  // if found...
        * If Scripts->ObjectName is NULL, then we can't find padr(this.className,80).
        if Empty(_app.qScripts.rowset.fields['ScriptFile'].value)
                if not Empty(_app.qScripts.rowset.fields['ScriptMemo'].value)
                                  runCb( _app.qScripts.rowset.fields['ScriptMemo'].value, form )
           endif
        else
                if FileInCurrentDirectory( _app.qScripts.rowset.fields['ScriptFile'].value )
                  runCbFil( _app.qScripts.rowset.fields['ScriptFile'].value, form )
                  * Note that in the script file, P1 refers to the form.
                endif
        endif
      endif
    endif
    *
    if type('_app.ULanguage')=='C' and _app.ULanguage#form.Language
      // The form is not currently in the user's preferred language.
      // Translate .Text properties, speedtips, etc. to user's preferred language:
      //
      // Note: This won't work if we're trying to translate a form from
      // French back to English.  For that to work properly, you must use the AppForm()
      // function, in Funclib.prg.
      //
      if type('_app.oLayoutManager') <> "O"
        set proc to LayOutMn.prg addi
        _app.oLayoutManager = new LayoutManager()
      endif
      Class::TranslateTexts((Form))
      Form.Text = Translate(Form.Text)   // A final touch.
      Form.Language = _app.ULanguage  // Important, in case the form is closed and re-opened.
    endif
    *
    form.FirstTime := false
    *
  return true   // Allow the form to open


  Proc TranslateTexts(InitialObject)
    local i

    local aContainers  //This will handle containers afterwards... - Serge...
    aContainers = new array()

    * Translate .Text properties to user's preferred language:
    private o
    o = InitialObject.First
    do
      do case
      case o.className == 'SBUTTON'
        * No text translation or recursion necessary.  But it might have a speedTip.
      case Type('o.Text')=='C'  // If this control object has a .Text property...
        o.Text = Translate(o.Text)
        if type('o.FontSizeFixed')#'L' or o.FontSizeFixed==false
          _app.oLayoutManager.FitTextInObject(o)
        endif
*      case Type('o.First')=='O' and type('o.UseTablePopUp')=='U'  // if o is a container or notebook (but not a form, because forms have a UseTablePopUp property) that contains other objects...
*        // Notebooks and containers do not have a Text property.
*???        CLASS::TranslateTexts(o)   // recurse through the objects of the container or notebook.
         ** The reason we commented this CASE and replaced it with the following one
         ** was to work around the problem of "Too many nested...".
      case typeOfObject(o) $ 'CONTAINER/NOTEBOOK'
			aContainers.add(o)
      case type('o.Columns')=='O'  // If this object is a grid...
        // Grid does not have a text property either.
        // Translate grid column headings:
        for i=1 to o.columns.size
          o.columns[i].headingControl.Value = Translate(o.columns[i].headingControl.Value)
        next
      endcase
      if Type('o.SpeedTip')=='C' and not empty(o.SpeedTip)  // If this control object has a .SpeedTip property...
        o.SpeedTip = Translate(o.SpeedTip)
      endif
      if type('o.DataSource')=='C' and left(upper(o.DataSource),6)=='ARRAY '
        * o is a ComboBox, ListBox, NoteBook, or Listbox.
        * Translate the array elements of the datasource:
        private a, c
        c = substr(o.DataSource,7)
        * NOte that c might be "{'aaa','bbb'}" or it might be "ExamStausChoices()"
        a = &c
        for i=1 to a.size
          a[i] = Translate(a[i])
        next
        o.DataSource="Array " + ArrayToString(a)
      EndIf
      o = o.before
    until o.name == InitialObject.First.name
    for i = 1 to aContainers.size
    	class::TranslateTexts(aContainers[i])
    next
  return


  Proc Init
  * Over-ride this procedure in forms that are inherited from this class,
  * to do any stuff that needs to happen when the form opens.
  * Use an OnOpen event handler only for the stuff that must be done AFTER
  * a form opens - i.e., stuff that requires a form window handle.


  Proc AfterControlInits
  * Over-ride this procedure in forms that are inherited from this class,
  * to do any stuff that needs to happen when the form opens but AFTER
  * the Init methods of its controls have been run.


  Proc RunControlInits(InitialObject)
    * Run the Init methods of any controls that have such methods.
    * (This is often faster than running OnOpen events, where the
    * form is already open.  This is especially effective for things
    * like entryPickers, which build buttons on-the-fly during their
    * onOpen/Init methods.)
    * For example, your form's Init method could have something like this:
    *    Form.zEntryField1.Init = CLASS::DoWhatever  // function pointer
    *    Form.zEntryPicker2.Init = {; EntryPicker::Init() ; CLASS::DoWhatever() }
    * In some cases, this can give you better performance than using the OnOpen
    * events of the controls. The reason is that when the Init methods are fired
    * the form is not yet open, so there are no GUI controls to refresh.
    private o
    o = InitialObject.First
    do
      if Type('o.Init') $ 'FP/CB'   // If this control object has an Init method...
        o.Init()
      endif
      If Type('o.First')=='O' and type('o.UseTablePopUp')=='U'  // if o is a container or notebook (but not a form, because forms have a UseTablePopUp property) that contains other objects...
        CLASS::RunControlInits(o)   // recurse through the objects of the container
      EndIf
      o = o.before
    until o.name == InitialObject.First.name
  return


//  Proc RegisterThisForm
//    * Create a global reference to this form, for debugging or other purposes:
//    * (Note: this only holds up for SDI forms.)
//    if type('_app.forms') = 'U'
//      _app.forms = new object()
//      * array()   // an array of form objects, e.g. { {'MyForm',<object>,nCountHowManyCurrentlyOpen},
//    endif
//    Private c
//    c = '_app.forms.' + this.ClassName + ' = This'
//    &c
//  return


  Procedure Repaint   // Call this method to refresh the form. More thorough than Form.Refresh().
    if type("UpdateWindow") # "FP"
      extern cLogical UpdateWindow(cHandle) user32
    endif
    if argcount()=1
       UpdateWindow(argVector(1))
    else
       UpdateWindow(form.hwnd)   // Or is it better to change the PageNo?
    endif
    *
    *  ** The old way:
    *  local nSavePageNo
    *  If Type('Form.Rowset')=='O'
    *    Form.Rowset.RefreshControls()
    *  Endif
    *  nSavePageNo = Form.PageNo
    *  Form.PageNo = 99
    *  Form.PageNo = nSavePageNo   // To solve the refresh problem.
    *
  Return

// // This doesn't work...
//  Func PreventMultipleConcurrentInstances   // Returns True if the form is already open and the user is not allowed to open it twice concurrently.
//    If form.AllowMultipleConcurrentInstances == false
//      If type('_app.forms.' + this.ClassName ) == 'O'  // If this form has been opened at least once during this session...
//        * Is it open right now?
//        If not Empty(form.hWnd)   // If this form has a Windows handle right now...
//          tmsgBox('This form is already open')
//          return True
//        Endif
//      Endif
//    Endif
//  return false


  procedure OnGotfocus
    form.lIsInFocus = true
    _app.activeform = form
    on key label f5 _app.activeform.PrintIt()
    form.repaint()
  return


  proc onClose
    * The following lines are to solve the glitch where the key associations
    * such as Up/Down/Enter persist
    * after closing the form while a picker's dropdownlist has focus.
    local f
    f = new form()
    f.top := -300
    f.open()
    f.close()
    rele f
  return


  procedure OnLostfocus
    form.lIsInFocus = true
    _app.activeform = null
    on key label f5
  return


  procedure PrintIt
    on key label f5     // Prevent multiple printouts for users with heavy fingers.
    *keyboard chr(13)   // Uncomment this to choose the default printer and avoid the printing dialog.
    _app.activeform.PrintThisRecord()
    on key label f5 _app.activeform.Printit()
  return

****
  Procedure PrintThisRecord

   if true
     return   // Feature disabled for now - needs more testing, etc..
   endif

   If (TYPE("Form.DataModRef1.Ref.aQueries")<> "A") AND (TYPE("Form.rowset")<> "O")
     tMsGBox('There is no query running','Nothing to print',16)
     retu
   EndIf

   tmpTable1 = RandomFileName(".DBF")
   tmpTableName1 = LEFT(tmpTable1,AT(".DBF",UPPER(tmpTable1))-1)

   SELE SELE()
   // Structure
   c = "CREATE TABLE '" + tmpTable1 + "' ("+;
         "SELECTED     BOOLEAN,"+;
         "FIELDNAME    CHAR(50),"+;
         "FIELDVALUE   BLOB(0,1))"
   &c

   USE (tmpTable1) EXCL
   _app.Rowset = Form.Rowset

   // All the fields in DataModRef1 *minus* the ones listed in
   //  the optional Form.aFieldsToExclude, OR:

   // The list contained in the optional Form.aFieldsToInclude

   lExcludeSomeFields = (TYPE("Form.aFieldsToExclude")=="A")
   aFields = New array()
   aFieldLinks = New array()

   If (TYPE('Form.DataModRef1')=='O')
     For i=1 to aLen(Form.DataModRef1.Ref.aQueries)
       _app.DataPrint=form.DataModRef1.Ref
       c = 'TYPE("Form.DATAMODREF1.ref.' + Form.DataModRef1.Ref.aQueries[i] + '")'
       c = &c

       If (c=="O")
         c="Form.DATAMODREF1.ref."+Form.DataModRef1.Ref.aQueries[i]+".Rowset.fields.size"
         nn=&c
         For j=1 to nn
  //         MsgBox(Form.DataModRef1.Ref.aQueries[i]+'.'+Form.DataModRef1.Ref.aQueries[i].Rowset.fields.fieldname)
           cc="Form.DATAMODREF1.ref." + Form.DataModRef1.Ref.aQueries[i] + ".Rowset.fields[j].fieldname"
           c=Form.DataModRef1.Ref.aQueries[i]+".'" + &cc + "'"
           aFields.Add(c)

           cc="Form.DATAMODREF1.ref." + Form.DataModRef1.Ref.aQueries[i] + ".Rowset.fields[j]"
           aFieldLinks.Add(&cc)
         Next
       EndIf
      Next
   ElseIf(TYPE('Form.Rowset')=='O')
      _app.DataPrint=New Object()
      _app.DataPrint.Rowset=Form.Rowset
      For j=1 to Form.Rowset.fields.size
        aFields.Add(Form.Rowset.fields[j].fieldname)
        aFieldLinks.Add(Form.Rowset.fields[j])
      Next
   EndIf


   If TYPE("Form.aFieldsToInclude")=="A"
     For i=1 to aLen(Form.aFieldsToInclude)
      replace selected with false,;
         fieldname with Form.aFieldsToInclude[i]
     Next
   Else
     For i = 1 to aFields.size  // _app.Rowset.fields.size

      lFound=.f.
      sFieldName=aFields[i] //_app.Rowset.fields[i].fieldName

      If (lExcludeSomeFields)
//        MsgBox('Cest grande probleme')
        For j = 1 to aLen(Form.aFieldsToExclude)
          If (Form.aFieldsToExclude[j]==sFieldName)
           lFound=.t.
           exit
          EndIf
        Next

      EndIf

      c = ""+aFieldLinks[i].Value
      If .not.lFound
        append blank
        replace selected with false,;
           fieldname with sFieldName,;
           FIELDVALUE with c
      EndIf
     Next
   EndIf

   use    // close &tmpTable1

   _app.tmpTableName = tmpTableName1

   Do printrec.wfm with true

   if type('_app.forms.printrecForm') == 'O'
     _app.forms.printrecForm = null   // Destroy reference to the form so that we can delete the temp table.
   endif

   tmpTableName1 = LEFT(tmpTable1,AT(".DBF",UPPER(tmpTable1))-1)

   If (SELE(tmpTableName1)<>0)
     select &tmpTableName1
     USE
   EndIf

   Delete table (tmpTable1)

  return


  Proc Form_OnSize
    class::RunOnFormResizeRoutines((Form))   // call recursive routine
  return


  Proc RunOnFormResizeRoutines(InitialObject)
    * Run the OnFormResize methods of any controls that have such methods.
    private o
    o = InitialObject.First
    do
      If Type('o.First')=='O' and type('o.UseTablePopUp')=='U'  // if o is a container or notebook (but not a form, because forms have a UseTablePopUp property) that contains other objects...
        CLASS::RunOnFormResizeRoutines(o)   // recurse through the objects of the container
      EndIf
      if Type('o.OnFormResize') $ 'FP/CB'   // If this control object has such a method...
        o.OnFormResize()
      endif
      o = o.before
    until o.name == InitialObject.First.name
  return


  Function UserStop(hWND)          // 32-bit version
    If Type("PeekMessage") # "FP"
       extern CLOGICAL PeekMessage(CSTRING, CHANDLE,;
                                   CLONG, CLONG, CLONG) User32;
                 From "PeekMessageA"
    EndIf
    #define PM_NOYIELD        2
    #define WM_LBUTTONDOWN    HToI("201")
    #define WM_LBUTTONUP      HToI("202")
     FORM.MsgBuff = Space(14)            // Buffer for MSG structure
  return PeekMessage(FORM.MsgBuff,;   // Scratch buffer for msg info.
                     hWND,;           // Look for msgs for specified hWND
                     WM_LBUTTONDOWN,; // Low msg filter (left button down)
                     WM_LBUTTONUP,;   // High msg filter
                     PM_NOYIELD)      // Don't yield to other applications

  Proc PrintControls
    // Invoked by LeftDblClick
    // Can also be called in a pushbutton's onClick, etc.
    //
    // Author: Peter Rorlick
    //
    // Print all of the controls on the form that are of these types:
    //  EntryField
    //  Editor
    //  Checkbox
    //  RadioButton
    //  Combobox
    //  Spinbox
    // Each of these controls will have a promptText
    //   If a control has a PromptTextForPrintout property, that becomes the promptText.
    //   Otherwise:
    //    For RadioButtons and checkBoxes, the promptText = oControl.text
    //    For each of the others, find out if the previous control in the z-order is
    //     a Text (or ZText etc) or Rectangle object.  If it is, then the promptText = oPreviousControl.text
    // The promptTexts will be underlined when they are printed.
    // Title for the report = form.text
    // What we intend to print will be shown in an Editor in a dialog form,
    // with buttons "Print" and "Close".  The user can edit the text (if desired) before printing.
    //
    if type('form.TitleForPrintout') == 'C'
      _app.stringToPrint = newLine() + form.TitleForPrintout + newLine() + newLine()
    else
      _app.stringToPrint = newLine() + form.text + newLine() + newLine()
    endif
    form.oPreviousControl = null
    form.BuildPrintString( form )
    *
//    local f
//    f = new Form()
//    with (f)
//      height = 21
//      width = 109
//      autoCenter = true
//    endwith
//    f.ed = new EDITOR(f)
//    with (f.ed)
//      anchor = 6	// Container
//      value = _app.stringToPrint
//    endWith
//    f.mdi := false
//    f.readModal()   // f.open()

    set proc to FormPrin.wfm addi
    local f
    f = new FormPrinForm()
    f.zEditor1.value = _app.stringToPrint
    f.mdi := false
    f.readModal()
    if f.printClicked = false  // if user did not click the Print button...
      return
    endif
    _app.stringToPrint = f.zEditor1.value
    *
    ChunkifyAndPrintHtmlString( _app.stringToPrint )
    form.setFocus()
  return

  Proc BuildPrintString( InitialObject )
    local cPromptText, cValue, nAt
    private o
    o = InitialObject.First
    do
      cTypeOfObject = TypeOfObject(o)
      if cTypeOfObject $ 'RADIOBUTTON|CHECKBOX|ENTRYFIELD|COMBOBOX|SPINBOX|EDITOR' and ;
         o.Printable
        if cTypeOfObject $ 'RADIOBUTTON|CHECKBOX'
          cPromptText = o.text
          cValue = iif( o.value, 'Yes', 'No' )
        else
          if type('o.PromptTextForPrintout') == 'C' // check for optional custom property
            cPromptText = o.PromptTextForPrintout
          elseif type('form.oPreviousControl.text') == 'C' ;
               and TypeOfObject(form.oPreviousControl) $ 'TEXT|RECTANGLE'
            cPromptText = form.oPreviousControl.text
            cPromptText = strTran(cPromptText,'<p>','')
            cPromptText = strTran(cPromptText,'</p>',' ')
            * cPromptText = strTran(cPromptText,crlf(),newline())
            cPromptText = strTran(cPromptText,'&','')
          else
            cPromptText = ''   // Hopefully this will be avoided most of the time.
            if not empty(o.dataLink)
              // Here are two types of datalinks:
              //   o.dataLink = "name"
              //   odataLink = form.patients1.rowset.fields["Name"]
              if type('o.dataLink.fieldName') == 'C'
                cPromptText = o.dataLink.fieldName
              elseif type('o.dataLink') == 'C'
                nAt = at('->',o.datalink)
                if nAt > 0
                  cPromptText = subs(o.dataLink,nAt+2)  // Take everything to the right of the "->"
                else
                  cPromptText = o.dataLink  // Take the whole string, which we assume is a field name.
                endif
              endif
              // Improve the the look of the field name (upper/lower casing, and spaces):
              if cPromptText == upper(cPromptText)
                // The field name is entirely upper case...
                cPromptText = Proper(cPromptText)
              else
                cPromptText = ConvertFieldNameToPrompt(cPromptText)
              endif
            endif
          endif
          cValue = new string( ''+o.value )
        endif
        if cTypeOfObject == 'EDITOR'
          // Multi-line value.  We'll indent it, on the line that follows the promptText:
          _app.stringToPrint += Underline() + trim(cPromptText) + UnderlineOff() + ': '
          _app.stringToPrint += BlockQuote() + strTran(cValue,crlf(),newline()) + BlockQuoteEnd() + NewLine()
        else
          // Single-line value.  We'll print it on the same line as the promptText:
          _app.stringToPrint += Underline() + trim(cPromptText) + UnderlineOff() + ': '
          _app.stringToPrint += cValue + newLine() + newLine()
        endif
      endif
      If Type('o.First')=='O' and type('o.UseTablePopUp')=='U'  // if o is a container or notebook (but not a form, because forms have a UseTablePopUp property) that contains other objects...
        form.BuildPrintString(o)   // recurse through the objects of the container
      EndIf
      form.oPreviousControl = o
      o = o.before
    until o.name == InitialObject.First.name
  return

endclass  // zform

* Note: When creating a new form where zForm is the baseFormClass,
* if you want to resize the form, change its position/size properties at
* least once in the inspector - otherwise (if you do it just with mouse
* resizing), it won't "stick".  This is due to a bug in VdB which Borland
* says will be eliminated in 7.01.
****************************************************************************
class zMenu(oOwner) of MENU(oOwner) custom

  this.BaseClassName = 'MENU'

  private o
  o = this.parent
  do while true  //  Move up the tree until we find the MENUBAR or the POPUP object...
    if type('o.aObjects')=='A'
      o.aObjects.add( this )   // Add a reference to this new menu object to the menuBar's aObjects custom property.
      * aObjects is used for language translation.
      * See class ZMenuBar and class ZPopup.
      exit
    endif
    if type('o.parent')=='O'
      o = o.parent
    else
      exit  // Shouldn't happen.
    endif
  enddo

endclass
****************************************************************************
class zMenuBar(oOwner,cRoot) of MENUBAR(oOwner,cRoot) custom

  this.BaseClassName = 'MENUBAR'
  this.aObjects = new Array()

  Proc OnInitMenu
    zPopup::TranslateTexts()       // See comments in zPopup class.
  return

  Proc TranslateTexts
    zPopup::TranslateTexts()       // See comments in zPopup class.
  return

  * Strange bug: When the form is called Modal-ly, any menu prompts that
  * have submenus do not get translated.  It's also a bit glitchy if the
  * form is opened non-modal.  Seems like the problem is WHEN the OnInitMenu
  * is being fired.  Popups work fine either way.
  * (This bug may have been fixed, since we do translation differently now.)

endclass
****************************************************************************
class zPopup(oForm,cName) of POPUP(oForm,cName) custom

  this.BaseClassName = 'POPUP'
  this.aObjects = new Array()

  Proc OnInitMenu
    class::TranslateTexts()       // See comments in zPopup class.
  return

  Proc TranslateTexts
   * msgbox('translating menu to '+_app.ULanguage)
   if type('_app.ULanguage')=='C'
     * Translate each of the menu prompts to a different language:
     private i
     for i = 1 to this.aObjects.size
       if type('this.aObjects[i].EnglishText') == 'U'
         this.aObjects[i].EnglishText = this.aObjects[i].text  // Store it into a custom property - we'll need this later on for a 2nd or subsequent translation.
       endif
       if _app.ULanguage == 'English'
         this.aObjects[i].text := this.aObjects[i].Englishtext
       else
         this.aObjects[i].text := Translate( this.aObjects[i].Englishtext )
       endif
     next
   endif
   * Now refresh the menu:
   shell(true,true)
   shell(false,true)
  return

  * For the translation to work, the programmer must use ZMenuBar, ZPopup, and
  * ZMenu classes throughout.

endclass
************************************************************************
class FontPicker(parentObj) of ZCONTAINER(parentObj) custom
   with (this)
      left = 6.2857
      top = 1.0909
      width = 52.2857
      height = 3.4091
      metric = 0        // Chars
   endwith


   this.TEXT1 = new ZBASETEXT(this)
   with (this.TEXT1)
      height = 1
      left = 1.7143
      top = 0.5909
      width = 5.8571
      wrap = false
      alignVertical = 1        // Middle
      text = "Font"
   endwith


   this.FontName = new ZENTRYFIELD(this)
   with (this.FontName)
      height = 1
      left = 8.4286
      top = 0.5909
      width = 41.1429
      value = "                         "
   endwith


   this.FontBtn = new sBUTTON(this)
   with (this.FontBtn)
      onClick = class::FONTBTN_onClick
      height = 0.8636
      left = 46.1429
      top = 0.6818
      width = 3.1429
      text = ""
      upBitmap = "RESOURCE #148"
      value = false
   endwith


   this.TEXT2 = new ZBASETEXT(this)
   with (this.TEXT2)
      height = 1
      left = 8.4286
      top = 1.7727
      width = 4.4286
      wrap = false
      alignVertical = 1        // Middle
      text = "Size"
   endwith


   this.FontSize = new ZSPINBOX(this)
   with (this.FontSize)
      height = 1
      left = 13.7143
      top = 1.8182
      width = 9.5714
      rangeMax = 100
      rangeMin = 1
      value = 1
   endwith


   this.FontBold = new ZCHECKBOX(this)
   with (this.FontBold)
      height = 1.0909
      left = 29.2857
      top = 1.7727
      width = 7
      text = "&Bold"
      value = false
   endwith


   this.FontItalic = new ZCHECKBOX(this)
   with (this.FontItalic)
      height = 1.0909
      left = 42
      top = 1.7727
      width = 7.1429
      text = "&Italic"
      value = false
   endwith


   function FONTBTN_onClick
      local cFont
      cFont = getfont()
      if not isblank(cFont)
        this.parent.FontName.Value = left(cFont,at(',',cFont)-1)
        cFont = subs(cFont,at(',',cFont)+1)   // now it's nSize,[B][I],cSwiss
        this.parent.FontSize.Value = val(left(cFont,at(',',cFont)-1))
        cFont = subs(cFont,at(',',cFont)+1)   // now it's [B][I],cSwiss
        this.parent.FontBold.Value = (left(cFont,1)=='B')
        this.parent.FontItalic.Value = ('I'$left(cFont,2))
        Form.Repaint()
        if not empty( this.parent.FontName.onChange )
          this.parent.FontName.onChange()
        endif
      endif
      return
endclass
************************************************************************
class zEntryLabel(parentObj) of ENTRYFIELD(parentObj) custom
   with (this)
      when = {||false}
      height = 0.8
      left = 5
      top = 2
      width = 16
      metric = 0	// Chars
      colorNormal = "BtnText/BtnFace"
      colorHighLight = "BtnText/BtnFace"
      border = false
      value = ""
      borderStyle = 3	// None
   endwith
   this.BaseClassName = 'ENTRYFIELD'

endclass
************************************************************************
Class GetSet of object

*  Getter/Setter class - a container object for properties that can
*  have triggers (BeforeGetValue, CanChange, OnChange, OnGotValue).
*  For example:
*
*  o = new getset()
*  o.setvalue('SecurityLevel',10)
*  ? o.getvalue('SecurityLevel')   // gives 10
*  ? o.SecurityLevel.value         // gives 10
*  o.SecurityLevel.canChange = {|newValue| ; this.value = 10*newValue ; retu .f. }
*  o.setvalue('SecurityLevel',4)
*  ? o.getvalue('SecurityLevel')   // gives 40
*
*  Warning: Watch out for the following:
*    o.SecurityLevel = 10            // Wrong
*    o.SecurityLevel.value := 10     // Right
*    o.setvalue('SecurityLevel',10)  // Right

  func GetValue( cPropertyName )   // Alternative syntax to: whatever.value
    private c
    c = 'this.'+cPropertyName+'.value'
  return &c

  func SetValue                    // Recommended alternative syntax to: whatever.value:=something
    param cPropertyName, newValue
    private c
    if type('this.'+cPropertyName) # 'O'
      * Create the new property:
      c = 'this.'+cPropertyName+' = new field()'
      &c
    endif
    * Now assign the new value:
    c = 'this.'+cPropertyName+'.value'
    &c := newValue
  return

endClass
************************************************************************
class zTableDef of TableDef

  this.alreadyLoaded = false

  func load
    if this.alreadyLoaded and 'dBASE 7.5'$version()
      // The bug that prevents you from Load()ing twice is slated to be fixed in dB2K.
      msgbox('Cannot Load() a tableDef object more than once','Error',48)
    else
      super::Load()
      this.alreadyLoaded = true
    endif
  return

endclass
******************************************************************
func GetGridFont  // Needed for the zGrid and zBrowse classes.
  if type('_app.GridFontName') == 'C'
    return  // These properties are already set.
  endif

  _app.GridFontName   = null
  _app.GridFontSize   = null
  _app.GridCellHeight = null

  if type('_app.oINI') == 'O'  // If there's an INI file...
   _app.GridFontName   = _app.oINI.GetValue('Startup','GridFontName')
   _app.GridFontSize   = val( _app.oINI.GetValue('Startup','GridFontSize') )
   _app.GridCellHeight = val( _app.oINI.GetValue('Startup','GridCellHeight') )
  endif

  // Use factory defaults if values were not available from the INI file:
  // (Tahoma 8-point, with cellHeight of .71 is also very nice.)
  if empty(_app.GridFontName)
    _app.GridFontName = 'Arial'  
  endif
  if empty(_app.GridFontSize)
    _app.GridFontSize = 10
  endif
  if empty(_app.GridCellHeight)
    _app.GridCellHeight = 0.8
  endif

return
***********************************************************************
* SButton is a small square button that has no text and does not retain focus - e.g., for for Calendar.wfm.
CLASS SBUTTON(FormObj) OF PUSHBUTTON(FormObj) Custom
	With (this)
      height = 1.25 // was 1.2
      width = 3.55 // was 3.5
      metric = 0
      tabStop = false
      group = true
      value = false
      text = " "
      SpeedBar = true
   EndWith
EndClass
******************************************************************
Func Tmsgbox(cMsg,cTitle,n1,n2)   // multilingual message box
retu ExMsgBox( Translate(cMsg), Translate(cTitle), n1, n2 )
*******************************************************************
func ExMsgBox
// By Peter Rorlick
// This is a Msgbox that accepts any data type as the message.

param cMsg, cTitle, nIcon

If (TYPE("cTitle")<>'C')
  cTitle=""
EndIf

If (TYPE("nIcon")<>'N')
  nIcon=0
EndIf

retu msgbox( iif(cMsg==null,'Null',''+cMsg), cTitle, nIcon)
**********************************************************************
Func FileInCurrentDirectory(cFile)   // Check for existence of a file in the current dir.
local retVal
private savepath
savepath = set('PATH')
set path to
retVal = File(cFile)
set path to &savepath
Return RetVal
********************************************
