// New version, to be submitted for the dUFLP. 
/*
  Create ODBC-DSN
  Create ODBC-Entrys during Runtime. Program is tested with MS Access, 
  MS Excel, Firebird ODBC-Driver and MySQL-ODBC-Driver. It can be used for all 
  other ODBC-Drivers. You must know the correct parameters for them 
  (Create a ODBC-Entry and look in the registry)

  Default is using german Language and driver. If you like to switch 
  to english-driver and Language as default you have to set the 
  variable cLang = "EN"
*/
function odbc
/*
   -----------------------------------------------------------------------
   Programmer..: Andreas Beckhaus
   Date........: July 6, 2004
   Notes.......: Creating an New Access-SYSTEM-DSN in ODBC
   Written for.: dBPlus
   Rev. History: 07/01/2004 0.1 - Original version 
                 07/06/2004 0.2 - Fixed Bug with aPara5
                                  Add Restart DBE so you do not need 
                                  a restart of the application 
                                  (Notice: There is also used
                                  CLOSE DATABASES)

                 May 17, 2006 -- Ken Mayer
                                 Fixed bug: uses :samples: source
                                            code alias for registry.prg
                                 Added code: to change BDE Registry
                                            setting for AUTO ODBC
                                            to TRUE, as this is 
                                            required for the ODBC
                                            and BDE Alias bits to
                                            work together properly.

                 April 21, 2007 -- A. Beckhaus, R. Wingerter
                                 Improved Excel support (DO CASE ...)
                                 Added support for binary datatype in 
                                 optional 4th parameter (options array)
                 					  
   Calls.......: registry.prg from dbasehome\samples
                 winreg.h from dbasehome\includes
                 windef.h from dbasehome\includes
                 may be copied to the same directory as this file
   Called by...: Any
   Usage.......: odbc(SYSTEM_DSN, PATH, DRIVER, OPTIONSARRAY)
   Example.....: 
                 set proc to odbc.prg additive 
                 (or remove the line "funtion odbc" and call odbc.prg directly)
                 Access-Entry (German driver):
                 ? odbc("TESTDB","c:\xx\xyz.mdb","Microsoft Access-Treiber (*.mdb)")
                 Access-Entry (english driver)
                 ? odbc("TESTDB","c:\xx\xyz.mdb","Microsoft Access Driver (*.mdb)")
                 Firebird-Entry
                 ? odbc("FBTESTDB","server:c:\xx\database.fdb","Firebird/InterBase(r) driver")

                 Sample for Excel:
					  ? odbc("EXCEL_ODBC","c:\xx\xyz.xls","Microsoft Excel Driver (*.xls)")	

                 Firebird-Entry with Character Set "NONE" AND USER "Fritz":
                 aTmp = NEW ARRAY(1,4)
                 aTmp[1,1] = ""
                 aTmp[1,2] = "CharacterSet"
                 aTmp[1,3] = "NONE"
                 aTmp[1,4] = "C"
                 aTmp.grow(1)
                 aTmp[2,1] = ""
                 aTmp[2,2] = "USER"
                 aTmp[2,3] = "Fritz"
                 aTmp[2,4] = "C"
                 ? odbc("FB1TESTDB","server:c:\xx\database.fdb","Firebird/InterBase(r) driver",aTmp)

                 MySql-Entry :
                 aTmp = NEW ARRAY(1,4)
                 x = 1
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Driver"
                 aTmp[x,3] = "C:\WINDOWS\System32\myodbc.dll"
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Description"
                 aTmp[x,3] = ""
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Database"
                 aTmp[x,3] = "mysqldb"
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Server"
                 aTmp[x,3] = "Servername"
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Server"
                 aTmp[x,3] = "Servername"
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "User"
                 aTmp[x,3] = "Username"
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Password"
                 aTmp[x,3] = "pass"
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Port"
                 aTmp[x,3] = ""
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Option"
                 aTmp[x,3] = "1"
                 aTmp[x,4] = "C"
                 aTmp.grow(1)
                 x++
                 aTmp[x,1] = ""
                 aTmp[x,2] = "Stmt"
                 aTmp[x,3] = ""
                 aTmp[x,4] = "C"
                 ? odbc("MYTESTDB","xxx","MySQL",aTmp)


   Returns.....: true or false
   Parameters..: cPara2 = String - Name of the new Alias
                 cPara3 = String - Name of the Datebase-File with 
                                   complete Path
                 cPara4 = String (optional) Name of the Databasedriver 
                          (default = Microsoft Access-Treiber (*.mdb)")
                          (english = Microsoft Access Driver (*.mdb)")
                 aPara5 = ARRAY (n,4) (optional) with values
                          [1] = "" -> Key direct in ODBC.INI\<ALIASNAME>
                                "\Engines\Jet" -> Key in ODBC.INI\<ALIASNAME>\Engines\Jet
                          [2] = Keyword
                          [3] = Value
                          [4] = "C" = String "N" = DWORD, "B" = REG_BINARY 
      
      						  Values to pass for datatype REG_BINARY
                          (thanks to Ivar B. Jessen, who came up with 
                          the idea to try chr(0)):

                          Passed value    registry value 	meaning
                          -----------------------------------------------
                          - chr(0) 			00    				false 
                          - chr(1) 			01 00 				true
   ---------------------------------------------------------------------
*/
       
   parameters cPara2, cPara3, cPara4, aPara5

   // Language: (default = "DE". switch to "EN" for using English 
   // Driver and Language)
   cLang = "DE"

   IF EMPTY(cPara2)
      // Kein Name / No DNS-Name
      DO CASE
         CASE cLang = "DE"
              MSGBOX("ODBC-Eintrag - Kein Datenbankname angegeben",;
                     "Datenprfung",16)
         CASE cLang = "EN"
              MSGBOX("ODBC-Entry - No DATABASENAME","",16)
         OTHER
              MSGBOX("ODBC-Eintrag - Kein Datenbankname angegeben",;
                     "Datenprfung",16)
      ENDCASE
      return false 
   ENDIF

   IF EMPTY(cPara3)
      // Kein Pfad / No Path
      DO CASE
         CASE cLang = "DE"
              MSGBOX("ODBC-Eintrag - Kein Pfad angegeben",;
                     "Datenprfung",16)
         CASE cLang = "EN"
              MSGBOX("ODBC-Entry - No DATABASE-FILE","",16)
         OTHER
              MSGBOX("ODBC-Eintrag - Kein Pfad angegeben",;
                     "Datenprfung",16)
      ENDCASE
      return false 
   ENDIF

   IF EMPTY(cPara4)
      // Kein Treiber / No Driver - using default
      DO CASE
         CASE cLang = "DE"
              cPara4 = "Microsoft Access-Treiber (*.mdb)"
         CASE cLang = "EN"
              cPara4 = "Microsoft Access Driver (*.mdb)"
         OTHER
              cPara4 = "Microsoft Access-Treiber (*.mdb)"
      ENDCASE
   ENDIF

   IF TYPE("aPara5") # "A"
      // Keine Optionen
      aPara5 = NEW ARRAY(1,4)
      aPara5.fill("")
   ENDIF

   #include <winreg.h>
   cAliasName = TRIM(cPara2)
   cAliasPfad = TRIM(cPara3)
   cAliasArt = TRIM(cPara4)
   aOptionen = aPara5

   // note :samples: source code alias:  
   SET PROCEDURE TO :samples:registry.prg ADDITIVE
   // code added by Ken Mayer to set the BDE's AUTO ODBC to true
   // this ensures that we can actually set things up 
   // properly:
   #define BDE_REG_KEY "SOFTWARE\Borland\Database Engine"
   oReg = new Registry(HKEY_LOCAL_MACHINE, ;
                       BDE_REG_KEY + ;
                       "\Settings\SYSTEM\INIT",false )
   xValue = oReg.setValue("AUTO ODBC","TRUE")

   // back to the original code:
   cKey =  "SOFTWARE\ODBC\ODBC.INI\ODBC Data Sources"
   TRY
      reg = new Registry(HKEY_LOCAL_MACHINE,cKey )
      reg.setValue(cAliasName,cAliasArt)
   CATCH (EXCEPTION E)
      return false 
   ENDTRY

   cSysDat = GETENV("WINDIR")
   DO CASE
      CASE  "WINDOWS NT VERSION" $UPPER(OS()) 
            cSysDat += "\system32"
       CASE  "WINDOWS VERSION" $UPPER(OS()) 
            cSysDat += "\system"
   ENDCASE
   cKey =  "SOFTWARE\ODBC\ODBC.INI\" + cAliasName

   TRY
      DO CASE
         CASE AT("ACCESS",UPPER(cAliasArt)) # 0
              // Microsoft Access
              cSysDat += "\odbcjt32.dll"
              reg = new Registry(HKEY_LOCAL_MACHINE,cKey )
              reg.setValue("UID","")
              reg.setValue("Driver",cSysDat)
              reg.setValue("DBQ",cAliasPfad)
              reg.setValue("DriverId","25",REG_DWORD)
              reg.setValue("FIL","MS Access;")
              reg.setValue("SafeTransactions","0",REG_DWORD)

              cKey = "SOFTWARE\ODBC\ODBC.INI\" + cAliasName + ;
                     "\Engines\Jet"
              reg = new Registry(HKEY_LOCAL_MACHINE,cKey )
              reg.setValue("ImplicitCommitSync","")
              reg.setValue("MaxBufferSize","256",REG_DWORD)
              reg.setValue("PageTimeout","5",REG_DWORD)
              reg.setValue("Threads","3",REG_DWORD)
              reg.setValue("UserCommitSync","Yes")

			CASE AT("EXCEL",UPPER(cAliasArt)) # 0
              // Microsoft Excel
              cAliasPfadExFileName = left(cAliasPfad, rat("\", cAliasPfad)-1)

              cSysDat += "\odbcjt32.dll" // same as for Access
              reg = new Registry(HKEY_LOCAL_MACHINE,cKey )
              reg.setValue("DBQ",cAliasPfad)
              reg.setValue("DefaultDir",cAliasPfadExFilename)
              reg.setValue("Driver",cSysDat)
              reg.setValue("DriverId","790",REG_DWORD) 
				  reg.setValue("Description","Created by :dUFLP:createODBC")              
              // reg.setValue("FIL","excel 8.0;")	
              reg.setValue("FIL","Excel;")	
				  // chr(1) = TRUE (Default)
              reg.setValue("ReadOnly", chr(1), REG_BINARY)  
              reg.setValue("SafeTransactions","0",REG_DWORD)
              reg.setValue("UID","")	  // empty	

              cKey = "SOFTWARE\ODBC\ODBC.INI\" + cAliasName + ;
                     "\Engines\Excel"
              reg = new Registry(HKEY_LOCAL_MACHINE,cKey )
              // chr(0) = 00 false ; chr(1) = 01 00 true
              reg.setValue("FirstRowHasNames",chr(1), REG_BINARY) 
              reg.setValue("ImplicitCommitSync","")
              reg.setValue("MaxScanRows","8",REG_DWORD)
              reg.setValue("Threads","3",REG_DWORD)
              reg.setValue("UserCommitSync","Yes")

         CASE AT("FIREBIRD",UPPER(cAliasArt)) # 0
              // Firebird-ODBC
              cSysDat += "\OdbcJdbc.dll"
              reg = new Registry(HKEY_LOCAL_MACHINE,cKey )
              reg.setValue("Client","")
              reg.setValue("User","")
              reg.setValue("Role","")
              reg.setValue("Driver",cSysDat)
              reg.setValue("Dbname",cAliasPfad)
              reg.setValue("CharacterSet","WIN1252")
              reg.setValue("ReadOnly","N")
              reg.setValue("NoWait","N")
              reg.setValue("Dialect","3")
              reg.setValue("QuotedIdentifier","Y")
              reg.setValue("Password",REPLICATE("A",80))
         OTHER
              // Nothing

      ENDCASE

      // Now user defined Values and Keys
      FOR odbcI = 1 TO aOptionen.size/4
          IF EMPTY (aOptionen[odbcI,2])
             LOOP
          ENDIF
          cKey =  "SOFTWARE\ODBC\ODBC.INI\" + cAliasName + ;
                  aOptionen[odbcI,1]
          reg = new Registry(HKEY_LOCAL_MACHINE,cKey )
          DO CASE
             CASE aOptionen[odbcI,4] = "C"
                  reg.setValue(aOptionen[odbcI,2],aOptionen[odbcI,3])

             CASE aOptionen[odbcI,4] = "N"
                  reg.setValue(aOptionen[odbcI,2],aOptionen[odbcI,3],REG_DWORD)

             // Added April 2007: Support for datatype REG_BINARY
       		 CASE aOptionen[odbcI,4] = "B"
						reg.setValue(aOptionen[odbcI,2],aOptionen[odbcI,3],REG_BINARY)

             OTHER
                  reg.setValue(aOptionen[odbcI,2],aOptionen[odbcI,3])
          ENDCASE
      ENDFOR

   CATCH (EXCEPTION E)
      return false 
   ENDTRY

   // Restart BDE
   // Thanks to Vic McClung for this routine in SetTableLevel.prg
   CLOSE DATABASES
   IF TYPE('DbiExit') # "FP"
      EXTERN cword DbiExit(cvoid) idapi32.dll
   ENDIF
   DbiExit()
   IF TYPE('DbiInitFn') # "FP"
      EXTERN cword DbiInitFn(cword, cptr) idapi32.dll
   ENDIF
   DbiInitFn(300, NULL)


return true
// EoF: odbc()
