/*
   ------------------------------------------------------------------
   Filename....: rcTimer.cc
   ClassName...: rcTimer
   Purpose.....: object for a high resolution timer
   Programmer..: Robert Cline
   Date........: January 18, 2003
   Notes.......: Should be run on a 386 or newer
                 Not all processors support
   Written for.: dBase 7.5 and above
   Rev. History: 
   Dependencies: None
   Calls.......: None
   Called by...: Any
   ------------------------------------------------------------------
   Usage.......: set procedure to :dUFLP:rcTimer.cc additive
                 oTimeChk = new RCTimer()
                 oTimeChk.GetStartCount()
                 // Do Some stuff
                 oTimeChk.GetEndCount()
                 ? oTimeChk.GetElapsed() + " seconds to run"
   ------------------------------------------------------------------
   Properties..:
      Freq:      High Performance Counter Ticks/Second
      Start:     Char Buffer of Start Point
      End:       Char Buffer of End Point
   Methods.....: 
      buf2DblLong()    Translates 4 char string to 64bit numeric
      getElapsed()     Calculates Time between Start and end
      getEndCount()    Gets the Finish Count
      getStartCount()  Gets the Start Count
      init(<char>)     Initialises needed info (no need to call)
*/
class RCTimer

   this.Freq  = 0
   this.Start = replicate(chr(0), 4)
   this.End   = replicate(chr(0), 4)
   this.init()
   //---------------------------------------------------------------//
   //  End Constructor
   //  Start Methods            
   //--------------------------------------------------------------//
   function Init
      local cBuffer

      if (TYPE("queryPerformanceFrequency") # "FP")
         extern	CLOGICAL QueryPerformanceFrequency(CPTR);
                           kernel32 from "QueryPerformanceFrequency"
      endif
      if (TYPE("queryPerformanceCounter") # "FP")
         extern	CLOGICAL QueryPerformanceCounter(CPTR);
                  kernel32 from "QueryPerformanceCounter"
      endif

      cBuffer = REPLICATE(CHR(0), 4)

      if (queryPerformanceFrequency(cBuffer))
         this.Freq = this.Buf2DblLong(cBuffer)
      else
         msgbox("High Performance Counter Not Supported", ;
                "rcTimer",;
                48)
      endif
   return
   //--------------------------------------------------------------//
   function GetStartCount
      queryPerformanceCounter(this.Start)
   return
   //--------------------------------------------------------------//
   function GetEndCount
      queryPerformanceCounter(this.End)
   return
   //--------------------------------------------------------------//
   function GetElapsed
      local nRet   

      nRet = 0
      if (this.Freq # 0)  // Verify there is a frequency (no /0)
         nRet = (this.Buf2DblLong(this.End) - ;
                 this.Buf2DblLong(this.Start);
                ) ;
                / this.Freq
      endif
   return(nRet)  // Elapsed Time or 0 if no frequency
   //--------------------------------------------------------------//
   function Buf2DblLong(cBuffer)
      local nHVal, nLVal

      // Get the High DWORD
      nHVal = BitLShift(cBuffer.GetByte(7), 24) + ;
              BitLShift(cBuffer.GetByte(6), 16) + ;
              BitLShift(cBuffer.GetByte(5), 08) + ;
              cBuffer.GetByte(4)
      // Get the Low DWORD
      nLVal = BitLShift(cBuffer.GetByte(3), 24) + ;
              BitLShift(cBuffer.GetByte(2), 16) + ;
              BitLShift(cBuffer.GetByte(1), 08) + ;
              cBuffer.GetByte(0)
   return (nHVal * 2 ^32 + nLVal)  // Return the value
endclass

