/*
   --------------------------------------------------------
   Filename.....: speedTip.cc
   ClassName....: speedTip
   Purpose......: SDI form for programmable speedtips.
                  Multi-line capable.
   Programmer...: Rick Miller -  http://www.treturn.com/dbase
   Date.........: January 04, 2005
   Notes........: See Notes below.
   Written for..: dBASE dB2K 0.1 and newer.
   Rev. History.: See Revisions below.
   Dependencies.: applicationIni.cc,
                  VStruct.prg or VStruct.cc or VStruct32Ex.cc,
                  and <VStruct.h>.
                  colorDlg.cc for use of selectColor().
                  fontDlg.cc for use of selectFont().
   Usage........: set procedure to speedTip.cc additive
   HelpFile.....: None.
   Example:.....: See Example below.
   SampleFile...: speedTip.wfm.
   --------------------------------------------------------
   Example:
   --------------------------------------------------------
      _app.speedTip  =  new speedTip()
      _app.speedTip.open()
      or
      form.speedTip  =  new speedTip()
      form.speedTip.open()
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
      When calling the set method:
         1) be sure the 2nd parameter is of proper type().
         2) passing an object property value is done by:
            ie:   set("padleft", int(form.sp.value))
            ie:   set("fontName", "" + form.ef.value)
      To reset properties to default values call reset().
      Protected methods begin with "x_".
      Protected properties begin with "~".
   --------------------------------------------------------
   Revisions:
   --------------------------------------------------------
      February 18, 2011
         1) default ini file adapted to following
            Windows security model and a standard User
            with the use of applicationIni.cc.
         2) added support for 2 monitors.
      March 11, 2005
         1) fix for preassignment of usersave property.
         2) increased alignment range to 5 so the speedTip
            can be displayed relative to:
            a) the mouse cursor (0 to 2).
            b) a GUI object (3 to 5).
      February 10, 2005
         1) renamed all external functions.
      January 19, 2005.
         1) added initialization of textSize.winColor
            in the textSize custom class.
   --------------------------------------------------------
   Keywords (parameter 1) for get and set methods:
   --------------------------------------------------------
      above, alignHorizontal, alignment, alignVertical,
      background, colorNormal, fontBold, fontColor,
      fontItalic, fontName, fontSize, fontStrikeout,
      fontUnderline, gutter, iniFile, mdiApp, padBottom,
      padHeight, padLeft, padRight, padTop, padWidth,
      text, time, usersave.
   --------------------------------------------------------
   Properties:
   --------------------------------------------------------
      iniFile           <char>   _app.exename.ini
      userSave          <bool>   TRUE
   --------------------------------------------------------
   Methods:
   --------------------------------------------------------
      get(<char> member)   returns a Keyword value.
      hide()               hide the speedTip.
      reset()              reset properties to default values.
      selectColor()        calls the colorDlg.
      selectFont()         calls the fontDlg.
      set(<char> member, <pValu> value)
                           set a Keyword value.
                           returns true/false.
      show(<int> hwnd)     displays the speedTip relative
                           to the GUI hwnd passed.
*/
// VStruct constants.
#include <VStruct.h>

// device parameters.
#define  LOGPIXELSX              88 /* Logical pixels/inch in X*/
#define  LOGPIXELSY              90 /* Logical pixels/inch in Y*/

// graphics modes.
#define  GM_ADVANCED             2
#define  GM_COMPATIBLE           1

#define  MWT_IDENTITY            1

// window adjustments.
#define  SWP_NOSIZE              0x0001
#define  SWP_NOMOVE              0x0002
#define  SWP_NOZORDER            0x0004
#define  SWP_NOACTIVATE          0x0010
#define  SWP_FRAMECHANGED        0x0020

// programmer functions.
#define  ghex_color(n)           right("000000" + ItoH(n), 6)
#define  zeros(n)                replicate(chr(0), n)
// programmer constants.
#define  SPEEDTIP_PUBLISHERINFO  "Rick Miller"
#define  SPEEDTIP_VERSIONINFO    "5.0311"
// bootstrap code when used with a do.
** END HEADER -- do not remove this line
//
// Generated on 01/04/2005
//
local f
f = new speedTip()
f.mdi = false // ensure not MDI
f.open()

class speedTip of Form
   with (this)
      canClose    =  class::form_canClose
      onMouseMove =  class::form_onMouseMove
      onOpen      =  class::form_onOpen
      onClose     =  class::form_onClose
      open        =  class::form_open
      metric      =  6  // Pixels
      colorNormal =  "InfoBk"
      height      =  18
      left        =  0
      top         =  -20
      width       =  500
      text        =  ""
      clientEdge  =  false
      topMost     =  true
      autoCenter  =  false
      mdi         =  false
      smallTitle  =  true
   endwith

   this.CT1 = new CONTAINER(this)
   with (this.CT1)
      height      =  0
      left        =  0
      top         =  0
      width       =  500
      transparent =  true
      colorNormal =  "InfoBk"
      borderStyle =  3 // None
      anchor      =  2  // Top
   endwith

   this.CT2 = new CONTAINER(this)
   with (this.CT2)
      height      =  18
      left        =  0
      top         =  0
      width       =  2
      transparent =  true
      colorNormal =  "InfoBk"
      borderStyle =  3  // None
      anchor      =  3  // Left
   endwith

   this.CT3 = new CONTAINER(this)
   with (this.CT3)
      height      =  18
      left        =  0
      top         =  1
      width       =  1
      transparent =  true
      colorNormal =  "InfoBk"
      borderStyle =  3  // None
      anchor      =  4  // Right
   endwith

   this.CT4 = new CONTAINER(this)
   with (this.CT4)
      height      =  0
      left        =  0
      top         =  0
      width       =  500
      transparent =  true
      colorNormal =  "InfoBk"
      borderStyle =  3  // None
      anchor      =  1  // Bottom
   endwith

   this.TX1 = new TEXTSIZE(this)
   with (this.TX1)
      onMouseMove =  {; this.parent.onMouseMove()}
      height      =  18
      left        =  2
      top         =  0
      width       =  497
      wrap        =  false
      transparent =  true
      text        =  "speedTip"
      fontname    =  iif(os().indexof("4.") == -1, "Tahoma", "MS Sans Serif")
      fontsize    =  8.0
      colorNormal =  "InfoText/InfoBk"
      anchor      =  6  // container
      alignVertical  =  0  // Top
      alignHorizontal   =  0  // Left
   endwith

   //------------------------------------------------------------//
   // this  =  form.
   function form_canClose
      if type("this.~timer") == "O"
         class::x_Timer_off()
         this.~timer.oparent  := null
         release object this.~timer
         this.~timer := null
      endif
      if (this.usersave)
         class::x_SaveIni()
      endif
   return   true

   //------------------------------------------------------------//
   // this  =  form.
   function form_onClose
      if type("this.~typeDef") == "O"
         release object this.~typeDef
         this.~typeDef   := null
      endif
      if type("this.~appIni") == "O"
         this.~appIni.close()
         this.~appIni   := null
      endif
      this.tx1.close()
      try
         this.release()
      catch(exception e)
      endtry
      class::x_CloseProcedure(program(1))
   return

   //------------------------------------------------------------//
   // this  =  form.
   function form_onMouseMove
      class::x_Timer_off()
   return

   //------------------------------------------------------------//
   // this  =  form.
   function form_onOpen
      // add timer object.
      this.~timer = new Timer()
      this.~timer.oParent  =  this
      this.~timer.interval =  1
      this.~timer.enabled  =  false

      if this.userSave
         class::x_LoadIni()
      endif

      Local nHwnd, nStyl
      nHwnd =  int(this.hwnd)
      // get current framestyle.
      nStyl =  RMM_GetWindowLong(nHwnd, -16)
      // change the framestyle.
      nStyl -= bitlshift(1, 22)  // dlgFrame.
      nStyl -= bitlshift(1, 18)  // thickFrame.
      RMM_SetWindowLong(nHwnd, -16, nStyl)
      RMM_SetWindowPos(nHwnd, 0, 0, 0, 0, 0,;
               SWP_FRAMECHANGED + SWP_NOZORDER +;
               SWP_NOACTIVATE + SWP_NOSIZE + SWP_NOMOVE)

      this.height := 18
   return

   //------------------------------------------------------------//
   // this  =  form.
   function form_open
      class::x_Protect()
      if not this.indesign
         this.visible   := false
         this.~formName =  "SpeedTip"
         class::x_Setup()
      endif
   return   super::open()

   //------------------------------------------------------------//
   // this  =  form.
   function get(cProp)
      Local cKey, pRet
      cKey  =  lower("" + cProp)
      pRet  =  ""
      if cKey == "above"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "alignment"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "alignhorizontal"
         pRet  := int(this.tx1.alignHorizontal)
      elseif cKey == "alignvertical"
         pRet  := int(this.tx1.alignVertical)
      elseif cKey == "background"
         pRet  := (this.backGround)
      elseif cKey == "colornormal"
         pRet  := (this.colorNormal)
      elseif cKey == "fontbold"
         pRet  := iif(this.tx1.fontBold, true, false)
      elseif cKey == "fontcolor"
         pRet  := int(this.tx1.winColor)
      elseif cKey == "fontitalic"
         pRet  := iif(this.tx1.fontItalic, true, false)
      elseif cKey == "fontname"
         pRet  := ("" + this.tx1.fontName)
      elseif cKey == "fontsize"
         pRet  := (0 + this.tx1.fontSize)
      elseif cKey == "fontstrikeout"
         pRet  := iif(this.tx1.fontStrikeout, true, false)
      elseif cKey == "fontunderline"
         pRet  := iif(this.tx1.fontUnderline, true, false)
      elseif cKey == "gutter"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "inifile"
         pRet  := ("" + this.inifile)
      elseif cKey == "mdiapp"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "padbottom"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "padleft"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "padright"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "padtop"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "padwidth"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "padheight"
         pRet  := this.~typeDef.getMem(cKey)
      elseif cKey == "text"
         pRet  := ("" + this.tx1.text)
      elseif cKey == "time"
         pRet  := int(this.~time)
      elseif cKey == "usersave"
         pRet  := iif(this.usersave, true, false)
      elseif cKey == "visible"
         pRet  := iif(this.visible, true, false)
      endif
   return   (pRet)

   //------------------------------------------------------------//
   // this  =  form.
   function hide
      class::x_Timer_off()
   return

   //------------------------------------------------------------//
   // this  =  form.
   function inspect
      inspect(this)
   return

   //------------------------------------------------------------//
   // this  =  form.
   function reset
      with (this.tx1)
         fontName       =  iif(os().indexof("4.") == -1, "Tahoma", "MS Sans Serif")
         fontSize       =  8.0
         fontBold       =  false
         fontItalic     =  false
         fontStrikeout  =  false
         fontUnderline  =  false
         colorNormal    =  "InfoText/InfoBk"
         alignVertical  =  0
         alignHorizontal   =  0
         wrap           =  false
      endwith
      this.~typeDef.setMem("above", false)
      this.~typeDef.setMem("mdiapp", true)
      this.~typeDef.setMem("alignment", 0)
      this.~typeDef.setMem("gutter", 0)
      this.~typeDef.setMem("padheight", 0)
      this.~typeDef.setMem("padwidth", 0)

      class::set("background", "")
      class::set("colornormal", "InfoBk")
      class::set("fontcolor", 0)
      class::set("inifile", "")
      class::set("padbottom", 0)
      class::set("padleft", 2)
      class::set("padright", 1)
      class::set("padtop", 0)
      class::set("time", 3)
      class::set("usersave", true)
      this.tx1.onFontChange()
   return

   //------------------------------------------------------------//
   // this  =  form.
   function selectColor
      Local bCont, bRet
      bCont =  true
      bRet  =  false
      try
         set procedure to colorDlg.cc additive
      catch(exception e)
         try
            set procedure to :dUFLP:colorDlg.cc additive
         catch(exception e)
            bCont := false
         endtry
      endtry
      if bCont
         Local oDlg
         oDlg  =  new colorDlg()
         // assign property values as desired.
         // to assign an initial color:
         oDlg.set("Color", "" + form.colorNormal)
         bRet  := oDlg.show()
         if bRet
            class::set("colornormal", int(oDlg.winColor))
         endif
         oDlg.release()
         oDlg  := null
      endif
   return   iif(bRet, true, false)

   //------------------------------------------------------------//
   // this  =  form.
   function selectFont
   return   this.tx1.selectFont()

   //------------------------------------------------------------//
   // this  =  form.
   function set(cProp, pValu)
      Local bRet, cKey, cType
      bRet  =  true
      cKey  =  lower("" + cProp)
      cType =  type("argVector(2)")
      if cKey.left(4) == "font"
         Local bChng
         bChng =  true
         if cType == "L"
            if cKey == "fontbold"
               this.tx1.fontBold := iif(pValu, true, false)
            elseif cKey == "fontitalic"
               this.tx1.fontItalic := iif(pValu, true, false)
            elseif cKey == "fontstrikeout"
               this.tx1.fontStrikeout := iif(pValu, true, false)
            elseif cKey == "fontunderline"
               this.tx1.fontUnderline := iif(pValu, true, false)
            else
               bChng := false
            endif
         elseif cType == "C"
            if cKey == "fontname" and not empty(pValu)
               this.tx1.fontName := pValu
            elseif cKey == "fontcolor" and not empty(pValu)
               this.tx1.colorNormal := "" + pValu + "/InfoBk"
               bChng := false
            else
               bChng := false
            endif
         elseif cType == "N"
            bChng := false
            if cKey == "fontcolor"
               this.tx1.winColor := int(pValu)
               this.tx1.colorNormal := ;
                        "0x" + ghex_color(pValu) + "/InfoBk"
            elseif cKey == "fontsize"
               if pValu > 5 and pValu < 25
                  this.tx1.fontSize := pValu
                  bChng := true
               endif
            endif
         else
            bChng := false
         endif
         if bChng
            this.tx1.onFontChange()
         elseif cKey == "fontcolor"
         else
            bRet  := false
         endif
      elseif cKey.left(3) == "pad"
         if cType == "N"
            if cKey == "padbottom"
               this.~typeDef.setMem(cKey, int(abs(mod(pValu, 100))))
               this.ct4.height   := this.~typeDef.getMem(cKey)
            elseif cKey == "padleft"
               this.~typeDef.setMem(cKey, int(abs(mod(pValu, 100))))
               this.ct2.width := this.~typeDef.getMem(cKey)
            elseif cKey == "padheight"
               this.~typeDef.setMem(cKey, int(abs(mod(pValu, 100))))
            elseif cKey == "padright"
               this.~typeDef.setMem(cKey, int(abs(mod(pValu, 100))))
               this.ct3.width := this.~typeDef.getMem(cKey)
            elseif cKey == "padtop"
               this.~typeDef.setMem(cKey, int(abs(mod(pValu, 100))))
               this.ct1.height   := this.~typeDef.getMem(cKey)
            elseif cKey == "padwidth"
               this.~typeDef.setMem(cKey, int(abs(mod(pValu, 100))))
            else
               bRet  := false
            endif
         else
            bRet  := false
         endif
      elseif cType == "C"
         if cKey == "background"
            this.backGround   := pValu
         elseif cKey == "colornormal"
            if empty(pValu)
               this.colorNormal  := "InfoBk"
            else
               this.colorNormal  := pValu
            endif
         elseif cKey == "inifile"
            Local cFold, cIni, nPos
            cIni  =  "" + pValu
            nPos  =  cIni.lastindexof("\")
            if nPos == -1
               cFold =  this.~appIni.getPath()
               cIni  := cFold + "\" + pValu
            else
               cFold =  cIni.left(nPos)
            endif
            bRet  := not empty(funique(cFold + "\????????"))
            if bRet
               this.iniFile   := "" + cIni
            endif
         elseif cKey == "text"
            this.tx1.text  := pValu
            this.tx1.onTextChange()
         else
            bRet  := false
         endif
      elseif cType == "L"
         if cKey == "above"
            this.~typeDef.setMem(cKey, pValu)
         elseif cKey == "mdiapp"
            this.~typeDef.setMem(cKey, pValu)
         elseif cKey == "usersave"
            this.usersave  := iif(pValu, true, false)
         else
            bRet  := false
         endif
      elseif cType == "N"
         if cKey == "alignment"
            this.~typeDef.setMem(cKey, int(abs(mod(pValu, 6))))
         elseif cKey == "alignhorizontal"
            this.tx1.alignHorizontal   := int(abs(mod(pValu, 3)))
         elseif cKey == "alignvertical"
            this.tx1.alignVertical  := int(abs(mod(pValu, 3)))
         elseif cKey == "colornormal"
            this.colorNormal := "0x" + ghex_color(pValu)
         elseif cKey == "gutter"
            this.~typeDef.setMem(cKey, int(mod(pValu, 100)))
         elseif cKey == "time"
            this.~time  =  int(mod(abs(pValu), 61))
            if this.~time == 0
               this.show   =  class::hide
            else
               this.show   =  class::show
            endif
         else
            bRet  := false
         endif
      else
         bRet  := false
      endif
   return   iif(bRet, true, false)

   //------------------------------------------------------------//
   // this  =  form.
   function show(nHwnd, bHwnd)
      if type("argVector(1)") == "N"
         if bHwnd
            bHwnd := this.~lastHwnd == nHwnd
         endif
         if not bHwnd
            this.~timer.enabled  := false
            this.~lastHwnd := nHwnd
            class::x_ShowForm(int(nHwnd))
            class::x_Timer_on()
            this.visible   := true
            this.topmost   := true
         endif
      endif
   return

   //------------------------------------------------------------//
   // this  =  form.
   function timer_onTimer
      this.~timecnt ++
      if this.~timecnt >= this.~time
         class::x_Timer_off()
      else
         if not this.~activeHwnd == RMM_GetActiveWindow()
            class::x_Timer_off()
         endif
      endif
   return

   //------------------------------------------------------------//
   // this  =  form.
   function x_CloseProcedure(cProc)
      try
         close procedure (cProc)
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   // this  =  form.
   function x_InitExterns
      if not type("RMM_GetActiveWindow") == "FP"
         extern   CHANDLE ;
                  RMM_GetActiveWindow(;
                  CVOID) ;
                  user32.dll from "GetActiveWindow"
      endif
      if not type("RMM_GetCapture") == "FP"
         extern   CHANDLE ;
                  RMM_GetCapture(;
                  CVOID) ;
                  user32 from "GetCapture"
      endif
      if not type("RMM_GetClientDesktopRect") == "FP"
         extern   CLOGICAL ;
                  RMM_GetClientDesktopRect(;
                  CUINT,;
                  CUINT,;
                  CPTR,;
                  CUINT) ;
                  user32 from "SystemParametersInfoA"
      endif
      if not type("RMM_GetSystemMetrics") == "FP"
         extern   CINT ;
                  RMM_GetSystemMetrics(;
                  CINT) ;
                  user32 from "GetSystemMetrics"
      endif
      if not type("RMM_GetWindowLong") == "FP"
         extern   CLONG ;
                  RMM_GetWindowLong(;
                  CHANDLE,;
                  CINT) ;
                  user32 from "GetWindowLongA"
      endif
      if not type("RMM_GetWindowRect") == "FP"
         extern   CLOGICAL ;
                  RMM_GetWindowRect(;
                  CHANDLE,;
                  CPTR) ;
                  user32 from "GetWindowRect"
      endif
      if not type("RMM_MoveWindow") == "FP"
         extern   CLOGICAL ;
                  RMM_MoveWindow(;
                  CHANDLE,;
                  CINT,;
                  CINT,;
                  CINT,;
                  CINT,;
                  CLOGICAL) ;
                  user32 from "MoveWindow"
      endif
      if not type("RMM_ReleaseCapture") == "FP"
         extern   CLOGICAL ;
                  RMM_ReleaseCapture(;
                  CVOID) ;
                  user32 from "ReleaseCapture"
      endif
      if not type("RMM_SetCapture") == "FP"
         extern   CHANDLE ;
                  RMM_SetCapture(;
                  CHANDLE) ;
                  user32 from "SetCapture"
      endif
      if not type("RMM_SetWindowLong") == "FP"
         extern   CLONG ;
                  RMM_SetWindowLong(;
                  CHANDLE,;
                  CINT,;
                  CLONG) ;
                  user32 from "SetWindowLongA"
      endif
      if not type("RMM_SetWindowPos") == "FP"
         extern   CLOGICAL ;
                  RMM_SetWindowPos(;
                  CHANDLE,;
                  CHANDLE,;
                  CINT,;
                  CINT,;
                  CINT,;
                  CINT,;
                  CUINT) ;
                  user32 from "SetWindowPos"
      endif
   return

   //------------------------------------------------------------//
   // this  =  form.
   // this method is called from the form_onOpen method.
   function x_LoadIni
      // retrieve values from the user's ini File.
      // direct assignments are made when a fontChange
      // would be made. a fontChange will take place at the
      // end of the form_onOpen.
      Local iVal
      class::set("above", iif(class::x_ReadIni("Above") == "1", true, false))
      class::set("alignment", val(class::x_ReadIni("Alignment")))
      class::set("alignhorizontal", val(class::x_ReadIni("AlignHorizontal")))
      class::set("alignvertical", val(class::x_ReadIni("AlignVertical")))
      class::set("background", class::x_ReadIni("BackGround"))
      class::set("colornormal", class::x_ReadIni("ColorNormal"))
      iVal  =  val(class::x_ReadIni("FontBold"))
      if iVal > 0
         this.tx1.fontBold := true
      endif
      class::set("fontcolor", val(class::x_ReadIni("FontColor")))
      iVal  =  val(class::x_ReadIni("FontItalic"))
      if iVal > 0
         this.tx1.fontItalic  := true
      endif
      iVal  =  class::x_ReadIni("FontName")
      if not empty(ltrim(trim(iVal)))
         this.tx1.fontName := iVal
      endif
      iVal  =  val(class::x_ReadIni("FontSize"))
      if iVal > 5 and iVal < 73
         this.tx1.fontSize := iVal
      endif
      iVal  =  val(class::x_ReadIni("FontStrikeout"))
      if iVal > 0
         this.tx1.fontStrikeout  := true
      endif
      iVal  =  val(class::x_ReadIni("FontUnderline"))
      if iVal > 0
         this.tx1.fontUnderline  := true
      endif
      class::set("gutter", val(class::x_ReadIni("Gutter")))
      class::set("mdiapp", iif(class::x_ReadIni("MdiApp") == "0", false, true))
      class::set("padbottom", val(class::x_ReadIni("PadBottom")))
      class::set("padheight", val(class::x_ReadIni("PadHeight")))
      iVal  =  class::x_ReadIni("PadLeft")
      class::set("padleft", iif(empty(iVal), 2, val(iVal)))
      iVal  =  class::x_ReadIni("PadRight")
      class::set("padright",iif(empty(iVal), 1, val(iVal)))
      class::set("padtop", val(class::x_ReadIni("PadTop")))
      class::set("padwidth", val(class::x_ReadIni("PadWidth")))
      iVal  =  class::x_ReadIni("Time")
      class::set("time", iif(empty(iVal), 3, val(iVal)))
   return

   //------------------------------------------------------------//
   // this  =  form.
   function x_Protect
      // protected methods.
      protect  x_CloseProcedure, x_InitExterns,;
               x_LoadIni, x_Protect, x_ReadIni, x_SaveIni,;
               x_Setup, x_ShowForm, x_Timer_off, x_Timer_on,;
               x_WriteIni
      // protected properties.
      protect  ~activeHwnd, ~appIni, ~formName, ~ini, ~lastHwnd,;
               ~runtime, ~time, ~timecnt, ~timer, ~typeDef,;
               ct1, ct2, ct3, ct4, tx1
   return

   //------------------------------------------------------------//
   // this  =  form.
   // read a value from an ini file.
   function x_ReadIni(cItem)
   return   this.~appIni.getValue(;
            this.~formName,;
            cItem,;
            this.iniFile)

   //------------------------------------------------------------//
   // save values to the user's ini File.
   // this  =  form.
   function x_SaveIni
      if this.~appIni.fileCheck(this.iniFile)
         //  save settings
         class::x_WriteIni("Above", iif(this.~typeDef.getMem('above'), "1", "0"))
         class::x_WriteIni("Alignment", "" + int(this.~typeDef.getMem('alignment')))
         class::x_WriteIni("AlignHorizontal", "" + this.tx1.alignHorizontal)
         class::x_WriteIni("AlignVertical", "" + this.tx1.alignVertical)
         class::x_WriteIni("BackGround", "" + this.backGround)
         class::x_WriteIni("ColorNormal", "" + this.colorNormal)
         class::x_WriteIni("FontBold", iif(this.tx1.fontBold, "1", "0"))
         class::x_WriteIni("FontColor", "" + int(this.tx1.winColor))
         class::x_WriteIni("FontItalic", iif(this.tx1.fontItalic, "1", "0"))
         class::x_WriteIni("FontName", "" + this.tx1.fontName)
         class::x_WriteIni("FontSize", "" + this.tx1.fontSize)
         class::x_WriteIni("FontStrikeout", iif(this.tx1.fontStrikeout, "1", "0"))
         class::x_WriteIni("FontUnderline", iif(this.tx1.fontUnderline, "1", "0"))
         class::x_WriteIni("Gutter", "" + int(this.~typeDef.getMem('gutter')))
         class::x_WriteIni("MdiApp", iif(this.~typeDef.getMem('mdiApp'), "1", "0"))
         class::x_WriteIni("PadBottom", "" + int(this.~typeDef.getMem('padBottom')))
         class::x_WriteIni("PadHeight", "" + int(this.~typeDef.getMem('padHeight')))
         class::x_WriteIni("PadLeft", "" + int(this.~typeDef.getMem('padLeft')))
         class::x_WriteIni("PadRight", "" + int(this.~typeDef.getMem('padRight')))
         class::x_WriteIni("PadTop", "" + int(this.~typeDef.getMem('padTop')))
         class::x_WriteIni("PadWidth", "" + int(this.~typeDef.getMem('padWidth')))
         class::x_WriteIni("Time", "" + int(this.~time))
      endif
   return

   //------------------------------------------------------------//
   // this  =  form.
   function x_Setup
      class::x_InitExterns()
      try
         set procedure to applicationIni.cc additive
      catch(exception e)
         try
            set procedure to :dUFLP:applicationIni.cc additive
         catch(exception e)
         endtry
      endtry
      this.tx1.winColor =  0
      this.~activeHwnd  =  0
      this.~lastHwnd    =  0
      this.~time        =  3
      this.~timecnt     =  0
      this.~typeDef     =  SPEEDTIP_Structures::speedTipStructure()
      this.~runtime  =  iif( ;
               version(0).toLowerCase().indexof("run") == -1,;
               false, true)
      this.~appIni   =  new applicationIni()
      this.iniFile   =  this.~appIni.value
      // to disable using an ini file,
      // assign the usersave (property) false before
      // the open method is called.
      if not type("this.usersave") == "L"
         this.usersave     =  true     // save settings
      endif
      this.versioninfo  =  SPEEDTIP_VERSIONINFO
      this.publishinfo  =  SPEEDTIP_PUBLISHERINFO
   return

   //------------------------------------------------------------//
   // this  =  form.
   function x_ShowForm(nHwnd)

      Local aText, nAdjh, nAdjv, nXcur, nYcur,;
               oHigh, oWide, oAbov, oAlig, oGut, oMdi, oPcx, oPcy,;
               sCurs, sRect, sType
      sCurs =  SPEEDTIP_Structures::pointStructure()
      RMM_GetCursorPos(sCurs)
      nXcur =  sCurs.getMem('x')
      nYcur =  sCurs.getMem('y')
      sRect =  SPEEDTIP_Structures::rectStructure()
      sType =  this.~typeDef

      nAdjh =  RMM_GetSystemMetrics(5)
      nAdjv =  RMM_GetSystemMetrics(6)

      oAbov =  sType.getMem('above')
      oMdi  =  sType.getMem('mdiapp')
      oAlgn =  sType.getMem('alignment')
      oGut  =  sType.getMem('gutter')
      oPcx  =  sType.getMem('padwidth')
      oPcy  =  sType.getMem('padheight')

      // add on width for the frames, ct2, and the width padding.
      oWide =  int(this.tx1.wide)
      oWide += nAdjh * 2 +;
               (this.ct2.width + this.ct3.width + oPcx)
      // calc the height for the number of strings.
      oHigh =  int(this.tx1.high)
      // add on height for the frames, ct1, and the height padding.
      oHigh +=  nAdjv * 2 +;
               (this.ct1.height + this.ct4.height + oPcy)

      Local nAlef, nAtop, nArig, nAbot
      // get the bounding rectangle to use.
      if oMdi
         // use the framewin for a boundary.
         RMM_GetWindowRect(_app.framewin.hwnd, sRect)
         nAdjh += RMM_GetSystemMetrics(32)
         nAdjv += RMM_GetSystemMetrics(33)
         if _app.framewin.windowState == 2
            nAdjh += RMM_GetSystemMetrics(5)
            nAdjv += RMM_GetSystemMetrics(6)
         endif
      else
         // use the desktop for a boundary.
         RMM_GetClientDesktopRect(48, 0, sRect, 0)
      endif

      // define the boundaries for display.
      nAlef =  int(sRect.getMem('left')   + nAdjh)
      nAtop =  int(sRect.getMem('top')    + nAdjv)
      nArig =  int(sRect.getMem('right')  - nAdjh)
      nAbot =  int(sRect.getMem('bottom') - nAdjv)

      Local nHtop, nHbot, nHlef, oLeft, oTop
      // position with the passed hWnd (nHwnd).
      sRect.string   := zeros(sRect.getSize())
      RMM_GetWindowRect(nHwnd, sRect)
      nHtop =  int(sRect.getMem('top'))
      nHbot =  int(sRect.getMem('bottom'))
      nHlef =  int(sRect.getMem('left'))

      // get the top position.
      if oAbov
         oTop  =  int(nHtop - oHigh - oGut)
      else
         oTop  =  int(nHbot + oGut)
      endif

      // get the left position.
      if oAlgn == 1
         oLeft =  int(nXcur - (oWide / 2))
      elseif oAlgn == 2
         oLeft =  int(nXcur - oWide)
      elseif oAlgn == 3
         oLeft =  int(sRect.getMem('left'))
      elseif oAlgn == 4
         oLeft =  ;
                  int(sRect.getMem('left') + ;
                  ((sRect.getMem('right') - sRect.getMem('left')) / 2) -;
                  (oWide / 2))
      elseif oAlgn == 5
         oLeft =  int(sRect.getMem('right') - oWide)
      else
         oLeft =  int(nXcur + iif(oAbov, 1, 6))
      endif

      if not oMdi
         if RMM_GetSystemMetrics(80) > 1
            Local    nPvx, nPvy, nPx, nPy
            // get primary screen size.
            // primary screen width.
            nPx   =  RMM_GetSystemMetrics(0)
            // primary screen height.
            nPy   =  RMM_GetSystemMetrics(1)
            // get the total virtual size.
            // the virtual width is the total of the 2 screens.
            // virtual width.
            nPvx  =  RMM_GetSystemMetrics(78)
            // the virtual height is the larger of the 2 screens.
            nPvy  =  RMM_GetSystemMetrics(79)
            // determine which screen is in use.
            // get hWndParent position.
            if nHlef > nPx - RMM_GetSystemMetrics(32)
               // using the 2nd screen.
               // the 2nd screen is positioned to the right.
               nAlef := int(nPx)
               nArig := int(nPvx)
               nAbot := int(nPvy)
            endif
         endif
      endif

      // insure the form is on the framewin or desktop.
      // re-assign the top or left if necessary.
      if (oTop + oHigh) > nAbot
         // will be too low.
         oTop  := (nAbot - oHigh)
      endif
      if oTop < nAtop
         // will be too high.
         oTop  := nAtop
      endif
      if (oLeft + oWide) > nArig
         // will be too far right.
         oLeft := (nArig - oWide)
      endif
      if oLeft < nAlef
         // will be too far left.
         oLeft := nAlef
      endif

      // insure the form does not cover the cursor.
      if nXcur >= oLeft and nXcur <= oLeft + oWide and ;
                  nYcur >= oTop and nYcur <= oTop + oHigh
         // will be covering the cursor or object.
         // move the top position.
         if oAbov
            oTop  := nHtop - oHigh
         else
            oTop  := nHbot
         endif
      endif

      // display form to user.
      RMM_MoveWindow(this.hwnd, oLeft, oTop, oWide, oHigh, true)
      // the next assignment re-sizes form objects
      // that are using the anchor property.
      this.height := oHigh

   return

   //------------------------------------------------------------//
   // this  =  form.
   function x_Timer_off
      try
         this.~timer.enabled  := false
         this.visible   := false
         this.topmost   := false
         this.~lastHwnd := 0
      catch(exception e)
      endtry
   return   iif(this.topmost, false, true)

   //------------------------------------------------------------//
   // this  =  form.
   function x_Timer_on
      this.~timecnt        := 0
      this.~activeHwnd     := RMM_GetActiveWindow()
      this.~timer.onTimer  := {; this.oparent.timer_onTimer()}
      this.~timer.enabled  := true
   return

   //------------------------------------------------------------//
   // this  =  form.
   // write value to ini file.
   function x_WriteIni(cItem, cValu)
   return   this.~appIni.setValue(;
            this.~formName,;
            cItem,;
            cValu,;
            this.iniFile)

endclass
//---------------------------------------------------------------//
//          end of speedTip
//---------------------------------------------------------------//
/*
   --------------------------------------------------------
   ClassName....: textSize
   Purpose......: To determine the size required for a text
                  object to display a single or multi-line
                  piece of text.
   Programmer...: Rick Miller -  http://www.treturn.com/dbase
   Date.........: January 04, 2005
   Notes........: See Notes below.
   Written for..: dBASE dB2K 0.1 and newer.
   Rev. History.: None.
   Dependencies.: VStruct.prg or VStruct.cc or VStruct32Ex.cc,
                  and VStruct.h.
                  :dUFLP:fontDlg.cc if using selectFont().
   Usage........: set procedure to :dUFLP:speedTip.cc additive
   HelpFile.....: None.
   Example:.....: See Example below.
   SampleFile...: None.
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
      Parsing takes place when passing a char
         to getTextExtent.
      The characters to use for multi-line separators
         are either chr(13) or chr(10)
      No parsing takes place when passing an array
         to getTextExtent.
   --------------------------------------------------------
   Example:
   --------------------------------------------------------
      form.textSize.text   := cVariable
      form.textSize.onTextChange()
      form.textSize.applyChange()
      or
      form.textSize.fontSize  := 14.0
      form.textSize.onFontChange()
      form.textSize.applyChange()
*/
class textSize(parentObj) of text(parentObj) custom
   with( this)
      onOpen         =  class::onOpen
      getTextExtent  =  class::getTextExtent
      fontName       =  "Arial"
      fontSize       =  10.0
      colorNormal    =  "BtnText/BtnFace"
      wrap           =  false
   endwith

   this.~lineWidth   =  0
   this.~lineHeight  =  16
   this.~procFile    =  class::x_GetProcFile()

   this.high   =  22
   this.wide   =  86
   this.winColor  =  0

   class::x_Protect()

   //------------------------------------------------------------//
   //       end of constructor   =  methods below.
   //------------------------------------------------------------//
   // call after a font or text change to have this
   // resized to accomodate the text property.
   function applyChange
      this.height := int(this.high)
      this.width  := int(this.wide)
   return

   //------------------------------------------------------------//
   function close
      class::x_CloseProcedure(this.~procFile)
      class::x_CloseProcedure(program(1))
   return

   //------------------------------------------------------------//
   // accepts an array, string, or nothing.
   function getTextExtent(cText)
      Local aText, i
      aText =  new array()
      if type("argVector(1)") == "A"
         for i = 1 to cText.size
            aText.add("" + cText[i])
         endfor
      else
         if not type("argVector(1)") == "C"
            cText := "" + this.text
         endif
         if empty(cText)
         elseif not cText.indexof(chr(13)) == -1
            aText := class::x_ParseText(cText, chr(13))
         elseif not cText.indexof(chr(10)) == -1
            aText := class::x_ParseText(cText, chr(10))
         else
            aText.add(cText)
         endif
      endif

      Local nDc, nFont, nHwnd, nLog, nRet, nTest
      nHwnd =  int(this.hwnd)
      nDc   =  RMM_GetDC(nHwnd)

      // create a logFont using the saved string.
      // returns the handle.
      nLog  =  RMM_CreateFontIndirect(this.~cLog)

      // put the logFont at nLog into the dc.
      // save the current logFont handle into nFont.
      nFont =  RMM_SelectObject(nDc, nLog)

      // get the length in pixels of the widest string in aText.
      nRet  =  0
      for i = 1 to aText.size
         nTest =  class::x_GetWidth(nDc, aText[i])
         if nTest > nRet
            nRet  := nTest
         endif
      endfor

      // select the logFont at nFont into the dc.
      // destroy the logFont at nLog.
      RMM_DeleteObject(RMM_SelectObject(nDc, nFont))
      RMM_ReleaseDC(nHwnd, nDc)

      // destroy font handles.
      nLog  := null
      nFont := null

      this.high   := int(ceiling(aText.size * this.~lineHeight)) +;
               class::x_GetBorderHeight()
      release object aText
      aText := null

   return   int(ceiling(nRet))

   //------------------------------------------------------------//
   // call when a change to this.font???? takes place.
   function onFontChange
      class::x_Initialize()
      class::onTextChange()
   return

   //------------------------------------------------------------//
   function onOpen
      class::x_InitExterns()
      class::onFontChange()
      class::applyChange()
   return

   //------------------------------------------------------------//
   // call when a change to this.text takes place.
   function onTextChange
      this.wide   := ceiling(class::getTextExtent() +;
               class::x_GetBorderWidth())
   return

   //------------------------------------------------------------//
   function selectFont
      Local bCont, bRet
      bCont =  true
      bRet  =  false
      try
         set procedure to fontDlg.cc additive
      catch(exception e)
         try
            set procedure to :dUFLP:fontDlg.cc additive
         catch(exception e)
            bCont := false
         endtry
      endtry
      if bCont
         Local bRels, cBg, cClr, cFg, oDlg
         cClr  =  "" + this.colorNormal
         cFg   =  cClr.left(cClr.indexof("/"))
         cBg   =  cClr.substring(cFg.length, cClr.length)
         oDlg  =  new fontDlg()
         oDlg.colorStyle   := 3
         oDlg.setColor(cFg)
         oDlg.fontName  := "" + this.fontName
         oDlg.bold      := iif(this.fontBold, true, false)
         oDlg.italic    := iif(this.fontItalic, true, false)
         oDlg.strikeout := iif(this.fontStrikeout, true, false)
         oDlg.underline := iif(this.fontUnderline, true, false)
         oDlg.size      := 0 + this.fontSize
         bRet  =  oDlg.show()
         if bRet
            with (this)
               fontName       := oDlg.fontName
               fontSize       := oDlg.size
               fontBold       := oDlg.bold
               fontItalic     := oDlg.italic
               fontStrikeout  := oDlg.strikeout
               fontUnderline  := oDlg.underline
               colorNormal    := oDlg.color + cBg
            endwith
            this.winColor  := int(oDlg.winColor)
         endif
         oDlg.release()
         oDlg  := null
         if bRet
            class::onFontChange()
         endif
      endif
   return   iif(bRet, true, false)

   //------------------------------------------------------------//
   function x_CloseProcedure(cProc)
      try
         close procedure (cProc)
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   function x_GetBorderHeight
      Local nRet
      nRet  =  0
      if this.borderStyle == 0      // Default
      elseif empty(this.text)
      elseif this.borderStyle == 1  // Raised
         nRet  := 4
      elseif this.borderStyle == 2  // Lowered
         nRet  := 4
      elseif this.borderStyle == 3  // None
      elseif this.borderStyle == 4  // Single
         nRet  := 4
      elseif this.borderStyle == 5  // Double
         nRet  := 12
      elseif this.borderStyle == 6  // Drop shadow
         nRet  := 8
      elseif this.borderStyle == 7  // Client
         nRet  := 5
      elseif this.borderStyle == 8  // Modal
         nRet  := 5
      elseif this.borderStyle == 9  // Etched in
         nRet  := 5
      elseif this.borderStyle == 10 // Etched out
         nRet  := 4
      endif
   return   int(nRet)

   //------------------------------------------------------------//
   function x_GetBorderWidth
      Local nRet
      nRet  =  0
      if this.borderStyle == 0      // Default
      elseif empty(this.text)
      elseif this.borderStyle == 1  // Raised
         nRet  := 4
      elseif this.borderStyle == 2  // Lowered
         nRet  := 4
      elseif this.borderStyle == 3  // None
      elseif this.borderStyle == 4  // Single
         nRet  := 4
      elseif this.borderStyle == 5  // Double
         nRet  := 17
      elseif this.borderStyle == 6  // Drop shadow
         nRet  := 8
      elseif this.borderStyle == 7  // Client
         nRet  := 13
      elseif this.borderStyle == 8  // Modal
         nRet  := 13
      elseif this.borderStyle == 9  // Etched in
         nRet  := 13
      elseif this.borderStyle == 10 // Etched out
         nRet  := 13
      endif
   return int(nRet)

   //------------------------------------------------------------//
   function x_GetLong(cStru, nByte)
      Local nRet
      nRet  =  cStru.getbyte(nByte) + ;
               BitLshift(cStru.getbyte(nByte + 1), 8)  + ;
               BitLshift(cStru.getbyte(nByte + 2), 16) + ;
               BitLshift(cStru.getbyte(nByte + 3), 24)
      if BitSet(nRet, 31)
         nRet  := - BitNot(nRet)
      endif
   return   nRet

   //------------------------------------------------------------//
   function x_GetProcfile
      Local cRet
      cRet  =  ""
      try
         set procedure to VStruct.cc additive
         cRet  := "VStruct.cc"
      catch(exception e)
         try
            set procedure to VStruct32Ex.cc additive
            cRet  := "VStruct32Ex.cc"
         catch(exception e)
            try
               set procedure to VStruct.prg additive
               cRet  := "VStruct.prg"
            catch(exception e)
               try
                  set procedure to :dUFLP:VStruct.prg additive
                  cRet  := ":dUFLP:VStruct.prg"
               catch(exception e)
                  cRet  := ""
               endtry
            endtry
         endtry
      endtry
   return   cRet

   //------------------------------------------------------------//
   function x_GetWidth(nDc, cText)
      Local cBuff, nRet
      // put the length of the string into a structure.
      nRet  =  0
      for i = 0 to cText.length - 1
         cBuff =  zeros(8)
         RMM_GetTextExtentPoint32( ;
                  nDc, cText.substring(i, i + 1), 1, cBuff)
         nRet  +=  class::x_GetLong(cBuff, 0)
         cBuff := null
      endfor
   return   int(nRet)

   //------------------------------------------------------------//
   function x_InitExterns
      if not type("RMM_CreateFontIndirect") == "FP"
         extern   CHANDLE ;
                  RMM_CreateFontIndirect(;
                  CPTR) ;
                  gdi32 from "CreateFontIndirectA"
      endif
      if not type("RMM_DeleteObject") == "FP"
         extern   CLOGICAL ;
                  RMM_DeleteObject(;
                  CHANDLE) ;
                  gdi32 from "DeleteObject"
      endif
      if not type("RMM_DPtoLP") == "FP"
         extern   CLOGICAL ;
                  RMM_DPtoLP(;
                  CHANDLE,;
                  CPTR,;
                  CINT) ;
                  gdi32 from "DPtoLP"
      endif
      if not type("RMM_GetCursor") == "FP"
         extern   CHANDLE ;
                  RMM_GetCursor(;
                  CVOID) ;
                  user32 from "GetCursor"
      endif
      if not type("RMM_GetCursorPos") == "FP"
         extern   CLOGICAL ;
                  RMM_GetCursorPos(;
                  CPTR) ;
                  user32 from "GetCursorPos"
      endif
      if not type("RMM_GetDC") == "FP"
         extern   CHANDLE ;
                  RMM_GetDC(;
                  CHANDLE) ;
                  user32 from "GetDC"
      endif
      if not type("RMM_GetDeviceCaps") == "FP"
         extern   CINT ;
                  RMM_GetDeviceCaps(;
                  CHANDLE,;
                  CINT) ;
                  gdi32 from "GetDeviceCaps"
      endif
      if not type("RMM_GetTextCharset") == "FP"
         extern   CINT ;
                  RMM_GetTextCharset(;
                  CHANDLE) ;
                  gdi32 from "GetTextCharset"
      endif
      if not type("RMM_GetTextExtentPoint32") == "FP"
         // handle of device context,
         // address of text string,
         // number of characters in string,
         // address of structure for string size
         extern   CLOGICAL ;
                  RMM_GetTextExtentPoint32(;
                  CHANDLE,;
                  CSTRING,;
                  CINT,;
                  CPTR) ;
                  gdi32 from "GetTextExtentPoint32A"
      endif
      if not type("RMM_GetTextMetrics") == "FP"
         extern   CLOGICAL ;
                  RMM_GetTextMetrics(;
                  CHANDLE,;
                  CPTR) ;
                  gdi32 from "GetTextMetricsA"
      endif
      if not type("RMM_ModifyWorldTransform") == "FP"
         extern   CLOGICAL ;
                  RMM_ModifyWorldTransform(;
                  CHANDLE,;
                  CPTR,;
                  CULONG) ;
                  gdi32 from "ModifyWorldTransform"
      endif
      if not type("RMM_PtInRect") == "FP"
         extern   CLOGICAL ;
                  RMM_PtInRect(;
                  CPTR,;
                  CPTR) ;
                  user32 from "PtInRect"
      endif
      if not type("RMM_ReleaseDC") == "FP"
         extern   CINT ;
                  RMM_ReleaseDC(;
                  CHANDLE,;
                  CHANDLE) ;
                  user32 from "ReleaseDC"
      endif
      if not type("RMM_RestoreDC") == "FP"
         extern   CLOGICAL ;
                  RMM_RestoreDC(;
                  CHANDLE,;
                  CINT) ;
                  gdi32 from "RestoreDC"
      endif
      if not type("RMM_SaveDC") == "FP"
         extern   CINT ;
                  RMM_SaveDC(;
                  CHANDLE) ;
                  gdi32 from "SaveDC"
      endif
      if not type("RMM_SelectObject") = "FP"
         extern   CHANDLE ;
                  RMM_SelectObject(;
                  CHANDLE,;
                  CHANDLE) ;
                  gdi32 from "SelectObject"
      endif
      if not type("RMM_SetGraphicsMode") == "FP"
         extern   CINT ;
                  RMM_SetGraphicsMode(;
                  CHANDLE,;
                  CINT) ;
                  gdi32 from "SetGraphicsMode"
      endif
      if not type("RMM_SetViewportOrgEx") == "FP"
         extern   CLOGICAL ;
                  RMM_SetViewportOrgEx(;
                  CHANDLE,;
                  CINT,;
                  CINT,;
                  CPTR) ;
                  gdi32 from "SetViewportOrgEx"
      endif
      if not type("RMM_SetWindowOrgEx") == "FP"
         extern   CLOGICAL ;
                  RMM_SetWindowOrgEx(;
                  CHANDLE,;
                  CINT,;
                  CINT,;
                  CPTR) ;
                  gdi32 from "SetWindowOrgEx"
      endif
   return

   //------------------------------------------------------------//
   function x_Initialize
      Local cName, nChr, nDc, nFont, nHwnd, nLog, nSize,;
               nXdpi, nYdpi, oLog, oMetr, oPt
      cName =  "" + this.fontName
      nHwnd =  int(this.hwnd)
      nSize =  (10 * this.fontSize)
      nDc   =  RMM_GetDC(nHwnd)
      nChr  =  RMM_GetTextCharset(nDc)
      RMM_SaveDC(nDc)
      // update the dc for WinNT.
      RMM_SetGraphicsMode(nDc, GM_ADVANCED)
      RMM_ModifyWorldTransform(nDc, 0, MWT_IDENTITY)
      RMM_SetViewportOrgEx(nDc, 0, 0, 0)
      RMM_SetWindowOrgEx(nDc, 0, 0, 0)

      // get the Decimal values for the dc.
      nXdpi =  RMM_GetDeviceCaps(nDc, LOGPIXELSX)
      nYdpi =  RMM_GetDeviceCaps(nDc, LOGPIXELSY)

      // create a point structure that will be Decimal.
      oPt   =  SPEEDTIP_Structures::pointStructure()
      // fill the point structure.
      oPt.setMem('x', int(ceiling(nSize * nXdpi / 72 + 0.5)))
      oPt.setMem('y', int(ceiling(nSize * nYdpi / 72 + 0.5)))

      // convert the point structure from Decimal to Logical values
      // so it can be placed into the logFont structure.
      RMM_DPtoLP(nDc, oPt, 1)

      // create a logFont structure.
      // cFont =  zeros(60)
      oLog  =  SPEEDTIP_Structures::logFontStructure()
      // Update the lfHeight, lfWeight, lfItalic,
      // lfUnderline, lfStrikeout, lfFaceName, lfCharSet members.
      oLog.setMem('lfHeight', - int(oPt.getMem('y') / 10 + 0.5))
      if this.fontBold
         oLog.setMem('lfWeight', 700)
      endif
      if this.fontItalic
         oLog.setMem('lfItalic', 255)
      endif
      if this.fontUnderline
         oLog.setMem('lfUnderline', 1)
      endif
      if this.fontStrikeout
         oLog.setMem('lfStrikeout', 1)
      endif
      oLog.setMem('lfFaceName', cName + chr(0))
      oLog.setMem('lfCharSet', nChr)

      // save the string for re-use.
      this.~cLog  =  oLog.string

      // create a logFont returning the handle.
      nLog  =  RMM_CreateFontIndirect(oLog)
      // put the logFont at nLog into the dc.
      // save the current logFont handle into nFont.
      nFont =  RMM_SelectObject(nDc, nLog)

      // create a textmetric structure.
      oMetr =  SPEEDTIP_Structures::textMetricStructure()
      // fill the textmetric structure.
      RMM_GetTextMetrics(nDc, oMetr)

      // save the tmHeight + tmExternalLeading for re-use.
      this.~lineHeight  := int(ceiling( ;
               oMetr.getMem('tmHeight') +;
               oMetr.getMem('tmExternalLeading')))

      // select the logFont at nFont into the dc.
      // destroy the logFont at nLog.
      RMM_DeleteObject(RMM_SelectObject(nDc, nFont))

      // restore the original dc.
      RMM_RestoreDC(nDc, -1)
      RMM_ReleaseDC(nHwnd, nDc)

      release object oLog
      release object oMetr
      release object oPt
      // destroy structures.
      oLog  := null
      oMetr := null
      oPt   := null

      // destroy font handles.
      nLog  := null
      nFont := null
   return

   //------------------------------------------------------------//
   // parses text that uses a chr(?) for line separation
   // into an array that is returned.
   function x_ParseText(cText, cSep)
      Local aRet, cTmp, nPos1, nPos2
      aRet  =  new array()
      cTmp  =  cText.toLowerCase()
      nPos1 =  0
      for i = 1 to 100
         nPos2 =  cTmp.indexof(cSep, nPos1)
         if nPos2 == -1
            aRet.add(cText.substring(nPos1, cText.length))
            i  := 105
         else
            aRet.add(cText.substring(nPos1, nPos2))
            nPos1 := nPos2 + cSep.length
         endif
      endfor
   return   aRet

   //------------------------------------------------------------//
   function x_Protect
      // protected properties.
      protect  ~cLog, ~lineHeight, ~lineWidth, ~procFile

      // protected methods.
      protect  onOpen, x_CloseProcedure, x_GetBorderHeight,;
               x_GetBorderWidth, x_GetLong, x_GetProcFile,;
               x_GetWidth, x_InitExterns, x_Initialize,;
               x_ParseText, x_Protect
   return

endclass
//---------------------------------------------------------------//
//          end of textSize.
//---------------------------------------------------------------//
class SPEEDTIP_Structures

   //------------------------------------------------------------//
   function logFontStructure
      Local sStruct
      if not VSTRUCT_ISDEF("LOGFONT")
         VSTRUCT_DEF('LOGFONT',;
                  'lfHeight',          TYPE_LONG,;
                  'lfWidth',           TYPE_LONG,;
                  'lfEscapement',      TYPE_LONG,;
                  'lfOrientation',     TYPE_LONG,;
                  'lfWeight',          TYPE_LONG,;
                  'lfItalic',          TYPE_BYTE,;
                  'lfUnderline',       TYPE_BYTE,;
                  'lfStrikeOut',       TYPE_BYTE,;
                  'lfCharSet',         TYPE_BYTE,;
                  'lfOutPrecision',    TYPE_BYTE,;
                  'lfClipPrecision',   TYPE_BYTE,;
                  'lfQuality',         TYPE_BYTE,;
                  'lfPitchAndFamily',  TYPE_BYTE,;
                  'lfFaceName',        TYPE_STRING(32))
      endif
      sStruct  =  new VStruct('LOGFONT')
   return   sStruct

   //------------------------------------------------------------//
   function pointStructure
      Local sStruct
      if not VSTRUCT_ISDEF("POINT")
         VSTRUCT_DEF('POINT',;
                  'x',                 TYPE_LONG,;
                  'y',                 TYPE_LONG)
      endif
      sStruct  =  new VStruct('POINT')
   return   sStruct

   //------------------------------------------------------------//
   function rectStructure
      Local sStruct
      if not VSTRUCT_ISDEF("RECT")
         VSTRUCT_DEF('RECT',;
                  'left',              TYPE_LONG,;
                  'top',               TYPE_LONG,;
                  'right',             TYPE_LONG,;
                  'bottom',            TYPE_LONG)
      endif
      sStruct  =  new VStruct('RECT')
   return   sStruct

   //------------------------------------------------------------//
   function speedTipStructure
      Local sStruct
      if not VSTRUCT_ISDEF("SPEEDTIP")
         VSTRUCT_DEF('SPEEDTIP',;
                  'lStructSize',          TYPE_DWORD,;
                  'alignment',            TYPE_SHORT,;
                  'padBottom',            TYPE_SHORT,;
                  'padHeight',            TYPE_SHORT,;
                  'padLeft',              TYPE_SHORT,;
                  'padRight',             TYPE_SHORT,;
                  'padTop',               TYPE_SHORT,;
                  'padWidth',             TYPE_SHORT,;
                  'gutter',               TYPE_SHORT,;
                  'above',                TYPE_BOOL,;
                  'mdiApp',               TYPE_BOOL)
      endif
      sStruct  =  new VStruct('SPEEDTIP')
      sStruct.setMem('lStructSize', sStruct.getSize())
   return   sStruct

   //------------------------------------------------------------//
   function textMetricStructure
      Local sStruct
      if not VSTRUCT_ISDEF("TEXTMETRIC")
         VSTRUCT_DEF('TEXTMETRIC',;
                  'tmHeight',          TYPE_LONG,;
                  'tmAscent',          TYPE_LONG,;
                  'tmDescent',         TYPE_LONG,;
                  'tmInternalLeading', TYPE_LONG,;
                  'tmExternalLeading', TYPE_LONG,;
                  'tmAveCharWidth',    TYPE_LONG,;
                  'tmMaxCharWidth',    TYPE_LONG,;
                  'tmWeight',          TYPE_LONG,;
                  'tmOverhang',        TYPE_LONG,;
                  'tmDigitizedAspectX',TYPE_LONG,;
                  'tmDigitizedAspectY',TYPE_LONG,;
                  'tmFirstChar',       TYPE_BCHAR,;
                  'tmLastChar',        TYPE_BCHAR,;
                  'tmDefaultChar',     TYPE_BCHAR,;
                  'tmBreakChar',       TYPE_BCHAR,;
                  'tmItalic',          TYPE_BYTE,;
                  'tmUnderlined',      TYPE_BYTE,;
                  'tmStruckOut',       TYPE_BYTE,;
                  'tmPitchAndFamily',  TYPE_BYTE,;
                  'tmCharSet',         TYPE_BYTE)
      endif
      sStruct  =  new VStruct('TEXTMETRIC')
   return   sStruct

endclass
//---------------------------------------------------------------//
//          end of SPEEDTIP_Structures.
//---------------------------------------------------------------//

