/*
      
   Version 1.3 
 
   UsvOut.prg
   Posted in dBASE, Inc. newsgroups, January 17, 2002
  

   This program will output a text-formatted file with
   User-Defined Separated Value from a table.

   ****

   Original Name: CsvOut.prg 
   Posted in dBASE, Inc. newsgroups, May 11, 1999
   by Gary White
 
   January 17, 2002
   E Sanders
   Version 1.0
   Added: PARAMETER cSeparator
   Because user can choose any separator,
   the program's Name is changed from CsvOut.prg to UsvOut.prg 

   October 24, 2007
   Version 1.1																 
   Ivar B. Jessen                                                     
   Added: PARAMETER bMemo                                             
   Allows user to include or not include data from memo fields        
   in the text-formatted file                                         

   February 26, 2008
   Version 1.2
   Roland Wingerter
   Changes:
   - Use CRLF instead of CR only
   - Replace double quotes with double double quotes in character data,
     because they are used as delimiters and otherwise cause problems 
     when importing. 
   - Added set proc to :dUFLP:stringEx.cc to be able to use strTran()

   January 18, 2016
   Version 1.3
   Mervyn Bick
   Added: If fieldnames are inclused in the .csv file and the first field 
      name is id or ID Excel sees the .csv file as a .slk file.  
      This can be prevented by prefixing the fieldname with a space.  
   Added: Version numbers  

   Usage:

   UsvOut( <rowset>, <filename>, <fieldnames>, <cSeparator>, <bMemo> )

   <rowset>     = the rowset object you want to export.
   <filename>   = character expression - the name of the file to create
   <fieldnames> = logical - true to include fieldnames, false,
   					if omitted, fieldnames are excluded.					
   <cSeparator> = character expression - the character used as the separator
                  if omitted, the comma is used as separator.
	<bMemo>		 = logical - true/false to include/exclude memo fields,
   					if omitted, memo fields are included                
   Examples:
     q = new query()
     q.sql              := "select * from MyTable"
     q.active           := true
     q.rowset.indexName := "SomeIndex"
     // if you want to filter the data:
     // q.rowset.filter = ...
     q.rowset.first()

     UsvOut( q.rowset, "MyCSVFile.CSV", false , "," )
   or  
     UsvOut( q.rowset, "MyCSVFile.CSV", false )
   or
     UsvOut( q.rowset, "MyASVFile.ASV", false , "'" )
   or
     UsvOut( q.rowset, "MyTSVFile.TSV", false , chr(9) )
   or                            
     UsvOut( q.rowset, "MyTSVFile.TSV", false , chr(9), false ) 
   or
     if not UsvOut( q.rowset, "MyCSVFile.TSV", false , chr(9) )
        msgbox("Problem with converting!","error",16)
     endif


*/

#define CR	chr(13)
#define CRLF	chr(13)+chr(10)	// Added R.W.

   parameters oRowset, cFileName, bFieldNames, cSeparator, bMemo	

   ** added R.W.
   set procedure to :dUFLP:stringEx.cc addi
   **

   private xValue, cSepa, cMem	

   cMem = ""                     

   do case                       
   case argcount() <= 3          
      cSepa = ","						
   case argcount() = 4           
      cSepa = cSeparator			
   otherwise                     
   	cSepa = cSeparator         
      if bMemo = false           
      	cMem = "MEMO"           
      endif                      
   endcase                       

   local bRetVal, i, f, cType
   bRetVal = true
   try
      if type('oRowset.filterOptions') == 'U'
         throw new csvError('Type mismatch - expecting rowset')
      endif
      if type('cFileName') # 'C' or empty(cFileName)
         throw new csvError('Invalid output file name')
      endif
      fOutFile = new file()
      fOutFile.create( cFileName, 'W' )
      oRowset.first()
      f = oRowset.fields
      if bFieldNames
         for i = 1 to f.size
            if f[i].fieldName # '_DBASELOCK' and ;
               not f[i].type $ 'OLE|BINARY|&cMem'
               *********** Added by Mervyn Bick  Jan 2016
               if i = 1 and upper(f[i].fieldname) = "ID"
                  fOutFile.write(" ")
               endif
               **********
               fOutFile.write(f[i].fieldName + ; 
                  iif( i < f.size, cSepa,'') )
            endif
         endfor
         fOutFile.write( CRLF )	// was CR -- changed by R.W. Feb 2008
      endif

      do while not oRowset.endOfSet
         for i = 1 to f.size

            if f[i].fieldName # '_DBASELOCK' and ;
               not f[i].type $ 'OLE|BINARY|&cMem'
               xValue = f[i].value
               if type('xValue') == 'C'
               	** Line added by R.W.
               	xValue := new stringEx().strTran(xValue, ["], [""])
                  ** 
                  xValue := ["]+ new string( trim( xValue) )+["]
               endif
               fOutFile.write( xValue + ;
                  iif( i < f.size, cSepa,'') )
            endif
         endfor
         fOutFile.write( CRLF )         // was CR -- changed by R.W. Feb 2008
         oRowset.next()
      enddo
   catch( exception e )
      msgbox( e.message, 'Error', 16 )
      bRetVal := false
   finally
      if type('fOutFile') == 'O' and ;
         fOutFile.handle > 0
         fOutFile.close()
         release object fOutFile
         release fOutFile
      endif
   endtry
   return bRetVal

class csvError( cMsg ) of exception
   this.message = cMsg
endclass
